/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "monitor.h"
#include "net.h"
#include "alert.h"

#define DO_TIMEOUT   0
#define DO_RESTART   1
#define DO_CHECKSUM  2
#define DO_RESOURCE  3
#define DO_STOP      4
#define DO_FAILED    5
#define DO_TIMESTAMP 6


/* Private Prototypes */
static int count(char *, char *);
static void copy_mail(Mail_T, Mail_T);
static void replace(char **, char *, char *);
static void smtp_alert(Process_T, int, char *, va_list);
static void substitute(Mail_T*, char *name, char *event);

/* Private Variables */
static char desc[][STRLEN]= {"timed out", "restarted", "checksum error", 
                             "matches resource limitation", "stopped",
                             "failed", "timestamp error"};

static char desclog[][STRLEN]= {"Timeout", "Restart", "Checksum error", 
                                "Resource limit matched", "Stop",
                                "Failed", "Timestamp error"};


/**
 *  Alert routines, at the moment only mail alert is used. In the
 *  future this could be a placeholder for other types of alert
 *  notification.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: alert.c,v 1.15 2003/02/14 08:22:33 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Send an alert timeout message to the email address for
 * this process
 * @param p A Process_T object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_timeout(Process_T p, char *m, ...) {

  va_list ap;

  ASSERT(p);
  
  va_start(ap, m);
  smtp_alert(p, DO_TIMEOUT, m, ap);
  va_end(ap);
  
  
}


/**
 * Send an alert checksum message to the email address for
 * this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_checksum(Process_T p, char *m, ...) {

  va_list ap;
  
  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_CHECKSUM, m, ap);
  va_end(ap);
  
}


/**
 * Send an alert restart message to the email address for this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_restart(Process_T p, char *m, ...) {

  va_list ap;
  
  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_RESTART, m, ap);
  va_end(ap);
  
}


/**
 * Send an alert resource message to the email address for this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_resource(Process_T p, char *m, ...) {

  va_list ap;
  
  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_RESOURCE, m, ap);
  va_end(ap);
  
}


/**
 * Send an alert stop message to the email address for this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_stop(Process_T p, char *m, ...) {

  va_list ap;
  
  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_STOP, m, ap);
  va_end(ap);
  
}


/**
 * Send an alert failed message to the email address for this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_failed(Process_T p, char *m, ...) {

  va_list ap;

  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_FAILED, m, ap);
  va_end(ap);

}


/**
 * Send an alert timestamp message to the email address for this process
 * @param p A process_t object
 * @param m An optional message string. May be NULL.
 */
void smtp_alert_timestamp(Process_T p, char *m, ...) {

  va_list ap;

  ASSERT(p);

  va_start(ap, m);
  smtp_alert(p, DO_TIMESTAMP, m, ap);
  va_end(ap);

}


/* ----------------------------------------------------------------- Private */


/*
 * Send a smtp notification 
 */
static void smtp_alert(Process_T p, int event, char *optmsg, va_list ap) {

  ASSERT(p);

  if(p->maillist) {

    Mail_T m;
    sigset_t ns, os;
    Mail_T list= NULL;

    /* Set a SIGUSR1 block here */
    sigemptyset(&ns);
    sigaddset(&ns, SIGUSR1);
    pthread_sigmask(SIG_BLOCK, &ns, &os);
    
    /*
     * Build a mail-list with recipients that has registered interest
     * for this event.
     */
    for(m= p->maillist; m; m= m->next) {
      
      int notify[]= { m->alert_on_timeout,
		      m->alert_on_restart,
		      m->alert_on_checksum,
		      m->alert_on_resource,
		      m->alert_on_stop,
		      m->alert_on_restart,
		      m->alert_on_timestamp};
      
      if ( notify[event] ) {

	Mail_T tmp= NEW(tmp);
	
	copy_mail(tmp, m);

	if(optmsg)
	    tmp->opt_message= format(optmsg, ap);
	
	substitute(&tmp, p->name, desc[event]);
	
	tmp->next= list;
	list= tmp;
	
        if ( Run.debug )
	    log("%s notification is sent to %s\n", desclog[event], m->to);

      }

    }

    if(list) {
      
      sendmail(list);
      gc_mail_list(&list);
      
    }

    /* Release the SIGUSR1 block */
    pthread_sigmask(SIG_SETMASK, &os, NULL);

  }
  
}


static void substitute(Mail_T *m, char *name, char *event) {

  char *now= get_ctime();
  char *host= get_localhostname();

  ASSERT(m && name && event);

  replace(&(*m)->from, "$HOST", host);
  replace(&(*m)->subject, "$DATE", now);
  replace(&(*m)->message, "$DATE", now);
  replace(&(*m)->subject, "$HOST", host);
  replace(&(*m)->message, "$HOST", host);
  replace(&(*m)->subject, "$PROGRAM", name);
  replace(&(*m)->message, "$PROGRAM", name);
  replace(&(*m)->subject, "$EVENT", event);
  replace(&(*m)->message, "$EVENT", event);
 
  free(now);
  free(host);

}


static void replace(char **src, char *old, char *new) {

  int i;
  int d;

  ASSERT(src && old && new);
  
  i= count(*src, old);
  d= strlen(new)-strlen(old);
  
  if(i==0)
      return;
  if(d>0)
      d*= i;
  else
      d= 0;
  
  {
    char *p, *q;
    int l= strlen(old);
    char buf[strlen(*src)+d+1];

    q= *src;
    *buf= '\0';
    
    while((p= strstr(q, old))) {
      
      *p= '\0';
      strcat(buf, q);
      strcat(buf, new);
      p+= l;
      q= p;
      
    }
    
    strcat(buf, q);
    free(*src);
    *src= xstrdup(buf);

  }
  
}


static int count(char *src, char *needle) {

  int i= 0;
  char *p= src;

  ASSERT(src && needle);
  
  while((p= strstr(p, needle))) { i++;  p++; }

  return i;

}
  

static void copy_mail(Mail_T n, Mail_T o) {

  ASSERT(n && o);
  
  n->to= xstrdup(o->to);
  n->from=
      o->from?
      xstrdup(o->from):
      Run.MailFormat.from?
      xstrdup(Run.MailFormat.from):
      xstrdup(ALERT_FROM);
  n->subject=
      o->subject?
      xstrdup(o->subject):
      Run.MailFormat.subject?
      xstrdup(Run.MailFormat.subject):
      xstrdup(ALERT_SUBJECT);
  n->message=
      o->message?
      xstrdup(o->message):
      Run.MailFormat.message?
      xstrdup(Run.MailFormat.message):
      xstrdup(ALERT_MESSAGE);
  n->opt_message= NULL;
  
}

