/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "monitor.h"
#include "net.h"
#include "ssl.h"


/* Private Prototypes */
static void wait_stop(Process_T);
static void wait_start(Process_T);
static void do_stop(Process_T);
static void do_start(Process_T);
static void do_depend(Process_T, char *);


/**
 *  Methods for controlling processes managed by monit.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Rory Toma, <rory@digeo.com>
 *  @author Martin Pala, <martin.pala@iol.cz>
 *
 *  @version \$Id: control.c,v 1.44 2003/02/14 08:22:33 martinp Exp $
 *
 *  @file
 */

/* ------------------------------------------------------------------ Public */


/**
 * Start/stop all processes in process list
 * @param action A string describing the action to execute
 */
void control(char *action) {

  Process_T p;

  ASSERT(action);

  for(p= processlist; p; p= p->next) {
    if(p->visited)
	continue;
    if(exist_daemon()) {
      d_check_process(p->name, action);
    } else {
      check_process(p->name, action);
    }
  }
  
  reset_depend();

}


/*
 * Start/stop all processes in a group
 * @param G group name
 * @param action A string describing the action to execute
 */
void control_group(char *G, char *action) {

  Process_T p;

  ASSERT(G);
  ASSERT(action);

  for(p= processlist; p; p= p->next) {
    if(p->visited)
	continue;
    if(is(p->group, G)) {
      if(exist_daemon()) {
	d_check_process(p->name, action);
      } else {
	check_process(p->name, action);
      }
    }
  }

  reset_depend();
 
}


/**
 * Pass on to methods in http/cervlet.c to start/stop processes
 * @param P A process name as stated in the config file
 * @param action A string describing the action to execute
 */
void d_check_process(char *P, char *action) {

  int s;
  char req[2*STRLEN];
  char *auth= get_basic_authentication_header();
  ssl_connection *ssl= NULL;

  ASSERT(P);
  ASSERT(action);

  if(Run.httpdssl) {
    
    ssl= new_ssl_connection(Run.httpsslpem, SSL_VERSION_AUTO);
    
  }

  s= create_socket(Run.bind_addr?Run.bind_addr:"localhost",
                   Run.httpdport, SOCK_STREAM);
  
  if(s < 0) {
    
    error("%s: Cannot connect to the monit daemon. "
          "Did you start it with http support?\n", prog);
    goto error;
    
  } else {
    snprintf(req, sizeof(req),
             "GET /%s?action=%s HTTP/1.0\r\n%s\r\n", P, action, auth);
    
    if(ssl) {
      
      if(!embed_ssl_socket(ssl, s)) {
	
	fprintf(stdout, "Failed to establish SSL communication to monit"
		" server\n");
	goto error;
      }
      
      send_ssl_socket(ssl, req, sizeof(req));
      
      close_ssl_socket(ssl);
      close_socket(s);
      
    } else {
      
      sock_send(s, req, sizeof(req), 0);
      close_socket(s);
      
    }
  }
  
  if(Run.httpdssl) {
    
    delete_ssl_socket(ssl);

  }
  
  error:
  free(auth);
  
}


/**
 * Check to see if we should try to start/stop process
 * @param P A process name as stated in the config file
 * @param action A string describing the action to execute
 */
void check_process(char *P, char *action) {

  Process_T p= NULL;

  ASSERT(P);
  ASSERT(action);

  if(NULL==(p= get_process(P))) {
    error("%s: Cannot %s program '%s' -- not found in %s\n",
          prog, action, P, Run.controlfile);
    return;
  }

  if(is(action, "start")) {
    
    if(is_process_running(p)) {

      if(!p->do_validate) {

        LOCK(Run.mutex)
          p->do_validate= TRUE;
        END_LOCK;

        if(Run.debug)
          log("Monitoring enabled -- process %s\n", p->name);

      }

      return;
    }
    
    if(!p->start) {
      error("%s: Start method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(p, "stop");
    do_start(p);
    do_depend(p, "start");
    
  } else if(is(action, "stop")) {
    
    if(!p->stop) {
      error("%s: Stop method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(p, "stop");
    do_stop(p);
    
  } else if(is(action, "restart")) {
    
    if(!p->start || !p->stop) {
      error("%s: Start or stop method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(p, "stop");
    do_stop(p);
    do_start(p);
    do_depend(p, "start");
    
  }
    
}


/*
 * Reset the visited flags used when handling dependencies
 */
void reset_depend() {

  Process_T p;
  
  for (p= processlist; p; p= p->next) {
    p->visited= FALSE;
    p->depend_visited= FALSE;
  }

}


/* ----------------------------------------------------------------- Private */


/*
 * This is a post- fix recursive function for starting every process
 * that p depends on before starting p.
 * @param p A Process_T object
 */
static void do_start(Process_T p) {

  ASSERT(p);

  if(p->visited)
      return;
  
  p->visited= TRUE;
  
  if(p->dependantlist) {
    
    Dependant_T d;
    
    for(d= p->dependantlist; d; d= d->next ) {
      
      Process_T dp= get_process(d->dependant);
      ASSERT(dp);
      do_start(dp);
      
    }
  }
  
  if(p->start && (!is_process_running(p))) {
    log("start: (%s) %s\n", p->name, p->start->arg[0]);
    spawn(p, p->start);
    wait_start(p);
  }
  
  LOCK(Run.mutex)
    p->do_validate= TRUE;
  END_LOCK;

  if(Run.debug)
    log("Monitoring enabled -- process %s\n", p->name);

}


/*
 * This function simply stops the process p.
 * @param p A Process_T object
 */
static void do_stop(Process_T p) {

  ASSERT(p);

  if(p->depend_visited)
      return;
  
  p->depend_visited= TRUE;
  
  LOCK(Run.mutex)
    p->do_validate= FALSE;
  END_LOCK;

  if(Run.debug)
    log("Monitoring disabled -- process %s\n", p->name);

  if(p->stop && is_process_running(p)) {
    log("stop: (%s) %s\n", p->name, p->stop->arg[0]);
    spawn(p, p->stop);
    wait_stop(p);
  }

  /* Reset the proc info object in case of a later restart */
  memset(p->procinfo, 0, sizeof *(p->procinfo));
  
}


/*
 * This is an in-fix recursive function called before p is started to
 * stop every process that depends on p, in reverse order *or* after p
 * was started to start again every process that depends on p. The
 * action parametere controls if this function should start or stop
 * the procceses that depends on p.
 * @param p A Process_T object
 * @param action An action to do on the dependant processes
 */
static void do_depend(Process_T p, char *action) {

  Process_T parent;
  
  ASSERT(p);

  for(parent= processlist; parent; parent= parent->next) {
    
    if(parent->dependantlist) {

      Dependant_T d;
    
      for(d= parent->dependantlist; d; d= d->next)
	  if(is(d->dependant, p->name))
	     break;
      
      if(d) {
	
	if(is(action, "start"))
	    do_start(parent);
	
	do_depend(parent, action);
	
	if(is(action, "stop"))
	    do_stop(parent);
	
      }
    }
  }
}
    

/*
 * This function suspend the control until the process p is running.
 * @param p A Process to wait for
 */
static void wait_start(Process_T p) {

  int max_tries= Run.polltime;
  
  ASSERT(p);

  while(max_tries--) {
    if(is_process_running(p))
	break;
    sleep(1);
  }
  
  if(!is_process_running(p))
      log("%s: Warning process '%s' was not started\n", prog, p->name);
  
}


/*
 * This function suspend the control until the process p is stopped.
 * @param p A Process to wait for
 */
static void wait_stop(Process_T p) {

  int max_tries= Run.polltime;
  
  ASSERT(p);

  while(max_tries--) {
    if(!is_process_running(p))
	break;
    sleep(1);
  }

  if(is_process_running(p))
      log("%s: Warning process '%s' was not stopped\n", prog, p->name);

}
