/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "monitor.h"
#include "cervlet.h" 
#include "engine.h"
#include "processor.h"
#include "base64.h"
#include "alert.h"
#include "monit_process.h"

#define ACTION(c) !strncasecmp(req->url, c, sizeof(c))


/* URL Commands supported */
#define HOME        "/"
#define STOP        "/_stop"
#define TEST        "/_monit"
#define ABOUT       "/_about"
#define PIXEL       "/_pixel"
#define RUN         "/_runtime"
#define VIEWLOG     "/_viewlog"

/* Private prototypes */
static void printPixel(HttpResponse);
static void doGet(HttpRequest, HttpResponse);
static void doPost(HttpRequest, HttpResponse);
static void do_home(HttpRequest, HttpResponse);
static void do_about(HttpRequest, HttpResponse);
static void not_found(HttpRequest, HttpResponse);
static void do_runtime(HttpRequest, HttpResponse);
static void do_viewlog(HttpRequest, HttpResponse);
static void handle_action(HttpRequest, HttpResponse);
static void print_status(Process_T, HttpResponse res);
static void is_monit_running(HttpRequest, HttpResponse);
static void do_process(HttpRequest, HttpResponse, char *);

extern ssl_server_connection * mySSLServerConnection;

/**
 *  Implementation of doGet and doPost routines used by the cervlet
 *  processor module. This particilary cervlet will provide
 *  information about the monit deamon and programs monitored by
 *  monit.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: cervlet.c,v 1.54 2003/02/14 08:22:33 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Callback hook to the Processor module for registering this modules
 * doGet and doPost methods.
 */
void init_service() {
  
  add_Impl(doGet, doPost);
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Called by the Processor (via the service method)
 * to handle a POST request.
 */
static void doPost(HttpRequest req, HttpResponse res) {

  doGet(req, res);

}


/**
 * Called by the Processor (via the service method)
 * to handle a GET request.
 */
static void doGet(HttpRequest req, HttpResponse res) {

  set_content_type(res, "text/html");

  if ( ACTION(HOME) ) {
    
    LOCK(Run.mutex)
	do_home(req, res);
    END_LOCK;
    
  }
  else if ( ACTION(RUN) ) {
    
    LOCK(Run.mutex)
	do_runtime(req, res);
    END_LOCK;
    
  }
  else if ( ACTION(TEST) ) {
    
    is_monit_running(req, res);
    
  }
  else if ( ACTION(VIEWLOG) ) {
    
    do_viewlog(req, res);
    
  }
  else if ( ACTION(ABOUT) ) {
    
    do_about(req, res);
    
  }
  else if ( ACTION(STOP) ) {
    
    send_error(res, SC_SERVICE_UNAVAILABLE, "The monit http server is stopped");
    stop_httpd();
    
  }  
  else if ( ACTION(PIXEL) ) {
    
    printPixel(res);
    
  } else {
    
    handle_action(req, res);
    
  }
  
   
}


/* ----------------------------------------------------------------- Helpers */


static void is_monit_running(HttpRequest req, HttpResponse res) {

  int monit= exist_daemon();
  int status;
  char *msg= NULL;

  if ( monit ) {
    
    status= SC_OK;
    msg= get_status_string(status);
    
  } else {
    
    status= SC_GONE;
    msg= get_status_string(status);
    
  }

  set_status(res, status, msg);

  free(msg);
  
}
    

static void do_home(HttpRequest req, HttpResponse res) {

  int on= TRUE;
  struct myprocess *p;
  char *uptime= get_process_uptime(Run.pidfile);
 
  if ( Run.doprocess ) {
  HEAD("", Run.polltime)
      out_print(res,
"<table cellspacing=\"0\" cellpadding=\"5\" width=\"100%%\" border=\"0\">"
" <tr bgcolor=\"#BBDDFF\">"
"  <td colspan=2 valign=\"top\" align=\"left\" bgcolor=\"#EFF7FF\" width=\"100%%\">"
"  <br><h2 align=\"center\">Monit Process Manager</h2>"
"  <p align=\"center\">Monit is <a href='/_runtime'>running</a> on %s "
"  with <i>uptime, %s</i> and monitoring:</p><br>"
"  </td>"
" </tr>"
"</table>"
"<table cellspacing=\"0\" cellpadding=\"0\" width=\"100%%\" border=\"0\">"
"  <tr valign=\"middle\" bgcolor=\"#6F6F6F\">"
"    <td><img src=\"/_pixel\" width=\"1\" height=\"1\" alt=\"\"></td>"
"  </tr>"
"</table>"
"<br><p>&nbsp;</p>"
"<table cellspacing=0 cellpadding=3 border=0 width=\"70%\">"
"<tr><td><h3><b>Server</b></h3></td>"
"<td align=\"right\"><h3><b>Status</b></h3></td>"
"<td align=\"right\"><h3><b>Uptime</b></h3></td>"
"<td align=\"right\"><h3><b>CPU</b></h3></td>"
"<td align=\"right\"><h3><b>Memory</b></h3></td>"
"</tr>"
"<tr><td>&nbsp;</td><td>&nbsp;</td></tr>", Run.localhostname, uptime);
  } else {
  HEAD("", Run.polltime)
      out_print(res,
"<table cellspacing=\"0\" cellpadding=\"5\" width=\"100%%\" border=\"0\">"
" <tr bgcolor=\"#BBDDFF\">"
"  <td colspan=2 valign=\"top\" align=\"left\" bgcolor=\"#EFF7FF\" width=\"100%%\">"
"  <br><h2 align=\"center\">Monit Process Manager</h2>"
"  <p align=\"center\">Monit is <a href='/_runtime'>running</a> on %s "
"  with <i>uptime, %s</i> and monitoring:</p><br>"
"  </td>"
" </tr>"
"</table>"
"<table cellspacing=\"0\" cellpadding=\"0\" width=\"100%%\" border=\"0\">"
"  <tr valign=\"middle\" bgcolor=\"#6F6F6F\">"
"    <td><img src=\"/_pixel\" width=\"1\" height=\"1\" alt=\"\"></td>"
"  </tr>"
"</table>"
"<br><p>&nbsp;</p>"
"<table cellspacing=0 cellpadding=3 border=0 width=\"70%\">"
"<tr><td><h3><b>Server</b></h3></td>"
"<td align=\"right\"><h3><b>Status</b></h3></td></tr>"
"<tr><td>&nbsp;</td><td>&nbsp;</td></tr>", Run.localhostname, uptime);
  }

  free(uptime);
  
  for (p= processlist; p; p= p->next) {
    
    int isrunning= is_process_running(p);
    char *uptime= get_process_uptime(p->pidfile);

    if ( Run.doprocess ) {
      ProcInfo_T pi= p->procinfo;
      if ( isrunning ) {
	out_print(res,"<tr %s><td><a href='/%s'>%s</a></td><td align=\"right\">"
		  "%s</td><td align=\"right\">%s</td>"
		  "<td align=\"right\">%.1f%%</td>"
		  "<td align=\"right\">%.1f%% [%ldkB]</td></tr>",
		  on?"bgcolor=\"#EFEFEF\"":"",
		  p->name, p->name,
		  p->has_checksum_error?
		  "<font color='#ff0000'>Cheksum Error</font>":	
		  isrunning?
		  "<font color=green>running</font>":
		  "<font color=red>not running</font>",
		  uptime,
		  pi->cpu_percent/10.0,
		  pi->mem_percent/10.0,pi->mem_kbyte
		  );
      } else {
	out_print(res,"<tr %s><td><a href='/%s'>%s</a></td>"
		  "<td align=\"right\">%s</td>"
		  "<td align=\"right\">-</td>"
		  "<td align=\"right\">- </td>"
		  "<td align=\"right\">- [-]</td></tr>",
		  on?"bgcolor=\"#EFEFEF\"":"",
		  p->name, p->name,
		  p->has_checksum_error?
		  "<font color='#ff0000'>Cheksum Error</font>":	
		  "<font color=red>not running</font>"
		  );
      }
    } else {
      out_print(res,"<tr %s><td><a href='/%s'>%s</a></td><td align=\"right\">"
		"%s %s</td></tr>",
		on?"bgcolor=\"#EFEFEF\"":"",
		p->name, p->name,
		p->has_checksum_error?
		"<font color='#ff0000'>Cheksum Error</font>":	
		isrunning?
		"<font color=green>running</font>":
		"<font color=red>not running</font>",
		isrunning?uptime:"");
    }


    on= on?FALSE:TRUE;
    free(uptime);
    
  }
  out_print(res, "</table>");
  FOOT
      
}


static void do_about(HttpRequest req, HttpResponse res) {

  out_print(res,
    "<html><head><title>about monit</title><body bgcolor=white>"
    "<br><br><center><a href='http://www.tildeslash.com/monit/'><font size=+2>"
    "monit " VERSION "</font></a></center><br><br>");
  out_print(res,
    "<ul>"
    "<li>Copyright &copy; 2000-2003 by <a "
    "href=\"http://www.tildeslash.com/monit/who.html\">Contributors</a> "
    "to the monit codebase. All Rights Reserved.<br> </small><p>"
    "<li>Portions of this software are copyright &copy; 1995, 1996 "
    "<a href='http://www.gnu.org/'>Free Software Foundation, Inc.</a>"
    "</ul><p><br><p>");
  out_print(res,
   "<table cellspacing=\"0\" cellpadding=\"0\" width=\"100%%\" border=\"0\">"
   " <tr bgcolor=\"#6F6F6F\">"
   " <td valign=\"bottom\"><img src=\"/_pixel\" width=\"1\" height=\"1\" "
   "alt=\"\"></td></tr></table>");
  out_print(res,
    "<p>This program is free software; you can redistribute it and/or "
    "modify it under the terms of the GNU General Public License "
    "as published by the Free Software Foundation; either version 2 of "
    "the License, or (at your option) any later version.<p>"
    "This program is distributed in the hope that it will be useful, but "
    "WITHOUT ANY WARRANTY; without even the implied warranty of "
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
    "<a href='http://www.tildeslash.com/gpl.html'>GNU "
    "General Public License</a> for more details.");

}


static void do_runtime(HttpRequest req, HttpResponse res) {

  int pid=  exist_daemon();
  
  HEAD("_runtime", 1000)
  out_print(res,
	    "<center><h3>monit runtime status</h3><center><br>");
  out_print(res,"<table cellspacing=0 cellpadding=3 border=1 width=\"90%\">"
	    "<tr><td width=\"40%\"><b>Parameter</b></td>"
	    "<td width=\"60%\"><b>Value</b></td></tr>");
  out_print(res,
	    "<tr><td>Host</a></td><td>%s</td></tr>",
	    Run.localhostname);
  out_print(res,
	    "<tr><td>Process id</a></td><td>%d</td></tr>", pid);
  out_print(res,
	    "<tr><td>Effective user running monit</a></td>"
	    "<td>%s</td></tr>", Run.Env.user);
  out_print(res,
	    "<tr><td>Controlfile</a></td><td>%s</td></tr>", Run.controlfile);
  out_print(res,
	    "<tr><td>Logfile</a></td><td>%s</td></tr>",
	    Run.logfile?Run.logfile:"Not defined");
  out_print(res,
	    "<tr><td>Pidfile</a></td><td>%s</td></tr>", Run.pidfile);
  out_print(res,
	    "<tr><td>Debug</a></td><td>%s</td></tr>",
	    Run.debug?"True":"False");
  out_print(res,
	    "<tr><td>Log</a></td><td>%s</td></tr>", Run.dolog?"True":"False");
  out_print(res,
	    "<tr><td>Use syslog</a></td><td>%s</td></tr>",
	    Run.use_syslog?"True":"False");
  out_print(res,
	    "<tr><td>Mail server</a></td><td>%s</td></tr>",
	    Run.mailserver?Run.mailserver:"Not defined");
  out_print(res,
	    "<tr><td>Default mail from</a></td><td>%s</td></tr>",
            Run.MailFormat.from?Run.MailFormat.from:"Not defined");
  out_print(res,
	    "<tr><td>Default mail subject</a></td><td>%s</td></tr>",
            Run.MailFormat.subject?Run.MailFormat.subject:"Not defined");
  out_print(res,
	    "<tr><td>Default mail message</a></td><td>%s</td></tr>",
            Run.MailFormat.message?Run.MailFormat.message:"Not defined");
  out_print(res,
	    "<tr><td>Poll time</a></td><td>%d seconds</td></tr>",
	    Run.polltime);
  out_print(res,
	    "<tr><td>httpd bind address</a>"
	    "</td><td>%s</td></tr>", Run.bind_addr?Run.bind_addr:"Any/All");
  out_print(res,
	    "<tr><td>httpd portnumber</a>"
	    "</td><td>%d</td></tr>", Run.httpdport);
  out_print(res,
	    "<tr><td>Use ssl encryption</a>"
	    "</td><td>%s</td></tr>", Run.httpdssl?"True":"False");
  if (Run.httpdssl) {
    out_print(res,
	      "<tr><td>PEM key/certificate file</a>"
	      "</td><td>%s</td></tr>", Run.httpsslpem);
    
    if (Run.httpsslclientpem!=NULL) {
      out_print(res,
		"<tr><td>Client PEM key/certification</a>"
		"</td><td>%s</td></tr>", "Enabled");
      out_print(res,
		"<tr><td>Client PEM key/certificate file</a>"
		"</td><td>%s</td></tr>", Run.httpsslclientpem);
    } else {
      out_print(res,
		"<tr><td>Client PEM key/certification</a>"
		"</td><td>%s</td></tr>", "Disabled");
    }
    out_print(res,
	      "<tr><td>Allow self certified certificates </a>"
	      "</td><td>%s</td></tr>", Run.allowselfcert?"True":"False");
  }
  
  out_print(res,
	    "<tr><td>httpd auth. style</a></td><td>%s</td></tr>",
	    Run.Auth.defined&&has_hosts_allow()?
	    "Basic Authentication and Host allow list":
	    Run.Auth.defined?"Basic Authentication":
	    has_hosts_allow()?"Host allow list":
	    "No authentication");
  out_print(res,"</table>");
  out_print(res, "<table cellspacing=16><tr nowrap><td><font size=+1>"
	    "<font color='#f0000'>Stop monit http server?</font></font>"
	    "<td align=right><form method=GET action=%s>"
	    "<input type=submit value=Go style='font-size: 12pt'></font>"
	    "</form></td>", STOP);
  if ( Run.dolog && !Run.use_syslog ) {
    out_print(res, "<td><font size=+1>"
	      "View monit logfile?</font>"
	      "<td align=right><form method=GET action=%s>"
	      "<input type=submit value=Go style='font-size: 12pt'></font>"
	      "</form></td>", VIEWLOG);
  }
  out_print(res, "</tr></table>");
  
  FOOT

}


static void do_viewlog(HttpRequest req, HttpResponse res) {

  HEAD("_viewlog", 100)
      
  if ( Run.dolog && !Run.use_syslog ) {
    
    struct stat sb;
    
    if ( !stat(Run.logfile, &sb) ) {
      
      FILE *f= fopen(Run.logfile, "r");
      if ( f ) {
	
	int n;
	char buf[8192];
	
	out_print(res, "<br><p><form><textarea cols=80 rows=30>");
	
	while ( (n= fread(buf, sizeof(char), 8192, f)) > 0 ) {
	  
	  buf[n]= 0;
	  out_print(res, "%s", buf);
	  
	}
	
        fclose(f);
	out_print(res, "</textarea></form>");
	
      } else {
	
	out_print(res, "Error opening logfile: %s", STRERROR);
	
      }
      
    } else {
      
      out_print(res, "Error stating logfile: %s", STRERROR);
      
    }
    
  } else {
    
    out_print(res,
    "<b>Cannot view logfile:</b><br>");
    if ( !Run.dolog ) {
      
      out_print(res, "monit was started without logging");
      
    } else {
      
      out_print(res, "monit uses syslog");
      
    }
    
  }
  
  FOOT
  
}


static void handle_action(HttpRequest req, HttpResponse res) {

  char *name= req->url;
  char *action= get_parameter(req, "action");
  
  if(exist_process(++name)) {
    
    if(action) {
      
      struct myprocess *p= get_process(name);
      
      if( is(action, "start") ) {

	if(p->start) {

	  LOCK(p->mutex)
	    check_process(name, action);
	  END_LOCK;

	} else {

          send_error(res, SC_BAD_REQUEST,
	             "Start method not defined for the process");
	  goto quit;

	}
	
      }
      
      if( is(action, "stop") ) {

	if(p->stop) {

	  LOCK(p->mutex)
	    check_process(name, action);
	  END_LOCK;

	} else {

          send_error(res, SC_BAD_REQUEST,
	             "Stop method not defined for the process");
	  goto quit;

	}

      }
	
      if( is(action, "restart") ) {

	if(p->start && p->stop) {
 
	  LOCK(p->mutex)
	    check_process(name, action);
	  END_LOCK;

	} else {

          send_error(res, SC_BAD_REQUEST,
	    "Start or stop method not defined for the process");
	  goto quit;

	}

      }
	
      if(is(action, "status")) {

	print_status(p, res);
	goto quit;
	
      }
      
    }
    
    LOCK(Run.mutex)
	do_process(req, res, name);
    END_LOCK;
    
  } else {
    
    not_found(req, res);
    
  }

  quit:
  free(action);
  reset_depend();
  
}


static void do_process(HttpRequest req, HttpResponse res, char *name) {
  
  struct myprocess *p= get_process(name);
  int run= is_process_running(p);

  HEAD(name, 1000)
  out_print(res,
    "<p><br><h3>Process status</h3><br>");
  out_print(res,"<table cellspacing=0 cellpadding=3 border=1 width=\"90%\">"
	    "<tr><td width=\"30%\"><b>Parameter</b></td><td width=\"70%\"><b>Value</b></td></tr>");
  out_print(res,
    "<tr><td>Name</a></td><td>%s</td></tr>", p->name);
  out_print(res,
    "<tr><td>Group</a></td><td><font color='#0000ff'>%s</font></td></tr>",
	    p->group?p->group:"(not defined)");
  out_print(res,
    "<tr><td>Process id </a></td><td>%d</td></tr>", run);
  out_print(res,
    "<tr><td>Process status</a></td><td>%s %s</td></tr>",
	    run?"<font color='#00ff00'>Running</font>":
	    "<font color='#ff0000'>Not Running</font>",
	    p->has_checksum_error?
	    "<font color='#ff0000'><b>Checksum error!</b></font>":"");
  out_print(res,
    "<tr><td>Pid file</a></td><td>%s</td></tr>",
	    p->pidfile);
  out_print(res,
           "<tr><td>Monitoring mode</a></td><td>%s</td></tr>",
           modenames[p->mode]);
  out_print(res,
           "<tr><td>Monitoring status</a></td><td>%s</td></tr>",
           statusnames[p->do_validate]);
  out_print(res,
    "<tr><td>Start program</a></td><td>%s</td></tr>",
	    p->start?p->start->arg[0]:"(not defined)");
  out_print(res,
    "<tr><td>Stop program</a></td><td>%s</td></tr>",
	    p->stop?p->stop->arg[0]:"(not defined)");
  {
    struct mychecksum *c;
    for(c= p->checksumlist; c; c= c->next)
	out_print(res, "<tr><td>Associated checksum</a></td><td>%s %s</td>"
		  "</tr>", c->md5, c->file);
  }
  {
    Dependant_T d;

    for(d= p->dependantlist; d; d= d->next) {
      if(d->dependant != NULL) {
	out_print(res,"<tr><td>Depends on Process </a></td>"
		  "<td> <a href=%s> %s </a></td></tr>",
		  d->dependant, d->dependant);
      }
    }
  }
  if ( Run.doprocess ) {
    ProcInfo_T pi= p->procinfo;

    out_print(res,
	      "<tr><td>CPU usage</a></td><td>%.1f%%</td></tr>",
	      pi->cpu_percent/10.0);
    out_print(res,
	      "<tr><td>Memory usage</a></td><td>%.1f%% [%ldkB]</td></tr>",
	      pi->mem_percent/10.0,pi->mem_kbyte);
  }
  {
    Port_T n;
    for(n= p->portlist; n; n= n->next) {

      if ( n->family == AF_INET ) {

	if ( n->ssl != NULL ) {

	  out_print(res,
		    "<tr><td>Host:Port</a></td><td>%s:%d%s [%s via SSL]</td></tr>",
		    n->hostname, n->port, n->request?n->request:"",
		    n->protocol->name);
	  
	  if ( n->certmd5 != NULL ) {

	    out_print(res,
		      "<tr><td>Server certificate md5 sum</a></td><td>%s</td></tr>",
		      n->certmd5);

	  }


	} else {

	  out_print(res,
		    "<tr><td>Host:Port</a></td><td>%s:%d%s [%s]</td></tr>",
		    n->hostname, n->port, n->request?n->request:"",
		    n->protocol->name);

	}

      } else if ( n->family == AF_UNIX ) {

	out_print(res,
		"<tr><td>Unix Socket</a></td><td>%s [%s]</td></tr>",
		n->pathname, n->protocol->name);

      }
    }
  }
  {
    struct mytimestamp *t;
    for(t= p->timestamplist; t; t= t->next)
	out_print(res,
	  "<tr><td>Associated timestamp</a></td>"
          "<td>If %s %s %d second(s) then %s</td></tr>",
	  t->pathname, operatornames[t->operator], t->time, actionnames[t->action]);
  }
  {
    Resource_T q;

    for (q= p->resourcelist; q; q= q->next) {
      switch (q->resource_id) {

      case RESOURCE_ID_CPU_PERCENT: 

	out_print(res,"<tr><td>CPU usage limit</a></td>"
		  "<td>If %s %.1f%% for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit/10.0, q->max_cycle, 
		  actionnames[q->action]);
	break;

      case RESOURCE_ID_MEM_PERCENT: 
	
	out_print(res,"<tr><td>Memory usage limit</a></td>"
		  "<td>If %s %.1f%% for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit/10.0, q->max_cycle, 
		  actionnames[q->action]);
	break;
	
      case RESOURCE_ID_MEM_KBYTE: 
      
	out_print(res,"<tr><td>Memory amount limit</a></td>"
		  "<td>If %s %ld for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit, q->max_cycle, 
		  actionnames[q->action]);
	break;
	
      case RESOURCE_ID_LOAD1: 
      
	out_print(res,"<tr><td>Load average (1min)</a></td>"
		  "<td>If %s %.1f for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit/10.0, q->max_cycle, 
		  actionnames[q->action]);
	break;
	
      case RESOURCE_ID_LOAD5: 
      
	out_print(res,"<tr><td>Load average (5min)</a></td>"
		  "<td>If %s %.1f for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit/10.0, q->max_cycle, 
		  actionnames[q->action]);
	break;
	
      case RESOURCE_ID_LOAD15: 
	out_print(res,"<tr><td>Load average (15min)</a></td>"
		  "<td>If %s %.1f for %d cycle(s) then %s</td></tr>", 
		  operatornames[q->operator],
		  q->limit/10.0, q->max_cycle, 
		  actionnames[q->action]);
	break;
	
      }    
    }
  }
  out_print(res,
	    "<tr><td>Check process</a></td><td>Every %d cycle</td></tr>",
	    p->every?p->every:1);
  out_print(res,
    "<tr><td>Timeout</a></td><td>Timeout if %d restart within %d cycles"
    "</td></tr>", p->to_start, p->to_cycle);
  {
    Mail_T r;
    for(r= p->maillist; r; r= r->next) {
      out_print(res,
        "<tr bgcolor=\"#EFEFEF\"><td>Alert mail to</a></td><td>%s</td></tr>",
		r->to?r->to:"");
      out_print(res,
	"<tr><td>Alert from</a></td><td>%s</td></tr>",
              r->from?r->from:"(default)");
      out_print(res,
        "<tr><td>Alert subject</a></td><td>%s</td></tr>",
	      r->subject?r->subject:"(default)");
      out_print(res,
	"<tr><td>Alert message</a></td><td>%s</td></tr>",
	      r->message?r->message:"(default)");
      out_print(res,
	"<tr><td>Alert on timeout</a></td><td>%s</td></tr>",
		r->alert_on_timeout?"yes":"no");
      out_print(res,
	"<tr><td>Alert on restart</a></td><td>%s</td></tr>",
		r->alert_on_restart?"yes":"no");
      out_print(res,
	"<tr><td>Alert on checksum</a></td><td>%s</td></tr>",
		r->alert_on_checksum?"yes":"no");
      out_print(res,
	"<tr><td>Alert on exceeded resource</a></td><td>%s</td></tr>",
		r->alert_on_resource?"yes":"no");
      out_print(res,
	"<tr><td>Alert on stop</a></td><td>%s</td></tr>",
		r->alert_on_stop?"yes":"no");
      out_print(res,
	"<tr><td>Alert on timestamp</a></td><td>%s</td></tr>",
		r->alert_on_timestamp?"yes":"no");
    }
  }
  out_print(res, "</table>");
  out_print(res, "<table cellspacing=16><tr nowrap><td><font size=+1>");
  /* Start program */
  if(p->start)
     out_print(res, 
       "<td><form method=GET action=/%s>"
       "<input type=hidden value='start' name=action>"
       "<input type=submit value='Start program' style='font-size: 12pt'></font>"
       "</form></td>", name);
   /* Stop program */
  if(p->stop)
    out_print(res, 
       "<td><form method=GET action=/%s>"
       "<input type=hidden value='stop' name=action>"
       "<input type=submit value='Stop program' style='font-size: 12pt'></font>"
       "</form></td>", name);
  if(p->start && p->stop)
    out_print(res, 
       "<td><form method=GET action=/%s>"
       "<input type=hidden value='restart' name=action>"
       "<input type=submit value='Restart program' style='font-size: 12pt'></font>"
       "</form></td>", name);
  out_print(res, "</tr></table>");

  FOOT

  
}


static void print_status(Process_T p, HttpResponse res) {

  pid_t  pid= -1;
  FILE *out= res->outputstream;
  
  res->is_committed= TRUE;
  
  if((pid= is_process_running(p))) {
    
    char *uptime= get_process_uptime(p->pidfile);
    ProcInfo_T pi= p->procinfo;

    if( Run.httpdssl ) {

      printf_ssl_socket(res->ssl, 
	"Process '%s' is running with pid [%d]\n\tUptime: %s "
	"CPU: %.1f%% Memory: %.1f%% [%ldkB]\n\tMonitoring status: %s\n",
	p->name, (int)pid, uptime, pi->cpu_percent/10.0,
	pi->mem_percent/10.0,pi->mem_kbyte, statusnames[p->do_validate]);
      
    } else {

      fprintf(out, 
	"Process '%s' is running with pid [%d]\n\tUptime: %s "
	"CPU: %.1f%% Memory: %.1f%% [%ldkB]\n\tMonitoring status: %s\n",
	p->name, (int)pid, uptime, pi->cpu_percent/10.0,
	pi->mem_percent/10.0,pi->mem_kbyte, statusnames[p->do_validate]);

    }    

    free(uptime);
    
  } else {
    
    if( Run.httpdssl ) {

      printf_ssl_socket(res->ssl, "Process '%s' is not running\n",  p->name);

    } else {

      fprintf(out, "Process '%s' is not running\n",  p->name);

    }
    
  }

}


static void printPixel(HttpResponse res) {

  static int l;
  static unsigned char *pixel= NULL;

  if ( ! pixel ) {

    pixel= xcalloc(sizeof(unsigned char), strlen(PIXEL_GIF));
    l= decode_base64(pixel, PIXEL_GIF);

  }
  if (l) {
    
    if ( Run.httpdssl ) {

      res->is_committed= TRUE;

      printf_ssl_socket(res->ssl, "HTTP/1.0 200 OK\r\n");
      printf_ssl_socket(res->ssl, "Content-length: %d\r\n", l);
      printf_ssl_socket(res->ssl, "Content-Type: image/gif\r\n");
      printf_ssl_socket(res->ssl, "Connection: close\r\n\r\n");
      
      send_ssl_socket(res->ssl, pixel, l);

    } else {

      FILE *out= res->outputstream;
      
      res->is_committed= TRUE;
      
      fprintf(out, "HTTP/1.0 200 OK\r\n");
      fprintf(out, "Content-length: %d\r\n", l);
      fprintf(out, "Content-Type: image/gif\r\n");
      fprintf(out, "Connection: close\r\n\r\n");
      
      fwrite(pixel, sizeof(unsigned char), l, out);
    
    }
  
  }
  
}


static void not_found(HttpRequest req, HttpResponse res) {
  
  send_error(res, SC_NOT_FOUND,
	     "The requested URL was not found on this server");
  
}

