/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>
#include <time.h>

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "http_utils.h" 
#include "engine.h" 

#define  HOST      1
#define  WRAPPER   2
#define  INETADDR  3

/* Private variables */
static int myServerSocket= 0;
ssl_server_connection * mySSLServerConnection= NULL;
static volatile int stopped= FALSE;
static HostsAllow hostlist= NULL;
static pthread_mutex_t hostlist_mutex= PTHREAD_MUTEX_INITIALIZER;

/* Private prototypes */
static void initialize_service();
static void check_Impl();
static int  authenticate(RequestWrapper);
static int  is_host_allow(char *);
static RequestWrapper socket_producer(int);
static RequestWrapper create_wrapper(struct in_addr, int);
static void destroy_host_allow(HostsAllow);

/**
 *  A http 1.0 server. The server delegates handling of a HTTP request
 *  and response to the processor module.
 *
 *  NOTE
 *    This server does not use threads or forks; Requests are
 *    serialized and pending requests will be popped from the
 *    connection queue when the current request finish.
 *
 *    Since this server is written for monit, low traffic is expected.
 *    Connect from not-authenicated clients will be closed down
 *    promptly. The authentication schema or access control is based
 *    on client name/address and only requests from known clients are
 *    accepted. Hosts allowed to connect to this server should be
 *    added to the access control list by calling add_host_allow().
 *
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: engine.c,v 1.23 2003/02/18 01:19:01 rory Exp $
 *
 *  @file 
 */


/* ------------------------------------------------------------------ Public */


/**
 * Start the HTTPD server
 * @param port The Port number to start the server at
 * @param backlog The maximum length of the incomming connection queue 
 * @param bindAddr the local address the server will bind to
 */
void start_httpd(int port, int backlog, char *bindAddr) {

  RequestWrapper W;

  stopped= FALSE;

 
  if ( (myServerSocket= create_server_socket(port, backlog, bindAddr)) < 0 ) {
    
    error("http server: Could not create a server socket at port %d -- %s\n",
	  port, STRERROR);
    
    error("monit HTTP server not available\n");

    if ( Run.init ) {

      sleep(1);
      kill_daemon(SIGTERM);

    }
    
  } else {
  
    initialize_service();

    if ( Run.httpdssl ) {

      mySSLServerConnection= init_ssl_server( Run.httpsslpem,
					      Run.httpsslclientpem );

      if ( mySSLServerConnection == NULL ) {

	error("http server: Could not initilize SSL engine\n");
    
	error("monit HTTP server not available\n");

	return;
      }

#ifdef HAVE_OPENSSL
      mySSLServerConnection->server_socket= myServerSocket;
#endif
    }

    while( !stopped ) {
      
      if ( NULL == (W= socket_producer(myServerSocket)) ) {
	
	continue;

      }
      
      if ( Run.httpdssl ) {

	if ( NULL == (W->ssl= 
		      insert_accepted_ssl_socket(mySSLServerConnection) )) {
	  
	  error("http server: Cannot create SSL connection -- %s\n",
		STRERROR);
	  destroy_wrapper(W);

	  continue;

	}

	if (! embed_accepted_ssl_socket(W->ssl, W->socket)) {

	  error("http server: Cannot establish SSL connection\n");
	  destroy_wrapper(W);

	  continue;

	}
	
      }

      if ( (!W->status) || (!authenticate(W)) ) {
	
	destroy_wrapper(W);
	
	continue;
	
      }
      
      if ( !Run.httpdssl ) {
	
	if ( !check_socket(W->socket) ) {
	  
	  error("http server: Client socket not ready -- %s\n", STRERROR);
	  destroy_wrapper(W);
	  
	  continue;
	  
	}

      }

      http_processor(W);
      
    }

  }

}


/**
 * Stop the HTTPD server. 
 */
void stop_httpd() {

  stopped= TRUE;
  close_socket(myServerSocket);
  delete_ssl_server_socket(mySSLServerConnection);  

}


/* -------------------------------------------------------------- Properties */


/**
 * Add hosts allowed to connect to this server. 
 * @param name A hostname (A-Record) or IP address to be added to the hosts allow list
 * @return FALSE if the given host does not resolve, otherwise TRUE
 */
int add_host_allow(char *name) {

  struct hostent *hp;

  if ( ! (hp= gethostbyname(name)) ) {
    
    return FALSE;
    
  } else  {
    
    HostsAllow h= NEW(h);
    
    while(*hp->h_addr_list) {

      h->name= xstrdup( inet_ntoa( *(struct in_addr *) *hp->h_addr_list++) );
    
      LOCK(hostlist_mutex)
	
      if ( hostlist ) {
      
        HostsAllow p, n;
      
        for ( n= p= hostlist; p; n= p, p= p->next) {
	
	  if ( !strcasecmp(p->name, name) ) {
	  
	    destroy_host_allow(h);
	    goto done;
	  
	  }
	
        }
      
        n->next= h;
      
      } else {
      
        hostlist= h;
      
      }

      done:
      END_LOCK;

    }

  }

  return TRUE;

}


/**
 * Are any hosts present in the host allow list?
 * @return TRUE if the host allow list is non-empty, otherwise FALSE
 */
int has_hosts_allow() {

  int rv;

  LOCK(hostlist_mutex)
      rv= (hostlist != NULL );
  END_LOCK;

  return rv;

}


/** 
 * Free the host allow list
 */
void destroy_hosts_allow() {

  if ( has_hosts_allow() ) {
    
    LOCK(hostlist_mutex)
	destroy_host_allow(hostlist);
        hostlist= NULL;
    END_LOCK;
    
  }

}


/* ----------------------------------------------------------------- Private */


/**
 * Setup the cervlet service and verify that a cervlet implementation
 * exist. Only one cervlet is supported in this version. In a standalone
 * versions this function will load cervlets from a repository and
 * initialize each cervlet.
 */
static void initialize_service() {

  init_service();
  check_Impl();

}


/**
 * Abort if no Service implementors are found
 */
static void check_Impl() {

  if ( (Impl.doGet == 0) || (Impl.doPost == 0) ) {
    
    error("http server: Service Methods not implemented\n");
    _exit(1);
    
  }
  
}


/**
 * Returns TRUE if remote host is allowed to connect, if not shutdown
 * the socket and return FALSE. If allow Basic Authentication is
 * defined in the Run.Auth object, authentication is delegated to the
 * processor module.
 */
static int authenticate(RequestWrapper w) {

  if ( is_host_allow(w->inetaddr->remote_host) ) {
    
    return TRUE;
    
  } 

  if ( ! has_hosts_allow() && Run.Auth.defined ) {

    return TRUE;

  }
  
  return FALSE;

}



/**
 * Returns TRUE if host is allowed to connect to
 * this server
 */
static int is_host_allow(char *name) { 

  HostsAllow p;
  int rv= FALSE;

  LOCK(hostlist_mutex)
  
  for ( p= hostlist; p; p= p->next) {

    if ( !strncasecmp(p->name, name, STRLEN) ) {
      
      rv= TRUE;
      break;
      
    }
    
  }

  END_LOCK;

  return rv;

}


/* --------------------------------------------------------------- Factories */


/**
 * Accept connections from Clients and create a RequestWraper object
 * for each successful accept. If accept fails, return a NULL object
 * which will terminate the server.
 */
static RequestWrapper socket_producer(int server) {

  int client;
  int len= sizeof(struct sockaddr_in);
  struct sockaddr_in in;

  again:
  errno= 0;
  if ( (client= accept(server, (struct sockaddr*)&in, &len)) < 0) {
    
    if ( stopped ) {
      
      log("http server: service stopped\n");
      
    } else if ( errno == EINTR ) {
      
      goto again;
      
    } else {
      
      error("http server: cannot accept connection -- %s\n", STRERROR);
      
    }
    
    return NULL;
    
  }

  return create_wrapper(in.sin_addr, client);

}


/**
 * Create a RequestWrapper. This function will always return a 
 * RequestWraper. If the build is successful the status variable in 
 * the Wrapper is set to TRUE otherwise it is set to FALSE. If the 
 * build fails, the client socket is also shutdown. The build fails
 * if the connecting client does not have a valid domain name.
 */
static RequestWrapper create_wrapper(struct in_addr iaddr, int client) {

  RequestWrapper w= NEW(w);
  InetAddress i= NEW(i);
    
  i->remote_host= xstrdup( inet_ntoa(iaddr) );
  i->local_host= get_localhostname();
  w->inetaddr= i;
  w->socket= client;
  w->status= TRUE;
    
  return (w);

}


/* ----------------------------------------------------------------- Cleanup */


/**
 * Free a (linked list of) host_allow ojbect(s). 
 */
static void destroy_host_allow(HostsAllow p) {
  
  HostsAllow a= p; 
  
  if ( a->next ) {
    
    destroy_host_allow(a->next);
    
  }

  free(a->name);
  free(a);
 
}

