/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>
#include <time.h>
#include <setjmp.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "processor.h"
#include "base64.h"
#include "ssl.h"

#define REQUEST   1
#define RESPONSE  2
#define HEADER    3
#define PARAM     4

/* Private variables */
extern ssl_server_connection * mySSLServerConnection;

/* Private prototypes */
static void do_service(RequestWrapper);
static void send_response(HttpResponse);
static char *get_date();
static char *get_server(); 
static HttpRequest create_HttpRequest(RequestWrapper);
static HttpResponse create_HttpResponse(RequestWrapper);
static void create_headers(HttpRequest);
static int create_parameters(HttpRequest req);
static void reset_response(HttpResponse res);
static int is_authenticated(HttpRequest, HttpResponse);
static int basic_authenticate(HttpRequest);
static void done(HttpRequest, HttpResponse);
static void destroy_HttpRequest(HttpRequest);
static void destroy_HttpResponse(HttpResponse);
static void destroy_entry(void *);
static void internal_error(int, int, char *);
static void internal_error_ssl(ssl_connection *, int, char *);
static HttpParameter parse_parameters(char *);
static int is_available(int);


/**
 *  A HTTP Processor module that can handle HTTP requests received
 *  from a client, and return responses based on those requests.
 *
 *  This Processor delegates the actual handling of the request and
 *  reponse to so called cervlets, which must implement two methods;
 *  doGet and doPost. We call them cervlets because they are small
 *  C-modules closely resembling Java Servlets.
 *
 *  NOTES
 *    This Processor is command oriented and if a second slash '/' is
 *    found in the URL it's asumed to be the PATHINFO. In other words
 *    this processor perceive an URL as:
 *
 *                      /COMMAND?args/PATHINFO
 *
 *     The doGet/doPost routines act's on the COMMAND. See the
 *     cervlet.c code in this dir. for an example. 
 *
 *     2) This design is based on the Java Servlet API Framework, and
 *     to some extent; the Jakarta Catelina servlet engine by the very
 *     talented Craig R. McClanahan and others.
 *
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: processor.c,v 1.14 2003/02/11 21:27:40 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------- Entry */


/**
 * Process a HTTP request. This is done by dispatching to the service
 * function. Before the dispatch, an alarm clock is setup for handling
 * request timeout.
 * @param wrapper A Wrapper object wrapping the client socket connection
 */
void *http_processor(void *wrapper) {
  
  RequestWrapper W= wrapper;

  if(! is_available(W->socket)) {
    
    internal_error(W->socket, SC_REQUEST_TIMEOUT,
		   "Time out when handling the Request");
    goto shutdown;
    
  }
  do_service(W);

  shutdown:
  destroy_wrapper(W);

  return NULL;
  
}


/**
 * Callback for implementors of cervlet functions.
 * @param doGetFunc doGet function
 * @param doPostFunc doPost function 
 */
void add_Impl(void *doGetFunc, void *doPostFunc) {

  Impl.doGet= doGetFunc;
  Impl.doPost= doPostFunc;

}


/* ------------------------------------------------------------------ Public */


/**
 * Send an error message
 * @param res HttpResponse object
 * @param code Error Code to lookup and send
 * @param msg Optional error message (may be NULL)
 */
void send_error(HttpResponse res, int code, char *msg) {

  char *err= get_status_string(code);
  char *server= get_server();
  
  reset_response(res);
  set_content_type(res, "text/html");
  set_status(res, code, err);
  out_print(res,
	   "<html><head><title>%d %s</title></head>"\
	   "<body bgcolor=#FFFFFF><h2>%s</h2>%s<p>"\
	   "<hr><a href='%s'><font size=-1>%s</font></a>"\
	   "</body></html>\r\n",
	    code, err, err, msg?msg:"", SERVER_URL, server);
  free(err);
  free(server);
	   
}


/**
 * Sends a temporary redirect response to the client using the
 * specified redirect Location URL. The url must be absolute.
 * @param res HttpResponse object
 * @param location An absolute url to redirect to
 */
void send_redirect(HttpResponse res, char *location) {

  char *string= get_status_string(SC_MOVED_TEMPORARILY);
  reset_response(res);
  set_status(res, SC_MOVED_TEMPORARILY, string);
  set_header(res, "Location", location);
  free(string);

}


/**
 * Prints a string into the given HttpResponse output buffer. The
 * actual response to the client is done in the service function,
 * which will call the private function send_response.  Cervlets
 * should use this function (i.e. out_print) for sending a response,
 * and never use the HttpResponse outputstream directly.
 *
 * Despite the above warning, IF the HttpResponse.outputstream was
 * used directly by a cervlet THEN the cervlet MUST set the
 * is_committed flag in the HttpResponse object and is responsible for
 * sending all HTTP headers and content by itself.
 * @param res HttpResponse object
 * @param format A formated string to be sent to the client
 */
void out_print(HttpResponse res, const char *format, ...) {

  if ( format ) {
    
    int n= 0;
    ssize_t need= 0;
    ssize_t have= 0;
    int size= RES_STRLEN;
    char *string= xmalloc(size);
    va_list ap;

    /* Ripped from the Linux vsnprintf man page */
    while (1) {
      
      va_start(ap, format);
      n= vsnprintf(string, size, format, ap);
      va_end(ap);
      
      if (n > -1 && n < size) {
	
	  break;

      }
      
      if (n > -1) {
	
	  size= n+1;

      } else {
	
	  size*= 2;

      }
      
      string= xresize(string, size);
      
    }

    /* Check for possible buffersize overflow and abort if detected */
    if ( (float)((float)(sizeof(char)*(res->bufsize + RES_STRLEN + n + 1))
		 > SSIZE_MAX) ) {
      
      log("out_print: Possible http response buffer overflow detected. "
	  "aborting \n");
      exit(1);
      
    }

    have= res->bufsize - res->bufused;
    need= sizeof(char)*( n + 1); 
      
    if ( have <= need ) {
      
      /* Geometrical expand the outputbuffer size */ 
      res->outputbuffer=
	  xresize(res->outputbuffer,(res->bufsize + need + RES_STRLEN));
      
      res->bufsize+= (need + RES_STRLEN);
      
      if ( !res->bufused ) {
	
	  memset(res->outputbuffer, 0, res->bufsize);

      }
      
      res->bufused+= need;
      
    } else {
      
      res->bufused+= need;
      
    }
    
    strcat(res->outputbuffer, string);
    
    free(string);
    
  }
  
}


/**
 * Free a RequestWrapper object
 * @param w A RequestWrapper object (Should not be NULL)
 */
void destroy_wrapper(RequestWrapper w) {
  
  if ( w && w->inetaddr ) {

    free(w->inetaddr->remote_host);
    free(w->inetaddr->local_host);
    free(w->inetaddr);

  }

  if ( Run.httpdssl ) {

    close_accepted_ssl_socket(mySSLServerConnection, w->ssl);

  } else {

    close_socket(w->socket);

  }

  free(w);

}


/* -------------------------------------------------------------- Properties */


/**
 * Adds a response header with the given name and value. If the header
 * had already been set the new value overwrites the previous one.
 * @param res HttpResponse object
 * @param name Header key name
 * @param value Header key value
 */
void set_header(HttpResponse res, char *name, char *value) {

  HttpHeader h= NEW(h);
  
  h->name= xstrdup(name);
  h->value= xstrdup(value);
  
  if ( res->headers ) {
    
    HttpHeader n, p;
    
    for ( n= p= res->headers; p; n= p, p= p->next) {
      
      if ( !strcasecmp(p->name, name) ) {
	
	free(p->value);
	p->value= xstrdup(value);
	destroy_entry(h);
	return;
	
      }
      
    }
    
    n->next= h;
    
  } else {
    
    res->headers= h;
    
  }

}


/**
 * Sets the status code for the response
 * @param res HttpResponse object
 * @param code A HTTP status code <100-510>
 * @param msg The status code string message 
 */
void set_status(HttpResponse res, int code, char *msg) {
  
  res->status= code;
  free(res->status_msg);
  res->status_msg= xstrdup(msg);
  
}


/**
 * Set the response content-type
 * @param res HttpResponse object
 * @param mime Mime content type, e.g. text/html
 */
void set_content_type(HttpResponse res, char *mime) {

  set_header(res, "Content-Type", mime);

}


/**
 * Returns the value of the specified header
 * @param req HttpRequest object
 * @param name Header name to lookup the value for
 * @return The value of the specified header, NULL if not found
 */
char *get_header(HttpRequest req, const char *name) {

  HttpHeader p;

  for ( p= req->headers; p; p= p->next) {
    
    if ( !strcasecmp(p->name, name) ) {
      
	return xstrdup(p->value);

    }
    
  }

  return NULL;

}


/**
 * Returns the value of the specified parameter
 * @param req HttpRequest object
 * @param name The request parameter key to lookup the value for
 * @return The value of the specified parameter, or NULL if not found
 */
char *get_parameter(HttpRequest req, const char *name) {

  HttpParameter p;

  for ( p= req->params; p; p= p->next) {
    
    if ( !strcasecmp(p->name, name) ) {
      
	return xstrdup(p->value);

    }
    
  }

  return NULL;

}


/**
 * Returns a string containing all (extra) headers found in the
 * response.  The headers are newline separated in the returned
 * string.
 * @param res HttpResponse object
 * @return A String containing all headers set in the Response object
 */
char *get_headers(HttpResponse res) {

  HttpHeader p;
  char buf[RES_STRLEN];
  char *b= buf;

  for ( p= res->headers; p; p= p->next) {
    
    b+= sprintf(b, "%s: %s\r\n", p->name, p->value);
    
  }
  
  return buf[0]?xstrdup(buf):NULL;

}


/**
 * Lookup the corresponding HTTP status string for the given status
 * code
 * @param status A HTTP status code
 * @return A default status message for the specified HTTP status
 * code.
 */
char *get_status_string(int status) {

  switch (status) {
  case SC_OK:
      return xstrdup("OK");
  case SC_ACCEPTED:
      return xstrdup("Accepted");
  case SC_BAD_GATEWAY:
      return xstrdup("Bad Gateway");
  case SC_BAD_REQUEST:
      return xstrdup("Bad Request");
  case SC_CONFLICT:
      return xstrdup("Conflict");
  case SC_CONTINUE:
      return xstrdup("Continue");
  case SC_CREATED:
      return xstrdup("Created");
  case SC_EXPECTATION_FAILED:
      return xstrdup("Expectation Failed");
  case SC_FORBIDDEN:
      return xstrdup("Forbidden");
  case SC_GATEWAY_TIMEOUT:
      return xstrdup("Gateway Timeout");
  case SC_GONE:
      return xstrdup("Gone");
  case SC_VERSION_NOT_SUPPORTED:
      return xstrdup("HTTP Version Not Supported");
  case SC_INTERNAL_SERVER_ERROR:
      return xstrdup("Internal Server Error");
  case SC_LENGTH_REQUIRED:
      return xstrdup("Length Required");
  case SC_METHOD_NOT_ALLOWED:
      return xstrdup("Method Not Allowed");
  case SC_MOVED_PERMANENTLY:
      return xstrdup("Moved Permanently");
  case SC_MOVED_TEMPORARILY:
      return xstrdup("Moved Temporarily");
  case SC_MULTIPLE_CHOICES:
      return xstrdup("Multiple Choices");
  case SC_NO_CONTENT:
      return xstrdup("No Content");
  case SC_NON_AUTHORITATIVE:
      return xstrdup("Non-Authoritative Information");
  case SC_NOT_ACCEPTABLE:
      return xstrdup("Not Acceptable");
  case SC_NOT_FOUND:
      return xstrdup("Not Found");
  case SC_NOT_IMPLEMENTED:
      return xstrdup("Not Implemented");
  case SC_NOT_MODIFIED:
      return xstrdup("Not Modified");
  case SC_PARTIAL_CONTENT:
      return xstrdup("Partial Content");
  case SC_PAYMENT_REQUIRED:
      return xstrdup("Payment Required");
  case SC_PRECONDITION_FAILED:
      return xstrdup("Precondition Failed");
  case SC_PROXY_AUTHENTICATION_REQUIRED:
      return xstrdup("Proxy Authentication Required");
  case SC_REQUEST_ENTITY_TOO_LARGE:
      return xstrdup("Request Entity Too Large");
  case SC_REQUEST_TIMEOUT:
      return xstrdup("Request Timeout");
  case SC_REQUEST_URI_TOO_LARGE:
      return xstrdup("Request URI Too Large");
  case SC_RANGE_NOT_SATISFIABLE:
      return xstrdup("Requested Range Not Satisfiable");
  case SC_RESET_CONTENT:
      return xstrdup("Reset Content");
  case SC_SEE_OTHER:
      return xstrdup("See Other");
  case SC_SERVICE_UNAVAILABLE:
      return xstrdup("Service Unavailable");
  case SC_SWITCHING_PROTOCOLS:
      return xstrdup("Switching Protocols");
  case SC_UNAUTHORIZED:
      return xstrdup("Unauthorized");
  case SC_UNSUPPORTED_MEDIA_TYPE:
      return xstrdup("Unsupported Media Type");
  case SC_USE_PROXY:
      return xstrdup("Use Proxy");
  default: {
     char buf[STRLEN];
     snprintf(buf, STRLEN, "HTTP Response Status %d", status);
     return xstrdup(buf);
    } 
  }

}


/* ----------------------------------------------------------------- Private */


/**
 * Receives standard HTTP requests from a client socket and dispatches
 * them to the doXXX methods defined in a cervlet module.
 */
static void do_service(RequestWrapper w) {

  volatile HttpResponse res= create_HttpResponse(w);
  volatile HttpRequest req= create_HttpRequest(w);

  if ( res && req ) {

    if ( is_authenticated(req, res) ) {
      
      if ( !strcmp(req->method, METHOD_GET) ) {

	Impl.doGet(req, res);
	
      }
      else if ( !strcmp(req->method, METHOD_POST) ) {

	Impl.doPost(req, res);
	
      }
      else {
	
	send_error(res, SC_NOT_IMPLEMENTED, "Method not implemented");
	
      }

    }

    if ( Run.httpdssl ) {

      send_response(res);
      
    } else {

      if ( check_socket(w->socket) ) {
	
	send_response(res);
      }

    }
    
  }

  done(req, res);
  
}


/**
 * Return a (RFC1123) Date string
 */
static char *get_date() {

  char date[STRLEN];
  time_t now;
  
  time(&now);
  
  if ( !strftime(date, STRLEN, DATEFMT, gmtime(&now)) ) {
    
      memset(date, 0, STRLEN);

  }
  
  return xstrdup(date);
  
}


/**
 * Return this server name + version
 */
static char *get_server() {

  char server[STRLEN];
  snprintf(server, STRLEN, "%s %s", SERVER_NAME, SERVER_VERSION);
  
  return xstrdup(server);
  
}


/**
 * Send the response to the client. If the response has already been
 * commited, this function does nothing.
 */
static void send_response(HttpResponse res) {

  if ( !res->is_committed ) {
    char *date= get_date();
    char *server= get_server();
    char *headers= get_headers(res);
    int  len= res->bufused?strlen(res->outputbuffer):0;

    if ( Run.httpdssl ) {

      printf_ssl_socket(res->ssl, "%s %d %s\r\n", res->protocol, res->status, 
			res->status_msg);
      printf_ssl_socket(res->ssl, "Date: %s\r\n", date);
      printf_ssl_socket(res->ssl, "Server: %s\r\n", server);
      printf_ssl_socket(res->ssl, "Content-length: %d\r\n", len);
      printf_ssl_socket(res->ssl, "Connection: close\r\n");
      if ( headers ) printf_ssl_socket(res->ssl, "%s", headers);
      printf_ssl_socket(res->ssl, "\r\n\r\n");
      
      if ( res->bufused ) send_ssl_socket(res->ssl, res->outputbuffer, len);

    } else {
 
      FILE *out= res->outputstream;
    
      fprintf(out, "%s %d %s\r\n", res->protocol, res->status, 
	      res->status_msg);
      fprintf(out, "Date: %s\r\n", date);
      fprintf(out, "Server: %s\r\n", server);
      fprintf(out, "Content-length: %d\r\n", len);
      fprintf(out, "Connection: close\r\n");
      if ( headers ) fprintf(out, "%s", headers);
      fprintf(out, "\r\n\r\n");
      
      if ( res->bufused ) fprintf(out, "%s", res->outputbuffer);
      
    }
    
    free(headers);
    free(date);
    free(server);
    
  }

}


/* --------------------------------------------------------------- Factories */


/**
 * Returns a new HttpRequest object wrapping the client request
 */
static HttpRequest create_HttpRequest(RequestWrapper W) {

  char line[REQ_STRLEN];
  char method[STRLEN];
  char url[REQ_STRLEN];
  char protocol[STRLEN]; 
  HttpRequest req;
  FILE *in= NULL;

  if ( Run.httpdssl ) {

    if ( gets_ssl_socket(W->ssl, line, sizeof(line)) == NULL ) {
      
      internal_error_ssl(W->ssl, SC_BAD_REQUEST, "No request found");
      return NULL;
      
    }

    if ( sscanf(line, "%s %s HTTP/%3[1.0]", method, url, protocol) != 3 ) {
      
      internal_error_ssl(W->ssl, SC_BAD_REQUEST, "Cannot parse request");
      return NULL;
      
    }

  } else {

    in= fdopen(dup(W->socket), "r");
      
    if ( in == NULL ) {
      
      internal_error(W->socket, SC_INTERNAL_SERVER_ERROR,
		     "Cannot create inputstream");
      return NULL;
      
    }
    
    if ( fgets(line, sizeof(line), in) == NULL ) {
      
      internal_error(W->socket, SC_BAD_REQUEST, "No request found");
      fclose(in);
      return NULL;
      
    }
    
    if ( sscanf(line, "%s %s HTTP/%3[1.0]", method, url, protocol) != 3 ) {
      
      internal_error(W->socket, SC_BAD_REQUEST, "Cannot parse request");
      fclose(in);
      return NULL;
      
    }
  }

  req= NEW(req);
  unescape_url(url);
  chomp(protocol);
  req->method= xstrdup(method);
  req->url= xstrdup(url);
  req->protocol= xstrdup(protocol); 
  req->ssl= W->ssl;

  if (! Run.httpdssl ) {
 
   req->inputstream= in;

  } else {

   req->inputstream= NULL;

  }

  req->inetaddr= W->inetaddr;

  create_headers(req);

  if ( !create_parameters(req) ) {
    
    if ( Run.httpdssl ) {

      internal_error_ssl(W->ssl, SC_BAD_REQUEST,
		     "Cannot parse Request parameters");

    } else {

      internal_error(W->socket, SC_BAD_REQUEST,
		     "Cannot parse Request parameters");
      fclose(in);

    }

    return NULL;
    
  }
    
  return req;
  
}


/**
 * Returns a new HttpResponse object wrapping a default response. Use
 * set_XXX methods to change the object.
 */
static HttpResponse create_HttpResponse(RequestWrapper W) {
  
  HttpResponse res;
  FILE *out= NULL;

  if ( ! Run.httpdssl ) {

    out = fdopen(dup(W->socket), "w");

    if ( out == NULL ) {
    
      internal_error(W->socket, SC_INTERNAL_SERVER_ERROR,
		     "Cannot create outputstream");
      
      return NULL;
    }

  }
  
  res= NEW(res);
  res->protocol= xstrdup(SERVER_PROTOCOL);
  res->status= 200;
  res->status_msg= get_status_string(SC_OK);
  res->outputstream= out;
  res->outputbuffer= NULL;
  res->bufsize= 0;
  res->bufused= 0;
  res->is_committed= FALSE;
  res->ssl= W->ssl;
  
  return res;
  
}


/**
 * Create HTTP headers for the given request
 */
static void create_headers(HttpRequest req) {

  char line[REQ_STRLEN];
  char *value;
  HttpHeader p;
  HttpHeader header;

  while ( 1 ) {

    if ( Run.httpdssl ) {

      if (NULL == gets_ssl_socket(req->ssl, line, sizeof(line))) {

	break;

      }

    } else {

      if (NULL == fgets(line, sizeof(line), req->inputstream)) {

	break;

      }

    }

    if ( !strcmp(line, "\r\n") || !strcmp(line, "\n") )
	break;
    
    if( NULL != (value= strchr(line, ':')) ) {
      
      header= NEW(header);
      *value++= '\0';
      trim(line);
      trim(value);
      chomp(value);
      header->name= xstrdup(line);
      header->value= xstrdup(value);
      
      if ( req->headers ) {
	
	for ( p= req->headers; p->next; p= p->next)
	    /* Empty */;
	p->next= header;
	
      } else {
	
	req->headers= header;
	
      }
      
    }
    
  }
  
}


/**
 * Create parameters for the given request. Returns FALSE if an error
 * occurs.
 */
static int create_parameters(HttpRequest req) {

  char *query_string= NULL;
  int alloc= FALSE;

  if ( !strcmp(req->method, METHOD_POST) ) {
    
    int len; char *l= get_header(req, "Content-Length");
    
    if ( l && (len= atoi(l)) ) {
      
      query_string= xmalloc(sizeof(char) * (len + 1));
      
      if ( fgets(query_string, (sizeof(char)*(len+1)),
		 req->inputstream) == NULL ) {
	
	free(query_string);
	return FALSE;
	
      }
      
      alloc= TRUE;
      
    }
    
  } else if ( !strcmp(req->method, METHOD_GET) ) {
    
    char *p;
    
    if( NULL != (p= strchr(req->url, '?')) ) {
      
      *p++= '\0';
      query_string= p;
      
    }
    
  }
  
  if ( query_string ) {
    
    char *p;
    
    if( NULL != (p= strchr(query_string, '/')) ) {
      
      *p++= '\0';
      req->pathinfo= p;
      
    }
    
    req->params= parse_parameters(query_string);
    
  }

  if ( alloc ) free(query_string);
  
  return TRUE;
  
}



/* ----------------------------------------------------------------- Cleanup */


/**
 * Clear the response output buffer and headers
 */
static void reset_response(HttpResponse res) {

  if ( res->headers ) destroy_entry(res->headers);
  memset(res->outputbuffer, 0, res->bufsize);
  res->headers= NULL; /* Release Pragma */
  
}


/**
 * Finalize the request and response object.
 */
static void done(HttpRequest req, HttpResponse res) {

  destroy_HttpRequest(req);
  destroy_HttpResponse(res);

}


/**
 * Free a HttpRequest object
 */
static void destroy_HttpRequest(HttpRequest req) {

  if ( req ) {
    
    if ( req->inputstream ) fclose(req->inputstream);
    free(req->method);
    free(req->url); /* req->pathinfo is freed with url */
    free(req->protocol);
    if ( req->headers ) destroy_entry(req->headers);
    if ( req->params ) destroy_entry(req->params);
    free(req);
    
  }
  
}


/**
 * Free a HttpResponse object
 */
static void destroy_HttpResponse(HttpResponse res) {

  if ( res ) {
    
    if ( res->outputstream ) fclose(res->outputstream);
    free(res->protocol);
    free(res->status_msg);
    free(res->outputbuffer);
    if ( res->headers) destroy_entry(res->headers);
    free(res);
    
  }
  
}


/**
 * Free a (linked list of) http entry ojbect(s). Both HttpHeader and
 * HttpParameter are of this type.
 */
static void destroy_entry(void *p) {
  
  struct entry *h= p; 
  
  if ( h->next ) {
    
    destroy_entry(h->next);
    
  }

  free(h->name);
  free(h->value);
  free(h);
 
}


/* ----------------------------------------------------- Checkers/Validators */


/**
 * Do Basic Authentication if this auth. style is allowed. 
 */
static int is_authenticated(HttpRequest req, HttpResponse res) {

  if ( Run.Auth.defined ) {

    if ( ! basic_authenticate(req) ) {
      
      send_error(res, SC_UNAUTHORIZED,
		 "You are <b>not</b> authorized to access <i>monit</i>. "
		 "Either you supplied the wrong credentials (e.g. bad "
		 "password), or your browser doesn't understand how to supply "
		 "the credentials required");
      set_header(res, "WWW-Authenticate", "Basic realm=\"monit\"");
      
      return FALSE;
      
    }

  }

  return TRUE;

}


/**
 * Authenticate the basic-credentials (uname/password) submitted by
 * the user.
 */
static int basic_authenticate(HttpRequest req) {

  int rv= FALSE;
  char *credentials= get_header(req, "Authorization");

  if ( credentials ) {

    unsigned char *cr= xcalloc(sizeof(unsigned char), strlen(credentials)+1);

    if ( decode_base64(cr, strchr(credentials, ' ')) ) {

      char buf[STRLEN];

      snprintf(buf, STRLEN, "%s:%s", Run.Auth.uname, Run.Auth.passwd);
      
      if ( !strncmp(cr, buf, strlen(buf)) ) {
	  
	  rv= TRUE;
	  
      }
	
    }
	
    free(cr);
    free(credentials);

  }
  
  return rv;
  
}


/* --------------------------------------------------------------- Utilities */


/**
 * Send an error message to the client. This is a helper function,
 * used internal if the service function fails to setup the framework
 * properly; i.e. with a valid HttpRequest and a valid HttpResponse.
 */
static void internal_error(int client, int status, char *msg) {

  char response[RES_STRLEN];
  char *server= get_server();
  char *date= get_date();
  char *status_msg= get_status_string(status);

  snprintf(response, RES_STRLEN,
	   "%s %d %s\r\n"
	   "Date: %s\r\n"
	   "Server: %s\r\n"
	   "Content-Type: text/html\r\n"
	   "Connection: close\r\n"
	   "\r\n"
	   "<html><head><title>%s</title></head>"
	   "<body bgcolor=#FFFFFF><h2>%s</h2>%s<p>"
	   "<hr><a href='%s'><font size=-1>%s</font></a>"
	   "</body></html>\r\n",
	   SERVER_PROTOCOL, status, status_msg, date, server,
	   status_msg, status_msg, msg, SERVER_URL, server);
  sock_send(client, response, strlen(response), 0);
  free(server);
  free(date);
  free(status_msg);
  
}

static void internal_error_ssl(ssl_connection *ssl, int status, char *msg) {

  char response[RES_STRLEN];
  char *server= get_server();
  char *date= get_date();
  char *status_msg= get_status_string(status);

  snprintf(response, RES_STRLEN,
	   "%s %d %s\r\n"
	   "Date: %s\r\n"
	   "Server: %s\r\n"
	   "Content-Type: text/html\r\n"
	   "Connection: close\r\n"
	   "\r\n"
	   "<html><head><title>%s</title></head>"
	   "<body bgcolor=#FFFFFF><h2>%s</h2>%s<p>"
	   "<hr><a href='%s'><font size=-1>%s</font></a>"
	   "</body></html>\r\n",
	   SERVER_PROTOCOL, status, status_msg, date, server,
	   status_msg, status_msg, msg, SERVER_URL, server);

  send_ssl_socket(ssl, response, strlen(response));
  free(server);
  free(date);
  free(status_msg);
  
}


/**
 * Parse request parameters from the given query string and return a
 * linked list of HttpParameters
 */
static HttpParameter parse_parameters(char *query_string) {

  HttpParameter head= NULL;
  int controller= 1000;
  
  while ( query_string[0] && controller-- ) {
    
    HttpParameter p= NEW(p);
    p->value= makeword(query_string,'&');
    p->name= makeword(p->value,'=');
    plustospace(p->value);
    unescape_url(p->value);
    plustospace(p->name);
    unescape_url(p->name);
    p->next= head;
    head= p;
    
  }

  return head;
  
}


/**
 * Check if data is available, if not, wait timeout seconds for data
 * to be present.
 * @param s A client socket
 * @return TRUE if data is available otherwise FALSE
 */
static int is_available(int s) {

  fd_set rset;
  struct timeval tv;

  if(s < 0)
      return FALSE;
  
  FD_ZERO(&rset);
  FD_SET(s, &rset);
  tv.tv_sec= REQUEST_TIMEOUT;
  tv.tv_usec= 0;

  return (select(s+1, &rset, NULL, NULL, &tv)>0);

}
 
