/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


%{
  
/*
 * DESCRIPTION
 *   Lexical grammar for tokenizing the control file. 
 *
 *  AUTHOR
 *   Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *   Olivier Beyssac, <ob@r14.freenix.org>
 *   Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  CVS INFO
 *     $Id: l.l,v 1.29 2003/02/11 21:27:39 martinp Exp $
 */

#include <config.h>
  
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#include "monitor.h"
#include "y.tab.h"

  int lineno= 1;
  
  /* Prototypes */
  void yyerror(const char*);
  static char *handle_quoted_path(char *);
  static char *handle_quoted_string(char *);

%}

ws          [ \t]+
wws         [ \t=;,()]+
str         [^\000-\041@:{}"';(),]+
noslash     [^\000-\041@/:{}"';(),]+
hostname    {str}(\.{str})*
greater     ("greater"|"gt"|">")
less        ("less"|"lt"|"<")
equal       ("equal"|"eq"|"==")
notequal    ("notequal"|"ne"|"!=")
loadavg1    load(avg)[ ]*(\([ ]*1[ ]*(m|min)?[ ]*\))?
loadavg5    load(avg)[ ]*\([ ]*5[ ]*(m|min)?[ ]*\)
loadavg15   load(avg)[ ]*\([ ]*15[ ]*(m|min)?[ ]*\)
startarg    start{wws}(program)?{wws}["]  
stoparg     stop{wws}(program)?{wws}["]   

%x ARGUMENT 

%%

{wws}             { /* Wide white space */ }
(#.*)?\\?\n       { lineno++; } 

i(f|s)            {/* EMPTY */}
are               {/* EMPTY */}
for               {/* EMPTY */}
on(ly)?           {/* EMPTY */}
with(in)?         {/* EMPTY */}
program(s)?       {/* EMPTY */}
and               {/* EMPTY */}
has               {/* EMPTY */}
using             {/* EMPTY */}
use               {/* EMPTY */}
the               {/* EMPTY */}
cycle(s)?         {/* EMPTY */}
sum               {/* EMPTY */}
restarts          {/* EMPTY */}
th(a|e)n          {/* EMPTY */}

{startarg}        { BEGIN(ARGUMENT); return START; }
{stoparg}         { BEGIN(ARGUMENT); return STOP; }

ssl               { return HTTPDSSL; }
enable            { return ENABLE; }
disable           { return DISABLE; }
set               { return SET; }
daemon            { return DAEMON; }
logfile           { return LOGFILE; }
syslog            { return SYSLOG; }
httpd             { return HTTPD; }
address           { return ADDRESS; }
clientpemfile     { return CLIENTPEMFILE; }
allowselfcertification  { return ALLOWSELFCERTIFICATION; }
certmd5           { return CERTMD5; }
pemfile           { return PEMFILE; }
init              { return INIT; }
allow             { return ALLOW; }
check             { return CHECK; }
pidfile           { return PIDFILE; }
start             { return START; }
stop              { return STOP; }
port(number)?     { return PORT; }
unix(socket)?     { return UNIXSOCKET; }
type              { return TYPE; }
proto(col)?       { return PROTOCOL; }
tcp               { return TCP; }
tcpssl            { return TCPSSL; }
udp               { return UDP; }
alert             { return ALERT; }
mail-format       { return MAILFORMAT; }
restart           { return RESTART; }
timeout           { return TIMEOUT; }
checksum          { return CHECKSUM; }
expect            { return EXPECT; }
mailserver        { return MAILSERVER; }
every             { return EVERY; }
host              { return HOST; }
default           { return DEFAULT; }
http              { return HTTP; }
ftp               { return FTP; }
smtp              { return SMTP; }
pop               { return POP; }
imap              { return IMAP; }
nntp              { return NNTP; }
ssh               { return SSH; }
dwp               { return DWP; }
ldap2             { return LDAP2; }
ldap3             { return LDAP3; }
autostart         { return AUTOSTART; }
mode              { return MODE; }
active            { return ACTIVE; }
passive           { return PASSIVE; }
manual            { return MANUAL; }
group             { return GROUP; }
request           { return REQUEST; }
depend(s)?        { return DEPENDS; }
cpuusage          { return CPUUSAGE; }
memusage          { return MEMUSAGE; }
memkbyte          { return MEMKBYTE; }
timestamp	  { return TIMESTAMP; }
second(s)?	  { return SECOND; }
minute(s)?	  { return MINUTE; }
hour(s)?  	  { return HOUR; }
day(s)?	  	  { return DAY; }
sslv2             { return SSLV2; }
sslv3             { return SSLV3; }
tlsv1             { return TLSV1; }
sslauto           { return SSLAUTO; }

{loadavg1}        { return LOADAVG1; }
{loadavg5}        { return LOADAVG5; }
{loadavg15}       { return LOADAVG15; }

{greater}         { return GREATER; }
{less}            { return LESS; }
{equal}           { return EQUAL; }
{notequal}        { return NOTEQUAL; }

[0-9]+            {
                    yylval.number= atoi(yytext); return NUMBER;
                  }

[0-9]+([.][0-9]+) {
                    yylval.real= atof(yytext); return REAL;
                  }

[a-zA-Z0-9]{str}  {
		    yylval.string= xstrdup(yytext);
		    return STRING;
                  }

\"[^"\n]*["\n]    {
		    yylval.string= handle_quoted_path(yytext);
		    return PATH;
                  }

\'[^'\n]*['\n]    {
		    yylval.string= handle_quoted_path(yytext);
		    return PATH;
                  }

{str}[@]{str}     {
		    yylval.string= xstrdup(yytext);
		    return MAILADDR;
                  }

[/]{str}          {
                     yylval.string= xstrdup(yytext);
	             return PATH;
                  }

{noslash}/[/]     {
  		      yyerror("A path must be fully qualified");
                  }

"from:"[ \t\n]*{str}[@]{str} {
		      char *p= yytext+strlen("from:");
                      yylval.string = trim(xstrdup(p));
                      return MAILFROM;
                  }
                     
"subject:"[^}\n]* {
		      char *p= yytext+strlen("subject:");
                      yylval.string = trim(xstrdup(p));
                      return MAILSUBJECT;
                  }

"message:"[^}]*   {
		      char *p= yytext+strlen("message:");
                      yylval.string = trim(xstrdup(p));
                      return MAILBODY;
                  }

{hostname}        {
                      yylval.string = xstrdup(yytext);
                      return STRING;
                  }

<ARGUMENT>{

  {ws}            ;

  [\n]            {
                    lineno++;
                  }

  \"              {
                      BEGIN(INITIAL);
                  }

  \'[^']*\'       {
                      yylval.string= handle_quoted_string(yytext);
	              return STRING;
		  }
  
  \'              {
                      yyerror("Unbalanced quotes");
                  }

  \{[^}]*\}       {  
                      yylval.string= xstrdup(yytext);
	              return STRING;
		  }

  [^ \t\n"]+      {
                      yylval.string= xstrdup(yytext);
	              return STRING;
		  }

}

<INITIAL,ARGUMENT>. {
                      return yytext[0];
                  }  


%%

static char *handle_quoted_string(char *string) {

  char *buf= xstrdup(string);

  trim_quotes(buf);
  handle_string_escapes(buf);

  return buf;

}

static char *handle_quoted_path(char *string) {

  char *buf= xstrdup(string);
  
  trim(buf);
  trim_quotes(buf);
  
  if((*buf != '/'))
     yyerror("A path must be fully qualified");
  
  return buf;

}
