/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "monitor.h"


/* Private variables */
static FILE *LOG= NULL;
static char time_str[STRLEN];

/* Private prototypes */
static int  open_log();
static char *timestr(void);


/**
 *  Implementation of a logger that appends log messages to a file
 *  with a preceding timestamp. Methods support both syslog or own
 *  logfile.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *
 *  @version \$Id: log.c,v 1.8 2003/02/11 21:27:39 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Initialize the log system and 'log' function
 * @return TRUE if the log system was successfully initialized
 */
int log_init() {

  if(!Run.dolog) {
    
    return TRUE;
    
  }

  if(!open_log()) {
    
    return FALSE;
    
  }

  /* Register log_close to be
     called at program termination */
  atexit(log_close);

  return TRUE;
  
}


/**
 * Log a message to monits logfile or syslog
 * @param format A formated (printf-style) string to log
 */
void log(const char *format, ...) {
  
  char msg[STRLEN];
  va_list ap;

  ASSERT(format);
  
  va_start(ap,format);
  vsnprintf(msg, STRLEN, format, ap);
  va_end(ap);

  if(! Run.dolog) goto nolog;

  if(Run.use_syslog) {
    
    syslog(LOG_ERR|LOG_USER, "%s", msg);
    
  } else if(LOG) {
    
    fprintf(LOG,"[%s] %s", timestr(), msg);
    fflush(LOG);
    
  }

  nolog:
  if(Run.have_tty) {
    
    fprintf(stdout, "%s", msg);
    fflush(stdout);
    
  }
  
}


/**
 * Close the log file or syslog
 */
void log_close() {
  
  if(Run.use_syslog) {

    closelog();

  }
  
  if(LOG  && (0 != fclose(LOG))) {
    
    error("%s: Error closing the log file -- %s\n",
	  prog, STRERROR);
    
  }

  LOG= NULL;
  
}


/**
 * @return the log filedescriptor or -1 if it's not open, i.e. syslog
 * is used or logging is not activated
 */
int get_log_fd() {

  return (LOG!=NULL?fileno(LOG):-1);

}


/* ----------------------------------------------------------------- Private */


/**
 * Open a log file or syslog
 */
static int open_log() {
 
  if(Run.use_syslog) {
    
    openlog(prog, LOG_PID, LOG_USER);
    
  } else {
    
    umask(LOGMASK);
    
    if((LOG= fopen(Run.logfile,"a+")) == (FILE *)NULL) {
      
      error("%s: Error opening the log file '%s' for writing -- %s\n",
	    prog, Run.logfile, STRERROR);
      
      return(FALSE);
      
    }
    
  }

  return TRUE;
  
}


/**
 * Returns the current time as a formated string, see the TIMEFORMAT
 * macro in the header file
 */
static char *timestr() {

  time_t now;
  
  time(&now);
  if(!strftime(time_str, STRLEN, TIMEFORMAT, localtime(&now))) {
    
    memset(time_str, 0, STRLEN);

  }
  
  return time_str;
  
}


  

