/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif


#include "monitor.h"
#include "net.h"
#include "ssl.h"
#include "monit_process.h"

/**
 *  DESCRIPTION
 *    monit - system for monitoring programs
 *
 *  SYNOPSIS
 *    monit [options] {arguments}
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: monitor.c,v 1.47 2003/02/17 15:34:19 hauk Exp $
 *
 *  @file
 */


/* --- Private Prototypes -------------------------------------------------- */
static void  do_init();                       /* Initialize this application */
static RETSIGTYPE do_reload(int);       /* Signalhandler for a daemon reload */
static void  do_reinit();           /* Re-initialize the runtime application */
static void  do_action(char **);         /* Dispatch to the submitted action */
static RETSIGTYPE  do_destroy(int);             /* Finalize this application */
static void  do_default();                              /* Do default action */
static RETSIGTYPE  do_wakeup(int); /* Signalhandler for a daemon wakeup call */
static int   do_wakeupcall();              /* Wakeup a sleeping monit daemon */
static void  handle_options(int, char **);         /* Handle program options */
static void  help();                 /* Print program help message to stdout */
static void  version();                         /* Print version information */
static void  stop_http();                                 /* Stop monit http */
static void  start_http();                               /* Start monit http */


/**
 * The Prime mover
 */
int main(int argc, char **argv) {

  prog= stripfilename(argv[0]);
  init_env();
  handle_options(argc, argv);
 
  do_init();
  do_action(argv); 
  do_destroy(SIGSTOP);

  exit(0);
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Initialize this application - Register signal handlers,
 * Parse the control file and initialize the program's
 * datastructures and the log system.
 */
static void do_init() {

  int status;

  /*
   * Register interest for the SIGTERM signal,
   * in case we run in daemon mode this signal
   * will terminate a running daemon.
   */
  signal(SIGTERM, do_destroy);

  /*
   * Register interest for the SIGUSER1 signal,
   * in case we run in daemon mode this signal
   * will wakeup a sleeping daemon.
   */
  signal(SIGUSR1, do_wakeup);

  /*
   * Register interest for the SIGHUP signal,
   * in case we run in daemon mode this signal
   * will reload the configuration.
   */
  signal(SIGHUP, do_reload);

  /*
   * Register no interest for the SIGPIPE signal,
   */
  signal(SIGPIPE, SIG_IGN);

  /*
   * Initialize the Runtime mutex. This mutex
   * is used to synchronize handling of global
   * process data
   */
  status= pthread_mutex_init(&Run.mutex, NULL);
  if(status != 0) {
    
    log("%s: Cannot initialize mutex -- %s\n", prog, strerror(status));
    exit(1);
    
  }

  /* 
   * Get the position of the control file 
   */
  if(! Run.controlfile) {
    
    Run.controlfile= find_rcfile();
    
  }
  
  /*
   * Initialize the process information gathering interface
   */
  Run.doprocess= init_process_info();

  /*
   * Start the Parser and create the process list. This will also set
   * any Runtime constants defined in the controlfile.
   */
  if(! parse(Run.controlfile)) {
    
    exit(1);
    
  }

  /*
   * Stop and report success if we are just validating the Control
   * file syntax. The previous parse statement exits the program with
   * an error message if a syntax error is present in the control
   * file.
   */
  if(Run.testing) {

    error("Control file syntax OK\n");
    exit(0);

  }

  /*
   * Initialize the log system 
   */
  if(! log_init()) {
    
    exit(1);
    
  }

  /* 
   * Did we find any process ?  
   */
  if(! processlist) {
    
    log("%s: No programs have been specified\n", prog);
    exit(0);
    
  }
  
  /* 
   * Initialize Runtime file variables 
   */
  init_files();

  /* 
   * Should we print debug information ? 
   */
  if(Run.debug && Run.have_tty) {
    
    printrunlist();
    printprocesslist();
    
  }

}


/**
 * Re-Initialize the application - called if a
 * monit daemon receives the SIGHUP signal.
 */
static void do_reinit() {

  char *bind_addr;
  int port= Run.httpdport;

  Run.doreload= FALSE;
  
  bind_addr= Run.bind_addr?xstrdup(Run.bind_addr):NULL;

  log("Awakened by the SIGHUP signal\n");
  log("Reinitializing %s - Control file '%s'\n",
      prog, Run.controlfile);
  
  /* Run the garbage collector */
  gc();

  finalize_files();

  if(! parse(Run.controlfile)) {
    
    log("%s daemon died\n", prog);
    exit(1);
    
  }

  /* Close the current log */
  log_close();

  /* Reinstall the log system */
  if(! log_init()) {
    
    exit(1);
    
  }

  /* Did we find any process ?  */
  if(! processlist) {
    
    log("%s: No programs have been specified\n", prog);
    exit(0);
    
  }

  /* Reinitialize Runtime file variables */
  init_files();

  if(! create_pidfile(Run.pidfile)) {
      
    log("%s daemon died\n", prog);
    exit(1);
      
  }

  if(! can_http()) {
	  
    stop_http();
	  
  } else if(!is(bind_addr, Run.bind_addr) || port != Run.httpdport) {
	  
    stop_http();
    start_http();
	  
  } else if(! check_httpd()) {

    start_http();
	  
  }

  free(bind_addr);

}


/**
 * Dispatch to the submitted action - actions are program arguments
 */
static void do_action(char **args) {
  
  char *action= args[optind];
  char *P= args[++optind];
  
  if(! action) {
    
    do_default();
    
  } else if(is(action, "start")) {
    
    if(P) {
      if(exist_daemon()) {
	d_check_process(P, "start");
      } else {
	check_process(P, "start");
      }
    } else {
      
      if(Run.mygroup)
	control_group(Run.mygroup, "start");
      else
	control("start");
    }
  
  } else if(is(action, "stop")) {
    
    if(P) {
      if(exist_daemon()) {
	d_check_process(P, "stop");
      } else {
	check_process(P, "stop");
      }
    } else {
      
      if(Run.mygroup)
	control_group(Run.mygroup, "stop");
      else
	control("stop");
    }
    
  } else if(is(action, "reload")) {
    
    error("Reinitializing monit daemon\n", prog);
    kill_daemon(SIGHUP);
    
  } else if(is(action, "restart")) {
    
    if(P) {
      
      if(! exist_process(P)) {
	
	error("%s: Cannot restart program '%s' -- not found in %s\n",
	      prog, P, Run.controlfile);
	
	return;
	
      }
      
      if(exist_daemon()) {
	d_check_process(P, "restart");
      } else {
	check_process(P, "restart");
      }
      
    } else {
      
      if(Run.mygroup)
	  control_group(Run.mygroup, "restart");
      else
	  control("restart");
    }
    
  } else if(is(action, "status")) {
    
    if(Run.mygroup) {
      status_group(Run.mygroup);
    } else {
      status();
    }
    
  } else if(is(action, "quit")) {
    
    error("Stopping monit daemon\n", prog);
    kill_daemon(SIGTERM);
    
  } else if(is(action, "validate")) {
    
    validate();
    
  } else {
    
    error("%s: invalid argument -- %s  (-h will show valid arguments)\n",
	  prog, action);
    exit(1);
    
  }
  
  reset_depend();
  
}


/**
 * Signalhandler for a daemon wakeup call
 */
static RETSIGTYPE do_wakeup(int sig) {

  signal(SIGUSR1, SIG_IGN);
  log("Awakened by User defined signal 1\n");
  signal(SIGUSR1, do_wakeup);

}


/**
 * Signalhandler for a daemon reload call
 */
static RETSIGTYPE do_reload(int sig) {

  signal(SIGHUP, SIG_IGN);
  Run.doreload= TRUE;
  signal(SIGHUP, do_reload);
  
}


/**
 * Wakeup a sleeping monit daemon.
 * Returns TRUE on success otherwise FALSE
 */
static int do_wakeupcall() {

  pid_t pid;
  
  if((pid= exist_daemon()) > 0) {
    
    kill(pid, SIGUSR1);
    error("%s daemon at %d awakened\n", prog, pid);
    
    return TRUE;
    
  }
  
  return FALSE;
  
}


/**
 * Finalize this application 
 */
static RETSIGTYPE do_destroy(int sig) {
  
  switch(sig) {
    
  case SIGTERM:
      
      stop_http();
      
      if(Run.isdaemon) {
	
	log("%s daemon with pid [%d] killed\n", prog, (int)getpid());
	
      }
      
      finalize_files();
      /* Fall trough */
      
  default:
      
      gc();
      log_close();
      exit(0);
      
  }
 
}


/**
 * Default action - become a daemon if defined in the Run object and
 * run validate() between sleeps. If not, just run validate() once.
 * Also, if specified, start the monit http server if in deamon mode.
 */
static void do_default() {

  if(Run.isdaemon) {
    
    if(do_wakeupcall()) {
      
      exit(0);
      
    }
  
    log("Starting %s daemon\n", prog);
    
    if(can_http()) {
      
      log("Starting httpd at [%s:%d]\n",
	  Run.bind_addr?Run.bind_addr:"*", Run.httpdport);
      
    }
    
    if(Run.init != TRUE)
	daemonize(); 
    
    if(! create_pidfile(Run.pidfile)) {
      
      log("%s daemon died\n", prog);
      exit(1);
      
    }

    if(can_http()) {

      start_http();
      
    }
    
    for(;;) {

      if(Run.validate != TRUE) {
	validate();
      } else {
	Run.validate= FALSE;
      }

      sleep(Run.polltime);

      if(Run.doreload)
	  do_reinit();

    }
    
  }
  else {
    
    validate();
    
  }

}


/**
 * Handle program options - Options set from the commandline
 * takes precedence over those found in the control file
 */
static void handle_options(int argc, char **argv) {
  
  int opt;

  opterr= 0;

  Run.mygroup=0;

  while((opt= getopt(argc,argv,"c:d:g:l:p:iItvVh")) != -1) {

    switch(opt) {

    case 'c':
        Run.controlfile= xstrdup(optarg);
        break;
	
    case 'd':
	Run.isdaemon= TRUE;
 	sscanf(optarg, "%d", &Run.polltime);
	if(Run.polltime<1) {
	  error("%s: option -%c requires a natural number\n", prog, opt);
	  exit(1);
	}
	break;

    case 'g':
        Run.mygroup= xstrdup(optarg);
        break;
	
    case 'l':
        Run.logfile= xstrdup(optarg);
	if(is(Run.logfile, "syslog"))
	    Run.use_syslog= TRUE;
	Run.dolog= TRUE;
        break;
   
    case 'p':
        Run.pidfile= xstrdup(optarg);
        break;

    case 'i':
        Run.validate= TRUE;
        break;

    case 'I':
	Run.init= TRUE;
	break;
      
    case 't':
        Run.testing= TRUE;
        break;
	
    case 'v':
        Run.debug= TRUE;
        break;
	
    case 'V':
        version();
        exit(0);
	break;
	
    case 'h':
        help();
        exit(0);
	break;
	
    case '?':
	switch(optopt) {
	  
	case 'c':
	case 'd':
	case 'g':
	case 'l':
	case 'p':
	    error("%s: option -- %c requires an argument\n", prog, optopt);
	    break;
	default:
	    error("%s: invalid option -- %c  (-h will show valid options)\n",
		  prog, optopt);
	    
	}
	
	exit(1);
	
    }
    
  }
  
}


/**
 * Print the program's help message
 */
static void help() {
  
  printf("Usage: %s [options] {arguments}\n", prog);
  printf("Options are as follows:\n");
  printf(" -c file     Use this control file\n");
  printf(" -d n        Run as a daemon once per n seconds\n");
  printf(" -g name     Set group name for start, stop, restart and status\n");
  printf(" -l logfile  Print log information to this file\n");
  printf(" -p pidfile  Use this lock file in daemon mode\n");
  printf(" -i          Validate mode, startup in validate mode\n");
  printf(" -I          Init mode, run from init\n");
  printf(" -t          Run syntax check for the control file\n");
  printf(" -v          Verbose mode, work noisy (diagnostic output)\n");
  printf(" -V          Print version number and patchlevel\n");
  printf(" -h          Print this text\n");
  printf("Optional action arguments for non-daemon mode are as follows:\n");
  printf(" start        - Start all programs listed in the control file\n");
  printf(" start name   - Only start the named program in the control file\n");
  printf(" stop         - Stop all programs listed in the control file\n");
  printf(" stop name    - Only stop the named program in the control file\n");
  printf(" reload       - Reinitialize monit\n");
  printf(" restart      - Stop and start all programs\n");
  printf(" restart name - Only restart the named program in the control file\n");
  printf(" status       - Print status information for each program\n");
  printf(" quit         - Kill monit daemon process\n");
  printf(" validate     - Check all programs and start if not running.\n");

}


/**
 * Print version information
 */
static void version() {

  printf("This is monit version %s\n", VERSION);
  printf("Copyright (C) 2000-2003 by Contributors to the monit codebase\n");
 
}


/**
 * Stop monit http
 */
static void stop_http() {

  monit_http(STOP_HTTP);

}


/**
 * Start monit http
 */
static void start_http() {

  monit_http(START_HTTP);

}


