/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#ifndef MONITOR_H
#define MONITOR_H

#include <config.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include <signal.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>

#include "ssl.h"

#define VERSION            PACKAGE_VERSION
#define MONITRC            "monitrc"
#define TIMEFORMAT         "%Z %b %e %T"
#define STRERROR            strerror(errno)
#define STRLEN             256
#define ARGMAX             64
/* Set log file mode: "-rw-rw-r--" */
#define LOGMASK            0112
/* Set pid file mode: "-rw-r--r--" */
#define PIDMASK            0122
#define MYPIDDIR           PIDDIR
#define MYPIDFILE          "monit.pid"

#define LOCALHOST          "localhost"
#define SMTP_PORT          25

#define SMTP_TIMEOUT       30
#define CHECK_TIMEOUT      30
#define SSL_TIMEOUT        5

#define START_HTTP         1
#define STOP_HTTP          2

#define TRUE               1
#define FALSE              0

#define MODE_ACTIVE        0
#define MODE_PASSIVE       1
#define MODE_MANUAL        2

#define OPERATOR_GREATER   0
#define OPERATOR_LESS      1
#define OPERATOR_EQUAL     2
#define OPERATOR_NOTEQUAL  3

#define TIME_SECOND	   1
#define TIME_MINUTE	   60
#define TIME_HOUR	   3600
#define TIME_DAY	   86400

#define ACTION_ALERT       1
#define ACTION_RESTART     2
#define ACTION_STOP        3

#define MAXIMUM(x,y)       ((x) > (y) ? (x) : (y))
#define is(a,b)		   ((a&&b)?!strcasecmp(a, b):0)


/** Replace the standard signal function with a more reliable using sigaction.*/
typedef void Sigfunc(int);
Sigfunc *signal(int signo, Sigfunc * func);
#if defined(SIG_IGN) && !defined(SIG_ERR)
#define SIG_ERR ((Sigfunc *)-1)
#endif

/** ------------------------------------------------- Synchronization macros */

#define LOCK(mutex) do { pthread_mutex_t *_yymutex = &(mutex); \
        pthread_mutex_lock(_yymutex);
#define END_LOCK pthread_mutex_unlock(_yymutex); } while (0)


/** ----------------------------------------------- Object Constructor macro */

#define NEW(p) (xcalloc(1, (long)sizeof *(p)))

/** ------------------------------------------ Simple Assert Exception macro */

#define ASSERT(e) if(!(e)) { error("AssertException: at %s:%d\naborting..\n", \
    __FILE__, __LINE__); abort(); }

/* --------------------------------------------------------- Data structures */

/** Defines data for application runtime */
struct myrun {
  char *controlfile;                /**< The file to read configuration from */
  char *logfile;                         /**< The file to write logdata into */
  char *localhostname;                      /**< The host name for localhost */
  char *mailserver;              /**< Mailserver used for alert notification */
  char *pidfile;                                  /**< This programs pidfile */
  char *mygroup;                              /**< Group Name of the Process */
  int  have_tty;                                 /**< Do we have a valid tty */
  int  debug;                   /**< Write debug information - TRUE or FALSE */
  int  use_syslog;                          /**< If TRUE write log to syslog */
  int  dolog;       /**< TRUE if program should log actions, otherwise FALSE */
  int  isdaemon;                 /**< TRUE if program should run as a daemon */
  int  polltime;        /**< In deamon mode, the sleeptime (sec) between run */
  int  dohttpd;                    /**< TRUE if monit HTTP server should run */
  int  httpdssl;                     /**< TRUE if monit HTTP server uses ssl */
  char *httpsslpem;                       /**< PEM file for the HTTPS server */
  int  clientssl;   /**< TRUE if monit HTTP server uses ssl with client auth */
  char *httpsslclientpem;      /**< PEM file/dir to check against at connect */
  int  servicessl;                 /**< TRUE if monit has ssl service checks */
  int  allowselfcert;   /**< TRUE if self certified client certs are allowed */
  int  httpdport;                    /**< The monit http server's portnumber */
  int  validate;	       /**< TRUE - use validate behaviour on startup */
  int  init; 		       /**< TRUE - don't background to run from init */
  int  doprocess;                 /**< TRUE if process status engine is used */
  char *bind_addr;                  /**< The address monit http will bind to */
  int  doreload;             /**< TRUE if a monit daemon should reinitialize */
  mode_t umask;                /**< The initial umask monit was started with */
  int  testing;   /**< Running in configuration testing mode - TRUE or FALSE */

  double loadavg[3];                                /**< Load average triple */

       /** An object holding program relevant "environment" data, see; env.c */
  struct myenvironment {
    char *user;             /**< The the effective user running this program */
    char *home;                                    /**< Users home directory */
    char *cwd;                                /**< Current working directory */
  } Env;
                      /** An object holding Basic Authentication information */
  struct myautentication {
    int  defined;              /**< TRUE if basic authentication can be used */
    char *uname;                 /**< User allowed to connect to monit httpd */
    char *passwd;                      /**< The users password in clear text */
  } Auth;
                                      /** User selected standard mail format */
  struct myformat {
    char *from;                          /**< The standard mail from address */
    char *subject;                            /**< The standard mail subject */
    char *message;                            /**< The standard mail message */
  } MailFormat;
                                          
  pthread_mutex_t mutex;    /**< Mutex used for process data synchronization */

};

/** Defines a port object */
typedef struct myport {
  volatile int socket;                       /**< Socket used for connection */
  int  type;                  /**< Socket type used for connection (UDP/TCP) */
  int  family;            /**< Socket family used for connection (INET/UNIX) */
  char *hostname;                                     /**< Hostname to check */
  int  port;                                                 /**< Portnumber */
  char *request;                              /**< Specific protocol request */
  char *pathname;                   /**< Pathname, in case of an UNIX socket */
  char *address;               /**< Human readable destination of the socket */

  ssl_connection * ssl;
  char *certmd5;
  /**< Object used for testing a port's service */
  struct myprotocol {
    const char *name;                                     /**< Protocol name */
    int(*check)(struct myport *);         /**< Service verification function */
  } *protocol;
  
  /** For internal use */
  struct myport *next;                               /**< next port in chain */
} *Port_T;

/** Defines a protocol object with protocol functions */
typedef struct myprotocol *Protocol_T;

typedef struct mydependant {
  char *dependant;                           /**< name of dependant process  */
  /** For internal use */
  struct mydependant *next;           /**< next dependant process in chain   */
} *Dependant_T;

/** Defines a checksum object */
typedef struct mychecksum {
  char *file;                  /**< A filename to compute a md5 checksum for */
  char *md5;                       /**< A md5 checksum computed for the file */

  /** For internal use */
  struct mychecksum *next;                       /**< next checksum in chain */
} *Checksum_T;

/** Defines a mailinglist object */
typedef struct mymail {
  char *to;                         /**< Mail address for alert notification */
  char *from;                                     /**< The mail from address */
  char *subject;                                       /**< The mail subject */
  char *message;                                       /**< The mail message */
  char *opt_message;                 /**< An optional message used in alerts */
  int  alert_on_timeout;       /**< If TRUE, alert user when process timeout */
  int  alert_on_restart;  /**< If TRUE, alert user when the process restarts */
  int  alert_on_checksum;    /**< If TRUE, alert user when the checksum fail */
  int  alert_on_resource;     /**< If TRUE, alert user when resources exceed */
  int  alert_on_stop;          /**< If TRUE, alert user when process stopped */
  int  alert_on_timestamp;  /**< If TRUE, alert user when the timestamp fail */

  /** For internal use */
  struct mymail *next;                          /**< next recipient in chain */
} *Mail_T;

/** Defines resource data */
typedef struct myresource {
  int  resource_id;                              /**< Which value is checked */
  long int limit;                                 /**< Limit of the resource */
  int  operator;                                    /**< Comparison operator */
  int  cycle;                                     /**< Cycle overrun counter */
  int  max_cycle;                                   /**< Cycle overrun limit */
  int  action;                                /**< Action in case of failure */

  struct myresource *next;
} *Resource_T;

/** Defines timestamp object */
typedef struct mytimestamp {
  char *pathname;				     /**< Path to the object */
  int  operator;				    /**< Comparison operator */
  int  time;					    /**< Timestamp watermark */
  int  action;				      /**< Action in case of failure */

  struct mytimestamp *next;
} *Timestamp_T;

/** Defines procfs (or other mechanism) data */
typedef struct myprocinfo {
  int  pid;
  int  status_flag;
  long mem_kbyte;    
  int  mem_percent;                                      /**< pecentage * 10 */
  int  cpu_percent;                                      /**< pecentage * 10 */
  /* private for calculating cpu_percent */
  double time_prev;                                        /**< 1/10 seconds */
  long  cputime;                                           /**< 1/10 seconds */
  long  cputime_prev;                                      /**< 1/10 seconds */
} *ProcInfo_T;

/**
 * Defines a command with ARGMAX optional arguments. The arguments
 * array must be NULL terminated and the first entry is the program
 * itself.
 */
typedef struct mycommand {
  char *arg[ARGMAX];                             /**< Program with arguments */
  int length;                         /**< The length of the arguments array */
} *Command_T;

/** Defines process data */
typedef struct myprocess {
  char *name;                                  /**< Process descriptive name */
  char *pidfile;                            /**< The pid file of the process */
  char *group;                                       /**< Process group name */
  Command_T start;                    /**< The start command for the process */
  Command_T stop;                      /**< The stop command for the process */
  int  do_validate;    /**< validation flag, if FALSE, no validation is done */
  int  mode;                            /**< Monitoring mode for the process */
  int  ncycle;                          /**< The number of the current cycle */
  int  nstart;           /**< The number of current starts with this process */
  int  to_start;                                  /**< Timeout start ceiling */
  int  to_cycle;                                  /**< Timeout cycle ceiling */
  int  every;                        /**< Check this program at given cycles */
  int  nevery;          /**< Counter for every.  When nevery == every, check */
  int  has_checksum_error;     /**< TRUE if the process has a checksum error */
  int  def_every;              /**< TRUE if every is defined for the process */
  int  def_timeout;          /**< TRUE if timeout is defined for the process */
  int  def_checksum;        /**< TRUE if checksum is defined for the process */
  int  def_procinfo;        /**< TRUE if procinfo is defined for the process */
  int  visited;      /**< Process visited flag, set if dependencies are used */
  int  depend_visited;/**< Depend visited flag, set if dependencies are used */

  Port_T portlist;                   /**< Portnumbers the process listens on */
  Checksum_T checksumlist;          /**< A list of file associated checksums */
  Mail_T maillist;                       /**< Alert notification mailinglist */
  Resource_T resourcelist;                           /**< Resouce check list */
  Dependant_T dependantlist;                      /**<Dependant process list */
  Timestamp_T timestamplist;                       /**< Timestamp check list */
  ProcInfo_T procinfo;                        /**< Data for the procfs check */
  
  /** For internal use */
  pthread_mutex_t mutex;          /**< Mutex used for action synchronization */
  struct myprocess *next;                         /**< next process in chain */
  struct myprocess *next_depend;           /**< next depend process in chain */
} *Process_T;


/* ------ Global variables ------------------------------------------------- */
Process_T processlist;                /**< The process list (created in p.y) */
char   *prog;                                  /**< The Name of this Program */
struct myrun Run;                      /**< Struct holding runtime constants */


/* Public prototypes */
void  error(const char *, ...);
char *trim(char *);
char *ltrim(char *);
char *rtrim(char *);
char *trim_quotes(char *);
int   starts_with(char *,char *);
void  handle_string_escapes(char *);
void  chomp(char *);
int   exist_process(char *);
Process_T get_process(char *);
void  printrunlist();
void  printprocess(Process_T);
void  printprocesslist();
pid_t get_pid(char *);
int   is_process_running(Process_T);
int   is_strdefined(char *);
int   isreg_file(char *);
char *stripfilename(char*);
int   exist_file(char *);
char *get_ctime();
char *get_RFC1123date(long *);
char *get_uptime(long delta);
char *get_process_uptime(char *);
char *get_md5sum(char *);
int   check_md5(char *, char *);
char *url_encode(char *uri);
char *get_basic_authentication_header();
int   set_md5sum(char **, char *);
int   parse(char *);
void  control(char *);
void  control_group(char *, char *);
void  d_check_process(char *, char *);
void  check_process(char *, char *);
void  setup_dependants();
void  reset_depend();
void  spawn(Process_T, Command_T);
void  status();
void  status_group(char *);
int   log_init() ;
void  log(const char *format, ...) ;
void  log_close();
int   get_log_fd();
void  validate() ;
void  daemonize() ;
void  gc();
void  gc_process(Process_T*);
void  gc_mail_list(Mail_T*);
void  init_files();
time_t get_timestamp(char *, mode_t);
void  finalize_files();
char *find_rcfile();    
int   create_pidfile(char *);
int   check_rcfile(char *);
int   kill_daemon(int);
int   exist_daemon(); 
void  sendmail(Mail_T);
int   sock_msg(int sock, char *, ...);
void  init_env();
void *xmalloc (int);
void *xcalloc(long, long);
char *xstrdup(const char *);
void *xresize(void *, long);
void  set_alarm_handler(void *);
void  monit_http(int);
int   check_httpd();
int   can_http();
char *format(const char *, va_list);
void  redirect_stdfd();
pid_t getpgid(pid_t);
int   check_file(char *, char *, int);

#endif
