/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif 

#include <fcntl.h>

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif 

#include <stdarg.h>
#include <signal.h>
#include <net/if.h>
#include <sys/un.h>

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif

#include "monitor.h"
#include "net.h"
#include "ssl.h"

/**
 *  General purpose Network and Socket methods.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: net.c,v 1.18 2003/02/11 21:27:39 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Test INET Socket connection to host on port with the
 * specified socket protocol type. The protocol is normaly 
 * either SOCK_STREAM or SOCK_DGRAM. 
 * @param hostname The host to connect to
 * @param port The portnumber to connect to
 * @param protocol The socket protocol to use
 * @return TRUE if the connect succeeded otherwise FALSE.
 */
int check_connect(char *hostname, int port, int protocol) {

  int socket;
  int rv= TRUE;

  ASSERT(hostname);
  
  if ((socket= create_socket(hostname, port, protocol)) < 0) {
    
    rv= FALSE;
    
  } else if (! check_socket(socket)) {

    rv= FALSE;

  }
  
  close_socket(socket);
  
  return rv;
  
}


/**
 * Check if the hostname resolves
 * @param hostname The host to check
 * @return TRUE if hostname resolves, otherwise FALSE
 */
int check_host(char *hostname) {
  
  struct hostent *hp;

  ASSERT(hostname);

  if ((hp = gethostbyname(hostname)) == NULL) {
    
    return FALSE;
    
  }
  
  return TRUE;
  
}


/**
 * Verify that the socket connection is ready for i|o. This test use
 * select for a stream based socket. For a dgram sockets the testing
 * technique is to send a 0 byte udp packet and "check" for an ICMP
 * error.
 * @param p A Port object
 * @return TRUE if the socket is ready, otherwise FALSE.
 */
int check_connection_io(Port_T p) {

  ASSERT(p);

  switch(p->type) {
   case SOCK_STREAM: return check_socket(p->socket);
   case SOCK_DGRAM:  return check_udp_socket(p->socket);
   default:          break;
  }

  return FALSE;
  
}


/**
 * Verify that the socket is ready for i|o
 * @param socket A socket
 * @return TRUE if the socket is ready, otherwise FALSE.
 */
int check_socket(int socket) {
  
  fd_set rset, wset;
  struct timeval tv;

  FD_ZERO(&rset);
  FD_ZERO(&wset);
  FD_SET(socket, &rset);
  FD_SET(socket, &wset);
  tv.tv_sec= SELECT_TIMEOUT;
  tv.tv_usec= 0;

  return (select(socket+1, &rset, &wset, NULL, &tv) > 0);
}


/**
 * Verify that the udp socket is ready for i|o. The given socket must
 * be a connected udp socket. The test is conducted by sending one
 * byte to the server and check for any returned ICMP error when
 * reading from the socket. A better test would be to send an empty
 * udp package to avoid possibly raising an error from the server we
 * are testing.
 * @param socket A socket
 * @return TRUE if the socket is ready, otherwise FALSE.
 */
int check_udp_socket(int socket) {

  int r;
  char buf[1]= {0};

  /*
   * R/W is asynchronous and we should probably loop and wait longer
   * for a possible ICMP error.
   */
  set_noblock(socket);
  write(socket, buf, 1);
  sleep(1);
  r= read(socket, buf, 1);
  set_block(socket);
  if(0>r) {
    switch(errno) {
    case ECONNREFUSED: return FALSE;
    default:           break;
    }
  }
  
  return TRUE;

}


/**
 * Open a socket against hostname:port with the given protocol.
 * The protocol is normaly either SOCK_STREAM or SOCK_DGRAM.
 * @param hostname The host to open a socket at
 * @param port The port number to connect to
 * @param protocol Socket protocol to use (SOCK_STREAM|SOCK_DGRAM)
 * @return The socket or -1 if an error occured.
 */
int create_socket(char *hostname, int port, int protocol) {

  int s;
  struct hostent *hp;
  struct sockaddr_in sin;
  
  ASSERT(hostname);

  if ((hp= gethostbyname(hostname)) == NULL) {
    
    return -1;
    
  }
  
  if ((s= socket(AF_INET, protocol, 0)) < 0) {
    
    return -1;
    
  }
  
  sin.sin_family= AF_INET;
  sin.sin_port= htons(port);
  memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);

  if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {

    close_socket(s);
    return -1;
    
  }

  return s;
  
}


/**
 * Open a socket with given Port_T structure
 * The protocol, destination and type are selected appropriately
 * @param pp connection description
 * @return The socket or -1 if an error occured.
 */
int create_generic_socket(Port_T pp) {

  int socket_fd= -1;

  ASSERT(pp);

  if(pp->family == AF_INET) {

    socket_fd= create_socket(pp->hostname, pp->port, pp->type);

  } else if(pp->family == AF_UNIX) {

    socket_fd= create_unix_socket(pp->pathname, pp->type);

  }

  return socket_fd;
  
}


/**
 * Open a socket against pathname with the given protocol.
 * The protocol is normaly either SOCK_STREAM or SOCK_DGRAM.
 * @param pathname The unix socket to connect to
 * @param protocol Socket protocol to use (SOCK_STREAM|SOCK_DGRAM)
 * @return The socket or -1 if an error occured.
 */
int create_unix_socket(char *pathname, int protocol) {

  int s;
  struct sockaddr_un unixsocket;
  
  ASSERT(pathname);

  if ((s= socket(PF_UNIX, protocol, 0)) < 0) {
    
    return -1;
    
  }
  
  unixsocket.sun_family= AF_UNIX;
  snprintf(unixsocket.sun_path, sizeof(unixsocket.sun_path), "%s", pathname);

  if (connect(s, (struct sockaddr *)&unixsocket, sizeof(unixsocket)) < 0) {

    close_socket(s);
    return -1;
    
  }

  return s;
  
}


/**
 * Creates a server socket (SOCK_STREAM type) and binds it to the
 * specified local port number, with the specified backlog. Set a
 * socket option to make the port reusable again. This is useful if
 * the server has been shut down, and then restarted right away while
 * sockets are still active on its port. If a bind address is given
 * the socket will only accept connect requests to this addresses. If
 * the bind address is NULL it will default accepting connections on
 * any/all local addresses
 * @param port The localhost port number to open
 * @param backlog The maximum queue length for incomming connections
 * @param bindAddr the local address the server will bind to
 * @return The socket ready for accept, or -1 if an error occured.
 */
int create_server_socket(int port, int backlog, char *bindAddr) {
  
  int s;
  int flag= 1;
  struct sockaddr_in myaddr;

  if ((s= socket(AF_INET, SOCK_STREAM, 0)) < 0)
      return -1;
  
  memset(&myaddr, 0, sizeof(struct sockaddr_in));
  myaddr.sin_family= AF_INET;
  myaddr.sin_port= htons(port);
  
  if(bindAddr) {
    struct hostent *h= gethostbyname(bindAddr);
    endhostent();
    if(h==NULL) {
      errno= h_errno;
      goto error;
    }
    myaddr.sin_addr= *(struct in_addr*)h->h_addr_list[0];
  } else {
    myaddr.sin_addr.s_addr= htonl(INADDR_ANY);
  }
  
  if (setsockopt(s, SOL_SOCKET,
                  SO_REUSEADDR, (char *)&flag, sizeof(int)) < 0) 
      goto error;

  if (bind(s, (struct sockaddr *)&myaddr, sizeof(struct sockaddr_in)) < 0)
      goto error;
  
  if (listen(s, backlog) < 0)
      goto error;
  
  if (fcntl(s, F_SETFD, FD_CLOEXEC) == -1)
      goto error;
  
  return s;

  error:
  close(s);

  return -1;

}


/**
 * Shutdown a socket and close the descriptor.
 * @param socket The socket to shutdown and close
 * @return TRUE if the close succeed otherwise FALSE
 */
int close_socket(int socket) {

  int rv= TRUE;

  if ((shutdown(socket, 2) < 0)) {
    
    rv= FALSE;
    
  }
  
  if (close(socket) < 0) {
    
    rv= FALSE;
    
  }

  return rv;

}


/**
 * Set a so_timeout (read/write) for the given socket with the
 * specified timeout in seconds.
 * @param socket A socket
 * @param timeout Read/Write timeout in seconds
 * @return TRUE if success otherwise FALSE
 */
int set_sotimeout(int socket, int timeout) {

  struct timeval tv= {timeout, 0};

  return ((setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv))==0) &&
          (setsockopt(socket, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv))==0));

}


/**
 * Enable nonblocking i|o on the given socket.
 * @param socket A socket
 * @return TRUE if success, otherwise FALSE
 */
int set_noblock(int socket) {
  
  int flag= 1;

  return (! ioctl(socket, FIONBIO, &flag));

}


/**
 * Disable nonblocking i|o on the given socket
 * @param socket A socket
 * @return TRUE if success, otherwise FALSE
 */
int set_block(int socket) {

  int flag= 0;

  return (! ioctl(socket, FIONBIO, &flag));

}

/**
 * Send the message 'msg' down a port in non-blocking mode and
 * with failover. Uses SSL if necessary.
 * @param p port handler
 * @param msg The message to send
 * @param len The length of the message in bytes
 * @param flags Optional send flags
 * @return The bytes sent or -1 if an error occured
 */
int port_send(Port_T p, const char *msg, int len, int flags) {

  ASSERT(p);
  ASSERT(msg);

  if (p->ssl!=NULL) {

    return send_ssl_socket(p->ssl, (void *) msg, len);

  } else {

    return sock_send(p->socket, msg, len, flags);

  }

}

/**
 * Read a message from the socket in non-blocking mode and
 * with failover. Uses SSL if necessary.
 * @param sock A port handler
 * @param buf The buffer to read a message into
 * @param len The length of the buffer
 * @param flags Optional recv flags
 * @return Returns the bytes read or -1 if an error occured
 */
int port_recv(Port_T p, char *buf, int len, int flags) {

  ASSERT(p);
  ASSERT(buf);

  if (p->ssl!=NULL) {
    int error;
    error = recv_ssl_socket(p->ssl, buf, len);

    return error;

  } else {

    return sock_recv(p->socket, buf, len, flags);

  }
}


/**
 * Send the message 'msg' down the socket in non-blocking mode and
 * with failover. Inspired by code from the Apache web server.
 * @param sock A socket
 * @param msg The message to send
 * @param len The length of the message in bytes
 * @param flags Optional send flags
 * @return The bytes sent or -1 if an error occured
 */
int sock_send(int sock, const char *msg, int len, int flags) {

  int rv;
  fd_set fdset;
  struct timeval t;
  int retry;

  ASSERT(msg);

  if (! set_noblock(sock)) {
    
    return -1;
    
  }

  rv= send(sock, msg, len, flags);
  if (rv <= 0) {
    
    if (errno == EWOULDBLOCK) 
	do {
	  
	  retry= FALSE;
	  FD_ZERO(&fdset);
	  FD_SET(sock, &fdset);
	  t.tv_sec= SELECT_TIMEOUT;
	  t.tv_usec= 0;
	  rv = select(sock+1, NULL, &fdset, NULL, &t);
	  if (rv <= 0) {
	    
	    rv= -1;
	    
	  }
	  else {
	    
	    rv = send(sock, msg, len, flags);
	    if (rv <= 0) {
	      
	      if(errno == EWOULDBLOCK) {
		
		retry= TRUE;
		sleep(1);
		
	      }
	      
	    }
	    
	  }
	  
	} while(retry);
    
  }
  
  set_block(sock);

  return rv;
  
}


/**
 * Read a message from the socket in non-blocking mode and
 * with failover. Inspired by code from the Apache web server.
 * @param sock A socket
 * @param buf The buffer to read a message into
 * @param len The length of the buffer
 * @param flags Optional recv flags
 * @return Returns the bytes read or -1 if an error occured
 */
int sock_recv(int sock, char *buf, int len, int flags) {

  int rv;
  fd_set fdset;
  struct timeval t;

  ASSERT(buf);

  if (! set_noblock(sock)) {
    
    return -1;
    
  }
  
  rv= recv(sock, buf, len, flags);
  if (rv <= 0) {
    
    if (errno == EWOULDBLOCK) {
      
      FD_ZERO(&fdset);
      FD_SET(sock, &fdset);
      t.tv_sec= SELECT_TIMEOUT;
      t.tv_usec = 0;
      rv = select(sock+1, &fdset, NULL, NULL, &t);
      if (rv <= 0) {
	
	rv= -1;
	
      }
      else {
	
	rv= recv(sock, buf, len, flags);
	
      }
      
    }
    
  }

  set_block(sock);

  return rv;

}


/**
 * Get the local host name
 * @return the name of the localhost.
 * NULL if lookup failed
 */
char *get_localhostname() {

  char hostname[256];

  if (gethostname(hostname, sizeof(hostname)) < 0) {
    
    return NULL;
    
  }

  return (strdup(hostname));

}


/**
 * Get the remote host name from the originating host.
 * @param socket A socket
 * @return the (official) name (A-Record) of the peer
 * or NULL if lookup failed.
 */
char *get_remote_host(int socket) {

  struct hostent *h;
  struct sockaddr_in r;
  int l= sizeof(struct sockaddr_in);

  if (getpeername(socket, (struct sockaddr*)&r, &l) < 0) {
    
    return NULL;
    
  } else {
    
    h=gethostbyaddr((char *)&r.sin_addr, sizeof(struct in_addr), AF_INET);
    if (h) {
      
      return strdup(h->h_name);
      
    }
    
  }

  return NULL;

}

