/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


%{
  
/*
 * DESCRIPTION
 *   Simple context-free grammar for parsing the control file. 
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Olivier Beyssac, <ob@r14.freenix.oraddmailg>
 *  @author Kianusch Sayah Karadji <kianusch.sayah.karadji@sk-tech.net>
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *  @author Rory Toma <rory@digeo.com>
 *  @version \$Id: p.y,v 1.60 2003/02/14 08:22:33 martinp Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
  
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
  
#include <sys/socket.h>

#ifdef HAVE_ASM_PARAM_H
#include <asm/param.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif


#include "net.h"
#include "monitor.h"
#include "protocol.h"
#include "engine.h"
#include "alert.h"
#include "monit_process.h"
#include "ssl.h"

#define YERROR(msg) { yyerror(msg); YYABORT; }

  struct IHavePrecedence {
    int daemon;
    int logfile;
    int pidfile;
  }; 

  struct MailFilter {
    char *from;
    char *subject;
    char *message;
    int  timeout;
    int  restart;
    int  checksum;
    int  resource;
    int  stop;
    int  timestamp;
  };

  struct PortSet {
    int  socket;
    char *hostname;
    int  port;
    int  type;
    int  family;
    int  ssl;
    int  sslversion;
    char *request;
    char *pathname;
    char *certmd5;
    Protocol_T protocol;
  };

  struct ResourceSet {
    int  resource_id;
    int  limit;
    int  operator;
    int  max_cycle;
    int  action;
  };

  struct TimestampSet {
    char *pathname;
    int  operator;
    int  time;
    int  action;
  };

  /* yacc interface */
  void  yyerror(const char *);

  /* lexer interface */
  int yylex(void);
  extern FILE *yyin;
  extern int lineno;
  extern char *yytext;
  
  /* Local variables */
  static Process_T tail= NULL;
  static int cfg_errflag= FALSE;
  static Process_T current= NULL;
  static Command_T command= NULL;
  static Process_T depend_list= NULL;
  static struct IHavePrecedence ihp= {FALSE, FALSE, FALSE};
  static struct MailFilter mtf= {NULL, NULL, NULL, FALSE, FALSE, FALSE,
				 FALSE, FALSE, FALSE};
  static struct PortSet portset= {-1, NULL, 0, SOCK_STREAM, AF_INET, FALSE, 
                                  SSL_VERSION_AUTO, NULL, NULL, NULL, NULL};
  static struct ResourceSet resourceset= {0, 0, OPERATOR_EQUAL, 1,
					  ACTION_ALERT};
  static struct TimestampSet timestampset= {NULL, OPERATOR_EQUAL, 0,
					    ACTION_ALERT};

  /* Private prototypes */
  static void initialize();
  static void addprocess(Process_T);
  static void addmail(char *, struct MailFilter *);
  static void createprocess(char *, char *);
  static void adddependant(char *);
  static void addchecksum(char *, char *);
  static void addport(struct PortSet *);
  static void addresource(struct ResourceSet *);
  static void addtimestamp(struct TimestampSet *);
  static void *addprotocol(int);
  static void addcommand(int);
  static void addargument(char *);
  static void setlogfile(char *);
  static void setpidfile(char *);
  static void setmailserver(char *);
  static void setuname(char *);
  static void setpasswd(char *);
  static void reset_mailfilter();
  static void reset_runmail();
  static void reset_portset();
  static void reset_resourceset();
  static void reset_timestampset();
  static void check_name(char *);
  static void check_timeout(int, int);
  static void check_every (int);
  static void check_hostname (char *);
  static void createprocinfo();
  static char *append_hostname(char *);
  static void cleanup_md5_string(char *);
  static void check_depend();
  static void order_depend(Process_T);
  static void validate_depend(Process_T, int *);
%}

%union {
  float real;
  int   number;
  char *string;
}

%token SET LOGFILE DAEMON SYSLOG MAILSERVER HTTPD ALLOW ADDRESS INIT
%token PEMFILE ENABLE DISABLE HTTPDSSL CLIENTPEMFILE ALLOWSELFCERTIFICATION
%token CHECK PIDFILE START STOP
%token HOST PORT TYPE UDP TCP TCPSSL PROTOCOL
%token ALERT MAILFORMAT UNIXSOCKET
%token TIMEOUT RESTART CHECKSUM EXPECT EVERY 
%token DEFAULT HTTP FTP SMTP POP IMAP NNTP SSH DWP LDAP2 LDAP3 CERTMD5
%token <string> STRING PATH MAILADDR MAILFROM MAILSUBJECT 
%token <string> MAILBODY
%token <number> NUMBER
%token <real> REAL
%token CPUUSAGE MEMUSAGE MEMKBYTE RESOURCE LOADAVG1 LOADAVG5 LOADAVG15
%token AUTOSTART MODE ACTIVE PASSIVE MANUAL
%token GROUP REQUEST DEPENDS
%token TIMESTAMP SECOND MINUTE HOUR DAY
%token SSLAUTO SSLV2 SSLV3 TLSV1

%left GREATER LESS EQUAL NOTEQUAL

%%
cfgfile         : /* EMPTY */
                | statement_list
                ;

statement_list  : statement
                | statement_list statement
                ;

statement       : setdaemon
                | setlog
                | setmta
                | setmailformat
                | sethttpd
                | setpid
                | setinit
                | check optionlist
                ;

optionlist      : /* EMPTY */
                | optionlist optionstatement
                ;

optionstatement : start
                | stop
                | host port type protocol {
                   addport(&portset);
                  }
                | unixsocket type protocol {
                   addport(&portset);
                  }
		| timestamp {
		   addtimestamp(&timestampset);
		  }
                | timeout
                | every
                | alert
		| checksum
		| autostart
		| mode
		| group
                | depend
		| resource resourcecycle action {
                   resourceset.action= $<number>3;
		   addresource(&resourceset);
		  }
                ;

setdaemon       : SET DAEMON NUMBER  {
                   if (!Run.isdaemon || ihp.daemon) {
		     ihp.daemon= TRUE;
		     Run.isdaemon= TRUE;
		     Run.polltime= $3;
		   }
                 }
                ;

setinit         : SET INIT {
                   Run.init= TRUE;
                 }
                ;

setlog          : SET LOGFILE PATH   {
                   if (!Run.logfile || ihp.logfile) {
		     ihp.logfile= TRUE;
		     setlogfile($3);
		     Run.use_syslog= FALSE;
		     Run.dolog=TRUE;
		   }
                 }
                | SET LOGFILE SYSLOG {
		   if (!Run.logfile || ihp.logfile) {
		     ihp.logfile= TRUE;
		     setlogfile(xstrdup("syslog"));
		     Run.use_syslog= TRUE;
		     Run.dolog=TRUE;
		   }
		 }
                ;

setpid          : SET PIDFILE PATH {
                   if (!Run.pidfile || ihp.pidfile) {
		     ihp.pidfile= TRUE;
                     setpidfile($3);
		   }
                 }
                ;

setmta          : SET MAILSERVER STRING   {
                   check_hostname($3);
		   setmailserver($3);
                 }
                ;

setmailformat   : SET MAILFORMAT '{' formatoptionlist '}' {
                   Run.MailFormat.from=
			mtf.from?
			mtf.from:
			xstrdup(ALERT_FROM);
                   Run.MailFormat.subject=
			mtf.subject?
			mtf.subject:
			xstrdup(ALERT_SUBJECT);
                   Run.MailFormat.message=
			mtf.message?
			mtf.message:
			xstrdup(ALERT_MESSAGE);
		   reset_mailfilter();
                 }
                ;

sethttpd        : SET HTTPD PORT NUMBER httpdlist {
                   Run.dohttpd= TRUE;
		   Run.httpdport= $4;
                 }
                ;


httpdlist	: /* EMPTY */
                | httpdlist httpdoption
                ; 

httpdoption     : ssl
		| bindaddress
                | allow
                ;

bindaddress     : ADDRESS STRING { Run.bind_addr= $2; }
                ;

ssl             : ssldisable { Run.httpdssl= FALSE; }
                | sslenable pemfile clientpemfile allowselfcert { 
		  Run.httpdssl= TRUE;                   
		  if(!have_ssl()) {
		    log("%s: SSL is not supported at line %d\n", 
			prog, lineno-1);
		     cfg_errflag++;
                   }
		 }
                ;

sslenable       : HTTPDSSL
                | HTTPDSSL ENABLE
                | ENABLE HTTPDSSL
                ;

ssldisable      : HTTPDSSL DISABLE
                | DISABLE HTTPDSSL
                | ssldisable PEMFILE PATH
                | ssldisable CLIENTPEMFILE PATH
                | ssldisable ALLOWSELFCERTIFICATION
                ;

pemfile         : PEMFILE PATH {
                     Run.httpsslpem= $2;
                     if(!check_file(Run.httpsslpem,
				    "SSL server PEM file", S_IRWXU))
			 yyerror("Parse error");
                  }
                ;

clientpemfile   : /* EMPTY */ 
                | CLIENTPEMFILE PATH {
                     Run.httpsslclientpem= $2; 
		     Run.clientssl= TRUE;
                     if(!check_file(Run.httpsslclientpem,
				    "SSL client PEM file", 
				    S_IRWXU | S_IRGRP | S_IROTH))
			 yyerror("Parse error");
                  }
                ;

allowselfcert   : /* EMPTY */ { 
                   Run.allowselfcert= FALSE; 
                   config_ssl(Run.allowselfcert); 
                 }
                | ALLOWSELFCERTIFICATION {   
                   Run.allowselfcert= TRUE;
                   config_ssl(Run.allowselfcert); 
                 }
                ;

allow           : ALLOW STRING':'STRING {
                   setuname($2);
		   setpasswd($4);
		   Run.Auth.defined= TRUE;
                 }
                | ALLOW STRING {
		   if(!add_host_allow($2)) {
		     log("%s: Hostname '%s' did not resolve at line %d\n",
			 prog, $2, lineno-1);
		     cfg_errflag++;
		   }
		   free($2);
		 }
                ;

check           : CHECK STRING PIDFILE PATH {
                   createprocess($2,$4);
		   check_name($2);
                  }
                ;

start           : START argumentlist { addcommand(START); }
                ;

stop            : STOP argumentlist { addcommand(STOP); }
                ;

argumentlist    : argument
                | argumentlist argument
                ;

argument        : STRING { addargument($1); } | PATH { addargument($1); }
                ;


host            : /* EMPTY */ { portset.hostname= xstrdup(LOCALHOST); }
                | HOST STRING { check_hostname($2); portset.hostname= $2; }
		;

port            : PORT NUMBER { portset.port= $2; portset.family= AF_INET; }
                ;

unixsocket      : UNIXSOCKET PATH {
                    portset.pathname= $2; portset.family= AF_UNIX;
                  }
                ;

type            : /* EMPTY */ { portset.type= SOCK_STREAM; }
                | TYPE TCP    { portset.type= SOCK_STREAM;  
		                Run.servicessl= TRUE; }
                | TYPE TCPSSL sslversion certmd5  { portset.type= SOCK_STREAM;
		                                    portset.ssl= TRUE; }
                | TYPE UDP    { portset.type= SOCK_DGRAM; }
                ;

certmd5         : /* EMPTY */    { portset.certmd5= NULL; }
                | CERTMD5 STRING { portset.certmd5= $2; }
                | EXPECT CERTMD5 STRING { portset.certmd5= $3; }
                ;

sslversion      : /* EMPTY */  { portset.sslversion= SSL_VERSION_AUTO; }
                | SSLV2        { portset.sslversion= SSL_VERSION_SSLV2; }
                | SSLV3        { portset.sslversion= SSL_VERSION_SSLV3; }
                | TLSV1        { portset.sslversion= SSL_VERSION_TLS; }
                | SSLAUTO      { portset.sslversion= SSL_VERSION_AUTO; }
                ;

protocol        : /* EMPTY */  { portset.protocol= addprotocol(P_DEFAULT); }
                | PROTOCOL DEFAULT { portset.protocol= addprotocol(P_DEFAULT); }
                | PROTOCOL HTTP request {portset.protocol= addprotocol(P_HTTP);}
                | PROTOCOL FTP { portset.protocol= addprotocol(P_FTP); }
                | PROTOCOL SMTP { portset.protocol= addprotocol(P_SMTP); }
                | PROTOCOL POP { portset.protocol= addprotocol(P_POP); }
                | PROTOCOL IMAP { portset.protocol= addprotocol(P_IMAP); }
                | PROTOCOL NNTP { portset.protocol= addprotocol(P_NNTP); }
                | PROTOCOL SSH  { portset.protocol= addprotocol(P_SSH); }
                | PROTOCOL DWP  { portset.protocol= addprotocol(P_DWP); }
                | PROTOCOL LDAP2 { portset.protocol= addprotocol(P_LDAP2); }
                | PROTOCOL LDAP3 { portset.protocol= addprotocol(P_LDAP3); }
                ;     
		   
request         : /* EMPTY */
                | REQUEST PATH  { portset.request= url_encode($2); free($2); }
                ;

timeout         : TIMEOUT NUMBER NUMBER {
                   check_timeout($2, $3);
		   current->def_timeout= TRUE;
		   current->to_start= $2;
		   current->to_cycle= $3;
                 }
                ;

alert           : alertmail '{' alertoptionlist '}' formatlist {
                   addmail($<string>1, &mtf);
                 }
                | alertmail formatlist {
                   addmail($<string>1, &mtf);
		 }
                ;

alertmail       : ALERT MAILADDR { $<string>$= $2; }
                | ALERT STRING { $<string>$= append_hostname($2); }
		;

alertoptionlist : alertoption
                | alertoptionlist alertoption
                ;

alertoption     : TIMEOUT   { mtf.timeout= TRUE; }
                | RESTART   { mtf.restart= TRUE; }
                | CHECKSUM  { mtf.checksum= TRUE; }
                | RESOURCE  { mtf.resource= TRUE; }
                | STOP      { mtf.stop= TRUE; }
                | TIMESTAMP { mtf.timestamp= TRUE; }
                ;

formatlist      : /* EMPTY */
	        | MAILFORMAT '{' formatoptionlist '}'
                ;

formatoptionlist: formatoption
                | formatoptionlist formatoption
                ;

formatoption    : MAILFROM { mtf.from= $1; }
                | MAILSUBJECT { mtf.subject= $1; }
                | MAILBODY { mtf.message= $1; }
                ;

every           : EVERY NUMBER {
                   check_every($2);
		   current->def_every= TRUE;
		   current->every= $2;
                 }
                ;

checksum        : CHECKSUM checksumlist { current->def_checksum= TRUE; }
                ;

checksumlist    : program
                | checksumlist program 
                ;

program         : PATH { addchecksum($1, NULL); }
                | PATH EXPECT STRING { addchecksum($1, $3); }
                ;

autostart       : AUTOSTART {
                    YERROR("Warning: 'autostart' is deprecated - "
			   "use the 'mode' statement instead");
                  }
                ;

mode            : MODE ACTIVE  { current->mode= MODE_ACTIVE; }
                | MODE PASSIVE { current->mode= MODE_PASSIVE; }
                | MODE MANUAL  { current->mode= MODE_MANUAL;
		    current->do_validate= FALSE;
		  }
		;

group		: GROUP STRING { current->group= $2; }
		;

depend          : DEPENDS dependlist
                ;

dependlist      : dependant
                | dependlist dependant
                ;
 
dependant       : STRING { adddependant($1); }
                ;

resource        : CPUUSAGE operator REAL { 
                    resourceset.resource_id= RESOURCE_ID_CPU_PERCENT;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($3 * 10.0); 
                  }
                | CPUUSAGE operator NUMBER {
		    yyerror("Expecting a real number (e.g. 10.0) Got");
		  }
                | MEMUSAGE operator REAL { 
                    resourceset.resource_id= RESOURCE_ID_MEM_PERCENT;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($3 * 10.0); 
		  }
                | MEMUSAGE operator NUMBER { 
		    yyerror("Expecting a real number (e.g. 15.0) Got");
		  }
                | MEMKBYTE operator NUMBER { 
                    resourceset.resource_id= RESOURCE_ID_MEM_KBYTE;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) $3; 
		  }
                | resourceload operator REAL { 
                    resourceset.resource_id= $<number>1;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($3 * 10.0); 
		  }
                | resourceload operator NUMBER { 
		    yyerror("Expecting a real number (e.g. 5.0) Got");
		  }
                ;

resourceload    : LOADAVG1  { $<number>$= RESOURCE_ID_LOAD1; }
                | LOADAVG5  { $<number>$= RESOURCE_ID_LOAD5; }
                | LOADAVG15 { $<number>$= RESOURCE_ID_LOAD15; }
                ;

resourcecycle   : /* EMPTY */
                | NUMBER { resourceset.max_cycle= $1; }
                ;

timestamp       : TIMESTAMP PATH operator NUMBER time action {
                    timestampset.pathname= $2;
		    timestampset.operator= $<number>3;
		    timestampset.time= ($4 * $<number>5);
		    timestampset.action= $<number>6;
                  }
                ;

operator        : /* EMPTY */ { $<number>$= OPERATOR_EQUAL; }
                | GREATER     { $<number>$= OPERATOR_GREATER; }
                | LESS        { $<number>$= OPERATOR_LESS; }
                | EQUAL       { $<number>$= OPERATOR_EQUAL; }
                | NOTEQUAL    { $<number>$= OPERATOR_NOTEQUAL; }
                ;

time            : /* EMPTY */ { $<number>$= TIME_SECOND; }
                | SECOND      { $<number>$= TIME_SECOND; }
                | MINUTE      { $<number>$= TIME_MINUTE; }
                | HOUR        { $<number>$= TIME_HOUR; }
                | DAY         { $<number>$= TIME_DAY; }
                ;

action          : ALERT       { $<number>$= ACTION_ALERT; }
                | RESTART     { $<number>$= ACTION_RESTART; }
                | STOP        { $<number>$= ACTION_STOP; }
                ;

%%


/* -------------------------------------------------------- Parser interface */


/**
 * Syntactic error routine
 */
void yyerror(const char *msg) {

  ASSERT(msg);

  log("%s: %s '%s' at line %d\n", prog, msg, yytext, lineno);
  cfg_errflag++;

}


/*
 * The Parser hook - start parsing the control file
 * Returns TRUE if parsing succeeded, otherwise FALSE
 */
int parse(char *controlfile) {

  ASSERT(controlfile);

  processlist= tail= current= NULL;
  
  if(! check_rcfile(controlfile)) {
    
    return (FALSE);
    
  }
  
  if ((yyin = fopen(controlfile,"r")) == (FILE *)NULL) {
    
    log("%s: Error opening the control file '%s' -- %s\n",
	prog, controlfile, STRERROR);
    return(FALSE);
    
  }

  /* Creation of the global process list is synchronized  */
  LOCK(Run.mutex)
      
      initialize();
      yyparse();
      fclose(yyin);

      /* If defined - add the last process to the process list */
      if(current) {
	
	createprocinfo();
	addprocess(current);
	
      }
      
      /* Check the sanity of any dependency graph */
      check_depend();
  END_LOCK;
    
  return(cfg_errflag == 0);
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Initialize objects. Placeholder for setting values that can change
 * between cycles (changed config file) and not easily supported by
 * the grammar implementation.
 */
static void initialize() {

  destroy_hosts_allow(); 
  Run.dolog= FALSE;
  Run.dohttpd= FALSE;
  Run.Auth.defined= FALSE;
  Run.servicessl= FALSE;
  Run.clientssl= FALSE;
  depend_list= NULL;
  reset_runmail();
  if(Run.bind_addr) {
    free(Run.bind_addr);
    Run.bind_addr= NULL;
  }
  Run.httpdssl= FALSE;
  
}


/*
 * Create a new process object and add any current objects to the
 * process list.
 */
static void createprocess(char *name, char *pidfile) {

  ASSERT(name);
  ASSERT(pidfile);

  if(current) {

    createprocinfo();
    addprocess(current);
    
  }
  else {
    
    current= NEW(current);
    
  }

  /* Reset the current object */
  memset(current, 0, sizeof(*current));

  /* Set default values */
  current->do_validate= TRUE;
  current->mode= MODE_ACTIVE;

  current->name= name;
  current->pidfile= pidfile;
  current->group= NULL;
  current->dependantlist= NULL;

  pthread_mutex_init(&current->mutex, NULL);

}


/*
 * Add a process object to the processlist
 */
static void addprocess(Process_T p) {
  
  Process_T n;

  ASSERT(p);
 
  /* if process is already running... set validate true  */ 
  if (is_process_running(p))
      p->do_validate= TRUE;
 
  n= NEW(n);
  memcpy(n, p, sizeof(*p));

  /* Add the process to the end of the process list */
  if(tail != NULL) {
    
    tail->next= n;
    
  } else {
    
    processlist= n;
    
  }
  
  tail= n;

}


/* 
 * Add a dependant entry to the current process dependant list
 *
 */
static void adddependant(char *dependant) {

  Dependant_T d= NEW(d);

  ASSERT(dependant);

  if (current->dependantlist != NULL) {
    d->next= current->dependantlist;
  }

  d->dependant=dependant;
  current->dependantlist= d;

}


/*
 * Add a filename and its associated md5 checksum
 * to the current process checksumlist
 */
static void addchecksum(char *filename, char *sum) {

  Checksum_T c= NEW(c);

  ASSERT(filename);

  c->file= filename;
  
  if(sum) {
    
    c->md5= sum;

    
  } else if(! set_md5sum(&c->md5, filename)) {
    
    log("%s: Cannot compute a checksum for %s at line %d\n",
	prog, filename, lineno-1);
    cfg_errflag++;

  }

  
  if (c->md5 != NULL) {
   
    cleanup_md5_string(c->md5);

  }

  c->next= current->checksumlist;
  current->checksumlist= c;

}


/*
 * Add the given mailaddress with the apropriate alert notification
 * mask { TIMEOUT RESTART CHECKSUM RESOURCE STOP TIMESTAMP } and mail
 * attributes to the current process's mailinglist.
 */
static void addmail(char *mailto, struct MailFilter *f) {

  Mail_T m= NEW(m);

  ASSERT(mailto);

  if(f->timeout  ||
       f->restart  ||
       f->checksum ||
       f->resource ||
       f->stop     ||
       f->timestamp) {
    
    m->alert_on_timeout= f->timeout;
    m->alert_on_restart= f->restart;
    m->alert_on_checksum= f->checksum;
    m->alert_on_resource= f->resource;
    m->alert_on_stop= f->stop;
    m->alert_on_timestamp= f->timestamp;
    
  } else {

    m->alert_on_timeout= TRUE;
    m->alert_on_restart= TRUE; 
    m->alert_on_checksum= TRUE;
    m->alert_on_resource= TRUE;
    m->alert_on_stop= TRUE;
    m->alert_on_timestamp= TRUE;
    
  }
  
  m->to= mailto;
  
  m->from= f->from;
  m->subject= f->subject;
  m->message= f->message;
  
  m->next= current->maillist;
  current->maillist= m;

  reset_mailfilter();

}


/*
 * Add the given portset to the current process's portlist
 */
static void addport(struct PortSet *pp) {

  char address[STRLEN];
  Port_T p= NEW(p);

  ASSERT(pp);

  p->socket= pp->socket;
  p->hostname= pp->hostname;
  p->port= pp->port;
  p->type= pp->type;
  p->family= pp->family;
  p->request= pp->request;
  p->protocol= pp->protocol;
  p->pathname= pp->pathname;
  p->certmd5= pp->certmd5;

  if (p->certmd5 != NULL) {
    
    cleanup_md5_string(p->certmd5);

  }

  if(pp->family == AF_INET) {
    snprintf(address, STRLEN, "INET[%s:%d]", pp->hostname, pp->port);
  } else if(pp->family == AF_UNIX) {
    snprintf(address, STRLEN, "UNIX[%s]", pp->pathname);
  }

  if(pp->ssl == TRUE) {

    if(!have_ssl()) {

      log("%s: The ssl check cannot be activated at line %d, " 
	  "SSL is not support.\n",
	  prog,lineno-1);
      cfg_errflag++;

    } else {

      p->ssl= new_ssl_connection(NULL, pp->sslversion);

    }

  } else {

    p->ssl= NULL;

  }

  p->address= xstrdup(address);
  
  p->next= current->portlist;
  current->portlist= p;
  
  reset_portset();

}


/*
 * Add a new resource object to the current process resource list
 */
static void addresource(struct ResourceSet *rr) {

  Resource_T r= NEW(r);

  ASSERT(rr);

  if(! Run.doprocess) {

    log("%s: Cannot activate process check at line %d\n"
	"\t(The process status engine was disabled. On certain"
	" systems you must\n\trun monit as root to utilize this"
	" feature)\n", prog, lineno-1);
    cfg_errflag++;
    
  }

  r->resource_id= rr->resource_id;
  r->limit= rr->limit;
  r->cycle= 0;
  r->max_cycle= rr->max_cycle;
  r->action= rr->action;
  r->operator= rr->operator;
  r->next= current->resourcelist;

  if(r->max_cycle < 1) {

    log("%s: The number of cycles must be greater then 0 in line %d\n",
	prog, lineno-1);
    cfg_errflag++;
    
  }

  current->resourcelist= r;
  reset_resourceset();

}


/*
 * Add a new file object to the current process timestamp list
 */
static void addtimestamp(struct TimestampSet *ts) {

  Timestamp_T t= NEW(t);

  ASSERT(ts);

  t->pathname= ts->pathname;
  t->operator= ts->operator;
  t->time= ts->time;
  t->action= ts->action;

  t->next= current->timestamplist;
  current->timestamplist= t;

  reset_timestampset();

}


/*
 * Adds procfs check data to current process procinfo
 */
static void createprocinfo() {

  ProcInfo_T pi= NEW(pi);

  current->procinfo=pi;
  
}


/*
 * Return a protocol object for the given protocol
 */
static void *addprotocol(int proto) {

  switch (proto) {
    
    case P_HTTP:  return create_http();
    case P_FTP:   return create_ftp();
    case P_SMTP:  return create_smtp();
    case P_POP:   return create_pop();
    case P_IMAP:  return create_imap();
    case P_NNTP:  return create_nntp();
    case P_SSH:   return create_ssh();
    case P_DWP:   return create_dwp();
    case P_LDAP2: return create_ldap2();
    case P_LDAP3: return create_ldap3();
	
  }

  return create_default();
   
}


/*
 * Add the current command object to the current process object
 */
static void addcommand(int what) {

  switch(what) {
    
    case START: current->start= command; break;
    case STOP:  current->stop= command; break;
	
  }

  command= NULL;
  
}

  
/*
 * Add a new argument to the argument array
 */
static void addargument(char *argument) {

  ASSERT(argument);

  if(! command) {
    
    command= NEW(command);

  }
  
  command->arg[command->length++]= argument;
  command->arg[command->length]= NULL;
  
  if(command->length >= ARGMAX) {
    
    log("%s: Exceeded maximum number of program arguments at line %d\n",
	prog, lineno-1);
    cfg_errflag++;
    
  }
  
}


/*
 * Reset the logfile if changed
 */
static void setlogfile(char *logfile) {

  if(Run.logfile) {
    
    if(is(Run.logfile, logfile)) {
      
      return;
      
    } else {
      
      free(Run.logfile);
      
    }
  }
  
  Run.logfile= logfile;
 
}


/*
 * Reset the pidfil if changed
 */
static void setpidfile(char *pidfile) {

  if(Run.pidfile) {
    
    if(is(Run.pidfile, pidfile)) {
      
      return;
      
    } else {
      
      free(Run.pidfile);
      
    }
  }
  
  Run.pidfile= pidfile;
 
}


/*
 * Set mailserver if changed
 */
static void setmailserver(char *mailserver) {

  ASSERT(mailserver);

  if(Run.mailserver) {
    
    if(is(Run.mailserver, mailserver)) {
      
      return;
      
    } else {
      
      free(Run.mailserver);
      
    }
  }
  
  Run.mailserver= mailserver;

}


/*
 * Set Basic Authentication username
 */
static void setuname(char *uname) {

  ASSERT(uname);

  if(Run.Auth.uname) {
    
    if(is(Run.Auth.uname, uname)) {
      
      return;
      
    } else {
      
      free(Run.Auth.uname);
      
    }
  }
  
  Run.Auth.uname= uname;

}

 
/*
 * Set Basic Authentication password
 */
static void setpasswd(char *passwd) {

  ASSERT(passwd);

  if(Run.Auth.passwd) {
    
    if(is(Run.Auth.passwd, passwd)) {
      
      return;
      
    } else {
      
      free(Run.Auth.passwd);
      
    }
  }
  
  Run.Auth.passwd= passwd;

} 


/*
 * Reset the Mail filter to default values
 */
static void reset_mailfilter() {

  mtf.from= NULL;
  mtf.subject= NULL;
  mtf.message= NULL;
  mtf.timeout= FALSE;
  mtf.restart= FALSE;
  mtf.checksum= FALSE;
  mtf.resource= FALSE;
  mtf.stop= FALSE;
  mtf.timestamp= FALSE;


}


/*
 * Reset the Port set to default values
 */
static void reset_portset() {

  portset.socket= -1;
  portset.hostname= NULL;
  portset.port= 0;
  portset.type= SOCK_STREAM;
  portset.family= AF_INET;
  portset.ssl= FALSE;
  portset.sslversion= SSL_VERSION_AUTO;
  portset.request= NULL;
  portset.protocol= NULL;
  portset.pathname= NULL;

}


/*
 * Reset the Proc set to default values
 */
static void reset_resourceset() {

  resourceset.resource_id= 0;
  resourceset.limit= 0;
  resourceset.max_cycle= 1;
  resourceset.action= ACTION_ALERT;
  resourceset.operator= OPERATOR_EQUAL;

}


/*
 * Reset and release the Run mail-format sub-structure
 */
static void reset_runmail() {

  free(Run.MailFormat.from);
  Run.MailFormat.from= NULL;
  free(Run.MailFormat.subject);
  Run.MailFormat.subject= NULL;
  free(Run.MailFormat.message);
  Run.MailFormat.message= NULL;

}


/*
 * Reset the Timestamp set to default values
 */
static void reset_timestampset() {

  timestampset.pathname= NULL;
  timestampset.operator= OPERATOR_EQUAL;
  timestampset.time= 0;
  timestampset.action= ACTION_ALERT;

}


/* ---------------------------------------------------------------- Checkers */


/*
 * Check for unique process name
 */
static void check_name(char *name) {

  ASSERT(name);

  if(exist_process(name)) {
    
    log("%s: Name conflict at line %d - check name '%s' already defined\n",
	prog, lineno, name);
    
    cfg_errflag++;
    
  }
  
}


/*
 * Timeout statement semantic check
 */
static void check_timeout(int s, int c) {

  if(s > c) {
    
    yyerror("The number of restarts must be less than poll cycles");
    
  }

  if(s <= 0 || c <= 0) {
    
    yyerror("Zero or negative values not allowed in a timeout statement");
    
  }
  
}


/*
 * Every statement semantic check
 */
static void check_every(int every) {

  if(every <= 1) {
    
    yyerror("An every statement must have a value greater than 1.");
    
  }
  
}


/*
 * Check hostname 
 */
static void check_hostname(char *hostname) {

  ASSERT(hostname);

  if(!check_host(hostname)) {
    
    yyerror("Hostname did not resolve");
    
  }
 
}


/*
 * Check the dependency graph for errors and if dependencies are
 * present reshuffle the process list in a depending order.
 */
static void check_depend() {

  Process_T p;
  int has_depend= FALSE;
  
  for (p= processlist; p; p= p->next) {
    if(p->visited)
	continue;
    validate_depend(p, &has_depend);
    reset_depend();
  }
  
  if(has_depend) {
    
    Process_T d;
    
    for (p= processlist; p; p= p->next) {
      if(p->visited)
	  continue;
      order_depend(p);
    }

    ASSERT(depend_list);
    processlist= depend_list;
    
    for(d= depend_list; d; d= d->next_depend)
	d->next= d->next_depend;
    
  }

  reset_depend();
  
}



/* -------------------------------------------------------------------- Misc */


/*
 * Create a real email address from a username
 */
static char *append_hostname(char *addr) {

  char buf[STRLEN];
  char * fulladdr;

  ASSERT(addr);

  if (gethostname(buf, STRLEN) != 0) {
     yyerror("Cannot find hostname for mailaddress");
  }

  fulladdr = xresize(addr, strlen(buf)+1);
  strcat(fulladdr, "@");
  strcat(fulladdr, buf);

  return fulladdr;

}


/*
 * Cleans up an md5 string, tolower and remove byte sperators
 */
static void cleanup_md5_string(char *md5string) {

  int i= 0, j= 0;

  ASSERT(md5string);

  while (md5string[i] != '\0') {

    if (isxdigit((int) md5string[i])) {
      
      md5string[j]=tolower(md5string[i]);
      j++;

    } 

    i++;

  }
  
  md5string[j]='\0';

  return;
 
}


/*
 * Search for any errors in the process dependency graph
 */
static void validate_depend(Process_T p, int *has_depend) {

  ASSERT(p);

  if(p->visited)
      return;
  
  if(p->dependantlist) {
    
    Dependant_T d;
    
    for(d= p->dependantlist; d; d= d->next) {
      
      Process_T dp= get_process(d->dependant);
      
      if(!dp) {
	log("%s: Error, depend process '%s' is not defined in the "
	    "control file\n", prog, d->dependant);
	exit(1);
      }
      
      if(dp->depend_visited) {
	log("%s: Error, found a depend loop in the control file involving "
	    "the process '%s'\n", prog, p->name);
	exit(1);
      }
      
      *has_depend= TRUE;
      dp->depend_visited= TRUE;
      validate_depend(dp, has_depend);
      
    }
  }
  
  p->visited= TRUE;

}


/*
 * Order the process list with the most "depending" process last and
 * the least first.
 */
static void order_depend(Process_T p) {

  ASSERT(p);
  
  if(p->visited)
      return;

  p->visited= TRUE;

  if(p->dependantlist) {
    
    Dependant_T d;
    
    for(d= p->dependantlist; d; d= d->next) {
      
      Process_T dp= get_process(d->dependant);
      
      order_depend(dp);
      
    }
  }

  p->next_depend= depend_list;
  depend_list= p;

}
