/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include <config.h>

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>
#include <signal.h>
#include <stdarg.h>
#include <errno.h>
#include <setjmp.h>
#include <stdlib.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif


#include "monitor.h"
#include "net.h"

/* Private variables */
static int  mysocket;
static char *server;
static sigjmp_buf die;

/* Private prototypes */
static void do_status();
static void do_timeout(int);
static void finalize_server();
static int  initialize_server();
static void do_send(const char *, ...);

/**
 *  Connect to a SMTP server and send mail.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *
 *  @version \$Id: sendmail.c,v 1.11 2003/02/11 21:27:39 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Send SMTP mail messages
 * @param mail A mymail struct object
 */
void sendmail(Mail_T mail) {

  Mail_T m;
  char *now = get_RFC1123date(NULL);
  char *localhost= get_localhostname();

  if(!initialize_server()) {
    
    goto exit;

  }
  
  if(sigsetjmp(die, TRUE)) {

    goto exit;

  }

  do_status();
  do_send("HELO %s\r\n", localhost); 
  do_status();

  for(m= mail; m; m= m->next) { 
    
    do_send("MAIL FROM: <%s>\r\n", m->from);
    do_status();
    do_send("RCPT TO: <%s>\r\n", m->to);
    do_status();
    do_send("DATA\r\n");
    do_status();
    do_send("From: %s\r\n", m->from);
    do_send("To: %s\r\n", m->to);
    do_send("Subject: %s\r\n", m->subject);
    do_send("Date: %s\r\n", now);
    do_send("X-Mailer: %s %s\r\n", prog, VERSION);
    do_send("Mime-Version: 1.0\r\n");
    do_send("Content-Type: text/plain; charset=\"iso-8859-1\"\r\n");
    do_send("Content-Transfer-Encoding: quoted-printable\r\n");
    do_send("\r\n");
    do_send("%s\r\n", m->message);
    if(m->opt_message)
	do_send("%s\r\n", m->opt_message);
    do_send(".\r\n");
    do_status();

    alarm(SMTP_TIMEOUT);

  }

  do_send("QUIT\r\n");
  
  exit:
  free(now);
  free(localhost);
  finalize_server();
  
}


/* ----------------------------------------------------------------- Private */


static int initialize_server() {
  
  struct sigaction sa;
  
  if((server= Run.mailserver) == NULL) {
    
    server= LOCALHOST;
    
  }
  
  sa.sa_handler= do_timeout;
  sa.sa_flags= 0;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGALRM, &sa, NULL);
  alarm(SMTP_TIMEOUT);

  if((mysocket= create_socket(server, SMTP_PORT, SOCK_STREAM)) <0) {

    log("Cannot open a connection to the mailserver '%s' -- %s\n",
	server, STRERROR);
    
    return FALSE;
    
  }
  
  return TRUE;
  
}


static void finalize_server() {
  
  alarm(0);
  
  if(mysocket && !close_socket(mysocket)) {
    
    log("Error shutting down socket connection to mailserver %s -- %s\n",
	server, STRERROR);
    
  }
  
}


static void do_timeout(int sig) {
  
  log("Sendmail: Connection to mailserver timed out\n");
  siglongjmp(die, TRUE);
  
}


void do_send(const char *format, ...) {
  
  char msg[2048];
  va_list ap;
  
  va_start(ap,format);
  vsnprintf(msg, 2048, format, ap);
  va_end(ap);
  
  if(sock_send(mysocket, msg, strlen(msg), 0) <= 0) {
    
    log("Sendmail: error sending data to the server '%s' -- %s\n",
	  server, STRERROR);
    siglongjmp(die, TRUE);
    
  }
  
}


static void do_status() {
  
  int  status;
  char buf[STRLEN];
  
  if(sock_recv(mysocket, buf, sizeof(buf), 0) <= 0) {
    
    log("Sendmail: error receiving data from the mailserver '%s' -- %s\n",
	server, STRERROR);
    siglongjmp(die, TRUE);
    
  }
  
  chomp(buf);
  
  sscanf(buf, "%d", &status);
  if(status >= 400) {
    
    log("Sendmail error: %s\n", buf);
    siglongjmp(die, TRUE);
    
  }
  
}


