/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "alert.h"
#include "monitor.h"
#include "engine.h"

/* Private prototypes */
static void do_free();
static Command_T copy_command(Command_T);
static Mail_T create_exec_alert_list(Process_T, Command_T);


/**
 *  Function for spawning of a process. This function fork's twice to
 *  avoid creating any zombie processes. Inspired by code from
 *  W. Richard Stevens book, APUE.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *
 *  @version \$Id: spawn.c,v 1.17 2003/02/11 21:27:39 martinp Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Execute the given command. If the execution fails, an alert message
 * is sent to the email addresses found in the process object that has
 * registred interest for restart alerts.
 * @param P A Process object
 * @param C A Command object
 */
void  spawn(Process_T P, Command_T C) {
  
  pid_t pid;
  sigset_t mask, save;

  ASSERT(P);
  ASSERT(C);
  
  /*
   * Block SIGCHLD
   */
  sigemptyset(&mask);
  sigaddset(&mask, SIGCHLD);
  pthread_sigmask(SIG_BLOCK, &mask, &save);
  
  if((pid= fork()) < 0) {
    
    log("Cannot fork of a new process\n");  
    exit(1);
    
  }  else if(pid == 0) {
    
    if((pid= fork()) < 0) {
      
      log("Cannot fork of a new process\n");  
      exit(1);
      
    }
    
    else if(pid > 0) {
      
      _exit(0);
      
    } else  {

      Command_T c= copy_command(C);
      Mail_T alert_list= create_exec_alert_list(P, C);
      
      /*
       * Reset all signals, so the child process is
       * *not* created with any inherited SIG_BLOCK
       */
      sigemptyset(&mask);
      pthread_sigmask(SIG_SETMASK, &mask, NULL);
      
      /* Unregister Signal handlers */
      signal(SIGTERM, SIG_DFL);
      signal(SIGUSR1, SIG_DFL);
      signal(SIGPIPE, SIG_DFL);

      /* Garbage collect */
      do_free();

      /*
       * Reset to the original umask so programs will inherit the
       * same file creation mask monit was started with
       */
      umask(Run.umask);

      /*
       * Redirect the standard file descriptors
       */
      redirect_stdfd();
      
      /*
       * The exec statement done by the second child
       */
      (void) execv(c->arg[0], c->arg);
      sendmail(alert_list);
      log("Could not execute %s\n", c->arg[0]);
      gc_mail_list(&alert_list);
      _exit(1);
      
    }
  }

  /* Wait for first child - aka second parent, to exit */
  if(waitpid(pid, NULL, 0) != pid) {
    
      log("Waitpid error\n");

  }

  /*
   * Restore the signal mask
   */
  pthread_sigmask(SIG_SETMASK, &save, NULL);
  
  /*
   * We do not need to wait for the second child since we forked twice,
   * the init system-process will wait for it. So we just return
   */
  return ;
  
} 


/* ----------------------------------------------------------------- Private */


/*
 * Free memory for the child process
 */
static void do_free() {
  
  gc();
  destroy_hosts_allow();
  free(Run.controlfile);
  free(Run.pidfile);
  free(Run.logfile);
  free(Run.localhostname);
  free(Run.Env.user);
  free(Run.Env.home);
  free(Run.Env.cwd);
  free(Run.Auth.uname);
  free(Run.Auth.passwd);
  free(Run.MailFormat.from);
  free(Run.MailFormat.subject);
  free(Run.MailFormat.message);
  free(Run.bind_addr);
  /* Run.mailserver is not freed since it's used in exec alert */

}


/*
 * Return a deep copy of the given command object
 */
static Command_T copy_command(Command_T C) {

  int i;
  Command_T P= NEW(P);

  for(i= 0; C->arg[i]; i++)
      P->arg[i]= xstrdup(C->arg[i]);
  P->arg[i]= NULL;

  return P;
  
}


/*
 * Create a list of mail addresses to send mail to in case the exec
 * fails. Only mail objects that has registred interest for restart
 * notification is used.
 */
static Mail_T create_exec_alert_list(Process_T p, Command_T c) {

  Mail_T l= NULL;
  Mail_T h= p->maillist;

  for(; h; h= h->next)
      if(h->alert_on_restart) {
	
	Mail_T n= NEW(n);
	char *message= xmalloc(STRLEN);
	
	n->to= xstrdup(h->to);
	n->from= xstrdup("monit@localhost");
	n->subject= xstrdup("monit alert -- exec error");
	snprintf(message, STRLEN,
	 "NOTE, THIS EMAIL TAKES PRECEDENCE OVER SUBSEQUENT RESTART ALERTS\n\n"
		 "Could not execute program '%s' for process %s\n\n"
		 "Your faithful employee,\nmonit",
		   c->arg[0]?c->arg[0]:"null", p->name);
	n->message= message;
	n->next= l;
	l= n;
	
      }

  return l;
  
}
