/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "net.h"
#include "monitor.h"
#include "monit_process.h"
#include "ssl.h"


/* Private Prototypes */
static void local_status(Process_T);
static int remote_status(Process_T, ssl_connection *);

/**
 *  Print the status of processes in the process list.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: status.c,v 1.24 2003/02/13 08:51:13 chopp Exp $
 *
 *  @file
 */

/* ------------------------------------------------------------------ Public */


/**
 * Show all processes in the process list
 */
void status() {

  Process_T p;
  char *uptime= get_process_uptime(Run.pidfile);
  int remote= TRUE;
  ssl_connection * ssl= NULL;

  if(Run.httpdssl) {

    ssl = new_ssl_connection(Run.httpsslpem, SSL_VERSION_AUTO);

  }

  fprintf(stdout, "monit daemon uptime: %s\n", uptime);
  free(uptime);
  
  for(p= processlist; p; p= p->next) {

      if(remote) {

	remote = remote_status(p, ssl);

      } else {

	local_status(p);

      }
    
  }

  if(Run.httpdssl) {

    delete_ssl_socket(ssl);

  }
  
}

/**
 * Show all processes in the group
 * @param G  group name
 */
void status_group(char *G) {

  Process_T p;
  int remote= TRUE;
  ssl_connection * ssl= NULL;

  ASSERT(G);
  
  if(Run.httpdssl) {

    ssl = new_ssl_connection(Run.httpsslclientpem, SSL_VERSION_AUTO);

  }

  for(p= processlist; p; p= p->next) {

    if(is(p->group, G)) {

      if(remote) {

	remote = remote_status(p, ssl);

      } else {

	local_status(p);

      }

    }
   
  }

  if(Run.httpdssl) {

    delete_ssl_socket(ssl);

  }
}

/* ----------------------------------------------------------------- Private */


/**
 * Print the process status for the given process
 * @param p  A Process_T object
 */
static void local_status(Process_T p) {
  
  pid_t  pid= -1;

  ASSERT(p);
  
  if((pid= is_process_running(p))) {

    char *uptime= get_process_uptime(p->pidfile);

    fprintf(stdout, "Process '%s' is running with pid [%d] Uptime: %s "
            "Monitoring status: %s\n",
            p->name, (int)pid, uptime, statusnames[p->do_validate]);
    
    free(uptime);
 
  } else {
    
    fprintf(stdout, "Process '%s' is not running\n",  p->name);
    
  }
  
}


/**
 * Get the process status for the given process from a monit server
 * @param p  A Process_T object
 * @return TRUE if the monit server could be connected 
 */
static int remote_status(Process_T p, ssl_connection * ssl) {

  ASSERT(p);
  
  if(exist_daemon()) {
    
    /* If a monit daemon exist we request status information from the server */
    
    int s= create_socket(Run.bind_addr?Run.bind_addr:"localhost",
			 Run.httpdport, SOCK_STREAM);
    if(s<0) {

      fprintf(stdout,
	      "Cannot connect to monit server to get"
	      " extended process data.\n");
      local_status(p);

      return FALSE;
            
    } else {

      int n;
      char req[2*STRLEN];
      char *auth= get_basic_authentication_header();
      char buf[STRLEN];

      snprintf(req, 2*STRLEN,
	       "GET /%s?action=status HTTP/1.0\r\n%s\r\n", p->name, auth);
      
      free(auth);

      if(ssl != NULL) {

	if(!embed_ssl_socket(ssl, s)) {

	  fprintf(stdout, "Failed to establish SSL communication to monit"
		  " server\n");
	  local_status(p);
      
	  return FALSE;
	}

	send_ssl_socket(ssl, req, sizeof(req));
	if(0>(n= recv_ssl_socket(ssl, buf, STRLEN))) {
	  
	  local_status(p);
	  close_ssl_socket(ssl);
	  close_socket(s);
	  
	  return TRUE;
	
	}

	close_ssl_socket(ssl);
	close_socket(s);

      } else {

	sock_send(s, req, sizeof(req), 0);
      
	if(0>(n= sock_recv(s, buf, STRLEN, 0))) {
	  
	  local_status(p);
	  close_socket(s);
	  
	  return TRUE;
	
	}
	
	close_socket(s);
      }
      

      /* If everything has gone well the returned string starts with
	 "Process " */

      buf[n]= 0;
      if(starts_with(buf, "Process ")) {

	fprintf(stdout, "%s", buf);
	
      } else {

	fprintf(stdout, "The monit server did not return a process record.\n");
	local_status(p);
    
	return TRUE;

      }

      return TRUE;
    }
    
  } else {
    
    /* No monit daemon exist, just print local status information */
    
    fprintf(stdout, "Cannot connect to the monit server to get extended "
	    "process data.\n");
    local_status(p);
    
    return FALSE;
  }

}
