/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <fcntl.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "monitor.h"
#include "engine.h"
#include "md5.h"
#include "base64.h"
#include "alert.h"
#include "monit_process.h"

/* Private prototypes */
static int is_unsafe(unsigned char *c);
static char *is_str_defined(char *);
static char *is_str_defined_default(char *);

/**
 *  General purpose utility methods.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author "Martin Pala" <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *  @author Michael Amster, <mamster@webeasy.com> 
 *
 *  @version \$Id: util.c,v 1.55 2003/02/14 08:22:33 martinp Exp $
 * 
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Print a formated message to stderr
 * or to the logfile if no tty exist
 * @param format A formate (printf-style) string
 */
void error(const char *format, ...) {
  
  char msg[1024];
  va_list ap;

  ASSERT(format);
  
  va_start(ap,format);
  vsnprintf(msg, 1024, format, ap);
  va_end(ap);
  
  if(Run.have_tty) {
    
    fprintf(stderr,"%s", msg);
    fflush(stderr);
    
  } else {
    
    log("%s", msg);
    
  }
  
}


/**
 * @return TRUE if the string parameter is defined, otherwise FALSE
 */
int is_strdefined(char *p) {
  
  return(p && *p);
  
}


/**
 * Strip the path and return only the filename
 * @param path A file path string
 * @return the basename
 */
char *stripfilename(char* path) {
  
  char *fname;

  ASSERT(path);

  fname= strrchr(path, '/');
  
  return(fname ? ++fname : path);
  
}


/**
 * Removes everything from the first newline (CR|LF)
 * @param string A string to be chomped
 */
void chomp(char *string) {
  
  char *p;

  ASSERT(string);

  if((p= strchr(string, '\r')) || (p= strchr(string, '\n'))) *p= 0;
  
}


/**
 * Remove leading and trailing space from the string
 * @param s A string
 * @return s with leading and trailing spaces removed
 */
char *trim(char *s) {

  ASSERT(s);
  
  ltrim(s);
  rtrim(s);

  return s;
  
}


/**
 * Remove leading white space [ \t\r\n] from the string. 
 * @param s A string
 * @return s with leading spaces removed
 */
char *ltrim(char *s) {

  char *t= s;

  ASSERT(s);

  while(*t==' ' || *t=='\t' || *t=='\r' || *t=='\n') t++;

  return strcpy(s, t);

}


/**
 * Remove trailing white space [ \t\r\n] from the string
 * @param s A string
 * @return s with trailing spaces removed
 */
char *rtrim(char *s) {
  
  char *t= s;

  ASSERT(s);

  while(*s) s++; 
  while(*--s==' ' || *s=='\t' || *s=='\r' || *s=='\n') *s= '\0';

  return t;
  
}


/**
 * Remove any enclosing quotes ["'] from the string
 * @param s A string
 * @return s with any enclosed quotes removed
 */

char *trim_quotes(char *s) {

  char *t= s;
  char tmp=0;

  ASSERT(s);

  while(*t==39 || *t==34) {

    tmp=*t;
    t++;
    break;

  }  

  if(t[strlen(t)-1]==tmp) t[strlen(t)-1]= '\0';

  return strcpy(s, t);

}


/**
 * Return TRUE if the string <i>a</i> starts with the string <i>b</i>.
 * @param a The string to search for b in
 * @param b The sub-string to test a against
 * @return TRUE if a starts with b, otherwise FALSE
 */
int starts_with(char *a, char *b) {

  if((!a || !b) || *a!=*b) return FALSE;

  while(*a && *b) {
    
    if(*a++ != *b++) return FALSE;
    
  }

  return TRUE;

}


/**
 * Exchanges \escape sequences in a string
 * @param buf A string
 */

void handle_string_escapes(char *buf) {

  int editpos;
  int insertpos;

  ASSERT(buf);

  for(editpos=insertpos=0; *(buf+editpos)!='\0'; editpos++, insertpos++) {

    if(*(buf+editpos) == '\\' ) {
      
      switch(*(buf+editpos+1)) {

      case 'n': 
        *(buf+insertpos)='\n';
        editpos++;
	break;

      case 't':
        *(buf+insertpos)='\t';
        editpos++;
	break;

      case 'r':
        *(buf+insertpos)='\r';
        editpos++;
	break;

      case ' ':
        *(buf+insertpos)=' ';
        editpos++;
	break;

      case '\\':
        *(buf+insertpos)='\\';
        editpos++;
	break;

      default:
        *(buf+insertpos)=*(buf+editpos);

      }  

    } else {

      *(buf+insertpos)=*(buf+editpos);

    }  

  }
  *(buf+insertpos)='\0';

}


/**
 * @param name A process name as stated in the config file
 * @return the named process or NULL if not found
 */
Process_T get_process(char *name) {

  Process_T p;

  ASSERT(name);

  for(p= processlist; p; p= p->next) {
    if(is(p->name, name)) {
      return p;
    }
  }

  return NULL;

}


/**
 * @param name A process name as stated in the config file
 * @return TRUE if the process name exist in the
 * processlist, otherwise FALSE
 */
int exist_process(char *name) {

  Process_T p;

  ASSERT(name);

  for(p= processlist; p; p= p->next)
      if(is(p->name, name))
	  return TRUE;

  return FALSE;

}


/**
 * Print the Runtime object
 */
void printrunlist() {
  
  printf("Runtime constants:\n");
  printf(" %-18s = %s\n", "Control file", is_str_defined(Run.controlfile));
  printf(" %-18s = %s\n", "Log file", is_str_defined(Run.logfile));
  printf(" %-18s = %s\n", "Pid file", is_str_defined(Run.pidfile));
  printf(" %-18s = %s\n", "Debug", Run.debug?"True":"False");
  printf(" %-18s = %s\n", "Log", Run.dolog?"True":"False");
  printf(" %-18s = %s\n", "Use syslog", Run.use_syslog?"True":"False");
  printf(" %-18s = %s\n", "Is Daemon", Run.isdaemon?"True":"False");
  printf(" %-18s = %s\n", "Use process engine", Run.doprocess?"True":"False");
  printf(" %-18s = %d seconds\n", "Poll time", Run.polltime);
  printf(" %-18s = %s\n", "Mail server", is_str_defined(Run.mailserver));
  printf(" %-18s = %s\n", "Mail from", is_str_defined(Run.MailFormat.from));
  printf(" %-18s = %s\n", "Mail subject",
	 is_str_defined(Run.MailFormat.subject));
  printf(" %-18s = %-.20s%s\n", "Mail message",
	 Run.MailFormat.message?
	 Run.MailFormat.message:"(not defined)",
	 Run.MailFormat.message?"..(truncated)":"");
	 
  printf(" %-18s = %s\n", "Start monit httpd", Run.dohttpd?"True":"False");


  
  if(Run.dohttpd) {
    
    printf(" %-18s = %s\n", "httpd bind address",
	   Run.bind_addr?Run.bind_addr:"Any/All");
    printf(" %-18s = %d\n", "httpd portnumber", Run.httpdport);
    printf(" %-18s = %s\n", "Use ssl encryption", Run.httpdssl?"True":"False");

    if(Run.httpdssl) {

      printf(" %-18s = %s\n", "PEM key/cert file", Run.httpsslpem);

      if(Run.httpsslclientpem!=NULL) {
	printf(" %-18s = %s\n", "Client cert file", Run.httpsslclientpem);
      } else {
	printf(" %-18s = %s\n", "Client cert file", "None");
      } 

      printf(" %-18s = %s\n", "Allow self certs", 
	     Run.allowselfcert?"True":"False");

    }

    printf(" %-18s = %s\n", "httpd auth. style",
	   Run.Auth.defined&&has_hosts_allow()?
	   "Basic Authentication and Host allow list":
	   Run.Auth.defined?"Basic Authentication":
	   has_hosts_allow()?"Host allow list":
	   "No authentication!");
     
  }

  printf("\n");
  
}


/**
 * Print a process object
 * @param p A Process_T object
 */
void printprocess(Process_T p) {
  
  Port_T n;
  Mail_T r;
  Resource_T q;
  Checksum_T c;
  Timestamp_T t;
  Dependant_T d;

  ASSERT(p);
 
  printf("%-21s = %s\n", "Process Name", p->name);
  printf(" %-20s = %s\n", "Group", is_str_defined(p->group));
  printf(" %-20s = %s\n", "Pid file", p->pidfile);
  printf(" %-20s = %s\n", "Monitoring mode", modenames[p->mode]);
  if(p->start)
    printf(" %-20s = %s\n", "Start program", is_str_defined(p->start->arg[0]));
  if(p->stop)
    printf(" %-20s = %s\n", "Stop program", is_str_defined(p->stop->arg[0]));

  for(c= p->checksumlist; c; c= c->next) {
    
    printf(" %-20s = %s %s\n", "Checksum", c->md5, c->file);
    
  }

  for(d= p->dependantlist; d; d= d->next)
    if(d->dependant != NULL)
      printf(" %-20s = %s\n", "Depends on Process", d->dependant);

  if(! p->portlist) {
    
    printf(" %-20s = (not defined)\n", "Host:Port");
    
  } else {
  
    for(n= p->portlist; n; n= n->next) {
    
      if(n->family == AF_INET) {

	if(n->ssl != NULL) {

	  printf(" %-20s = %s:%d%s [protocol %s via SSL]\n", "Host:Port",
		 n->hostname, n->port, n->request?n->request:"",
		 n->protocol->name);

	  if(n->certmd5 != NULL) {

	    printf(" %-20s = %s\n", 
		   "Server cert md5 sum",
		   n->certmd5);

	  }

	} else {

	  printf(" %-20s = %s:%d%s [protocol %s]\n", "Host:Port",
		 n->hostname, n->port, n->request?n->request:"",
		 n->protocol->name);

	}

      } else if(n->family == AF_UNIX) {

        printf(" %-20s = %s [protocol %s]\n", "Unix Socket",
               n->pathname, n->protocol->name);

      }
    
    }

  }
  
  for(t= p->timestamplist; t; t= t->next) {
    
    printf(" %-20s = if %s %s %d second(s) then %s\n",
           "Timestamp",
           t->pathname,
           operatornames[t->operator],
           t->time,
           actionnames[t->action]);
    
  }

  if(! p->resourcelist) {

    printf(" %-20s = (not defined)\n", "Resource Limits");

  } 
    
  for(q= p->resourcelist; q; q= q->next) {

    switch(q->resource_id) {

    case RESOURCE_ID_CPU_PERCENT: 

      printf(" %-20s = if %s %.1f%% for %d cycle(s) then %s\n", 
	     "CPU usage limit", 
	     operatornames[q->operator], 
	     q->limit/10.0, q->max_cycle, actionnames[q->action]);
      break;

    case RESOURCE_ID_MEM_PERCENT: 

      printf(" %-20s = if %s %.1f%% for %d cycle(s) then %s\n", 
	     "Memory usage limit", 
	     operatornames[q->operator], q->limit/10.0, q->max_cycle, 
	     actionnames[q->action]);
      break;

    case RESOURCE_ID_MEM_KBYTE: 

      printf(" %-20s = if %s %ldkB for %d cycle(s) then %s\n", 
	     "Memory amount limit", 
	     operatornames[q->operator], q->limit, q->max_cycle, 
	     actionnames[q->action]);
      break;

    case RESOURCE_ID_LOAD1: 

      printf(" %-20s = if %s %.1f for %d cycle(s) then %s\n", 
	     "Load avg. (1min)", 
	     operatornames[q->operator], q->limit/10.0, q->max_cycle, 
	     actionnames[q->action]);
      break;

    case RESOURCE_ID_LOAD5: 

      printf(" %-20s = if %s %.1f for %d cycle(s) then %s\n", 
	     "Load avg. (5min)", 
	     operatornames[q->operator], q->limit/10.0, q->max_cycle, 
	     actionnames[q->action]);
      break;

    case RESOURCE_ID_LOAD15: 

      printf(" %-20s = if %s %.1f for %d cycle(s) then %s\n", 
	     "Load avg. (15min)", 
	     operatornames[q->operator], q->limit/10.0, q->max_cycle, 
	     actionnames[q->action]);
      break;

    }    
  }

  if(p->def_every)
      
    printf(" %-20s = Check process every %d cycles\n", "Every", p->every);
  
  else {
    
    printf(" %-20s = (not defined)\n", "Every");
    
  }

  if(p->def_timeout) {
    
    printf(" %-20s = Do timeout if %d restart within %d cycles\n",
	   "Timeout", p->to_start, p->to_cycle);
    
  } else {
    
    printf(" %-20s = (not defined)\n", "Timeout");
    
  }

  for(r= p->maillist; r; r= r->next) {
    
    printf(" %-20s = %s\n", "Alert mail to", is_str_defined(r->to));
    printf("   %-18s = %s\n", "alert from", is_str_defined_default(r->from));
    printf("   %-18s = %s\n", "alert subject",
	   is_str_defined_default(r->subject));
    printf("   %-18s = %-.20s%s\n", "alert message",
	   is_str_defined_default(r->message),
	   r->message?"..":"");
    printf("   %-18s = %s\n", "alert on timeout",
	   r->alert_on_timeout?"yes":"no");
    printf("   %-18s = %s\n", "alert on restart",
	     r->alert_on_restart?"yes":"no");
    printf("   %-18s = %s\n", "alert on checksum",
	   r->alert_on_checksum?"yes":"no");
    printf("   %-18s = %s\n", "alert on resource",
	   r->alert_on_resource?"yes":"no");
    printf("   %-18s = %s\n", "alert on stop",
	   r->alert_on_stop?"yes":"no");
    printf("   %-18s = %s\n", "alert on timestamp",
	   r->alert_on_timestamp?"yes":"no");

  }

  printf("\n");
  
}


/**
 * Print all the processes in the processlist
 */
void printprocesslist() {

  Process_T p;
  char ruler[STRLEN];
  
  printf("The process list contains the following entries:\n\n");
  
  for(p= processlist; p; p= p->next) {
    
    printprocess(p);
    
  }

  memset(ruler, '-', STRLEN);
  printf("%-.79s\n", ruler);
  
}


/**
 * Open and read the pid from the given pidfile.
 * @param pidfile A pidfile with full path
 * @return the pid (TRUE) or FALSE if the pid could
 * not be read from the file
 */
pid_t get_pid(char *pidfile) {
  
  FILE *file= NULL;
  int pid= -1;

  ASSERT(pidfile);

  if(! exist_file(pidfile)) {
    
    return(FALSE);
    
  }

  if(! isreg_file(pidfile)) {
    
    log("%s: pidfile '%s' is not a regular file\n",prog, pidfile);
    return(FALSE);
    
  }
  
  if((file= fopen(pidfile,"r")) == (FILE *)NULL) {
    
    log("%s: Error opening the pidfile '%s' -- %s\n",
	prog, pidfile, STRERROR);
    return(FALSE);
    
  }

  fscanf(file, "%d", &pid);
  fclose(file);

  if(pid == -1) {
    
    log("%s: pidfile `%s' does not contain a valid pidnumber\n",
	prog, pidfile);
    
    return (FALSE);
    
  }

  return (pid_t)pid;
  
}


/**
 * @return TRUE (i.e. the running pid id)  if
 * the process is running, otherwise FALSE
 */
int is_process_running(Process_T p) {
  
  pid_t  pid;
  int kill_return= 0;

  ASSERT(p);
  
  errno= 0;

  if((pid= get_pid(p->pidfile))) {
    
    if((kill_return= getpgid(pid)) > 0 || errno == EPERM)
	
      return pid;
    
  }

  memset(p->procinfo, 0, sizeof *(p->procinfo));

  return FALSE;
  
}


/**
 * Returns a (RFC1123) Date string. If the given date
 * is NULL compute the date now.
 * @param date
 * @return a date string alligned with RFC1123
 */
char *get_RFC1123date(long *date) {

  struct tm *tm_now;
  char D[STRLEN];
  time_t now= (date && (*date>0))?*date:time(&now);
  long timezone_h;
  long timezone_m;
  int datelen;

  time(&now);
  tzset();
  tm_now = localtime(&now);

#if HAVE_STRUCT_TM_TM_GMTOFF
  timezone_h = tm_now->tm_gmtoff/3600;
  timezone_m = abs(tm_now->tm_gmtoff/60)%60;
#else
  timezone_h = -(timezone/3600)+(tm_now->tm_isdst>0);
  timezone_m = abs(timezone/60)%60;
#endif

  datelen=strftime(D, STRLEN-7, "%a, %d %b %Y %H:%M:%S", tm_now);

  if(! datelen) {

    memset(D, 0, STRLEN);
               
  } else {
   
    snprintf(D+datelen, 7, " %+03ld%02ld", timezone_h, timezone_m);

  }

  return xstrdup(D);
  
}


/**
 * Get a non \n terminated ctime
 * @return a ctime string (i.e. date)
 */
char *get_ctime() {
  
  time_t now;
  char *buf = (char*)xmalloc(STRLEN);
  
  time(&now);
  snprintf(buf, STRLEN, "%s", ctime(&now));

  chomp(buf);
  
  return buf;
    
}

/**
 * Compute an uptime string for a process based on the ctime
 * from the pidfile. The caller must free the returned string.
 * @param pidfile A process pidfile
 * @return an uptime string
 */
char *get_process_uptime(char *pidfile) {

  time_t ctime;

  ASSERT(pidfile);

  if( (ctime= get_timestamp(pidfile, S_IFREG)) ) {
    
    time_t now= time(&now);
    time_t since= now-ctime;

    return get_uptime(since);
    
  }

  return xstrdup("");
  
}

  
/**
 * Compute an uptime string based on the delta time in seconds. The
 * caller must free the returned string.
 * @param delta seconds. 
 * @return an uptime string
 */
char *get_uptime(time_t delta) {

  static int min= 60;
  static int hour= 3600;
  static int day= 86400;
  long rest_d;
  long rest_h;
  long rest_m;
  char buf[STRLEN];
  char *p= buf;

  *buf= 0;
  
  if((rest_d= delta/day)>0) {
    p+= sprintf(p, "%ldd ", rest_d);
    delta-= rest_d*day;
  }
  if((rest_h= delta/hour)>0 || (rest_d > 0)) {
    p+= sprintf(p, "%ldh ", rest_h);
    delta-= rest_h*hour;
  }
  
  rest_m= delta/min;
  p+= sprintf(p, "%ldm ", rest_m);
  delta-= rest_m*min;

  return xstrdup(buf);

}


/**
 * Compute a md5 checksum for the given file and save the result
 * in an allocated area pointed to by dest. The caller is responsible
 * for freeing dest.
 * @param dest a pointer to an allocated area with the computed md5 sum
 * @param file The name of file to compute a md5 sum for
 * @return TRUE if success otherwise FALSE.
 */
int set_md5sum(char **dest, char *file) {

  ASSERT(dest);
  ASSERT(file);

  if(! (*dest= get_md5sum(file)))
    return FALSE;
  
  return TRUE;

}
    

/**
 * @return a md5 checksum for the given file, or NULL if error.
 */
char *get_md5sum(char *file) {

  ASSERT(file);

  if(isreg_file(file)) {
    
    FILE *f= fopen(file, "r");
    
    if(f) {
      
      int i;
      unsigned char md5buf[16];
      char result[STRLEN];
      char *r= result;
      
      if(md5_stream(f, md5buf)) {
	
	fclose(f);
	
	return NULL;
	
      }
      
      fclose(f);
      
      for(i= 0; i < 16; ++i) {
	
	r+= sprintf(r, "%02x", md5buf[i]);
	
      }
    
      return (xstrdup(result));
      
    }
    
  }
  
  return NULL;

}


/**
 * @param file A file to open and compute a md5 checksum for
 * @param sum A previous checksum computed for this file
 * @return TRUE if computation of md5sum on the given file, is equal
 * to the given sum, otherwise return FALSE. FALSE is also returned if
 * md5_stream fails.
 */
int check_md5(char *file, char *sum) {

  char *newSum;

  ASSERT(file);
  ASSERT(sum);

  newSum= get_md5sum(file);

  if(newSum) {
    
    int rv;
    
    rv= (!strncmp(sum, newSum, 31));
    free(newSum);
    
    return (rv);
    
  }

  return FALSE;
  
}


/**
 * Escape an uri string converting unsafe characters to a hex (%xx)
 * representation.  The caller must free the returned string.
 * @param uri an uri string
 * @return the escaped string
 */
char *url_encode(char *uri) {

  static unsigned char hexchars[]= "0123456789ABCDEF";
  register int x, y;
  unsigned char *str;

  ASSERT(uri);

  str= (unsigned char *)xmalloc(3 * strlen(uri) + 1);

  for(x = 0, y = 0; uri[x]; x++, y++) {

    if(is_unsafe(&uri[x])) {
      str[y++] = '%';
      str[y++] = hexchars[(unsigned char) uri[x] >> 4];
      str[y] = hexchars[(unsigned char) uri[x] & 0xf];
    } else str[y]= (unsigned char)uri[x];

  }

  str[y] = '\0';

  return ((char *) str);

}


/**
 * @return a Basic Authentication Authorization string (RFC 2617),
 * with credentials from the Run object, The string "\r\n" if
 * credentials are not defined.
 */
char *get_basic_authentication_header() {

  if(Run.Auth.defined) {
    
    char *b64;
    char buf[STRLEN];
    char *auth= xmalloc(STRLEN+1);
    
    snprintf(buf, STRLEN, "%s:%s", Run.Auth.uname, Run.Auth.passwd);
    encode_base64(&b64, strlen(buf), buf);
    snprintf(auth, STRLEN, "Authorization: Basic %s\r\n", b64);
    free(b64);
    
    return auth;

  }
      
  return xstrdup("\r\n");

}


/**
 * Do printf style format line parsing
 * @param s format string
 * @param ap variable argument list
 * @return buffer with parsed string
 */
char *format(const char *s, va_list ap) {

  int n;
  int size= STRLEN;
  char *buf= xmalloc(size);

  ASSERT(s);
    
  while(TRUE) {
      
    n= vsnprintf(buf, size, s, ap);
    
    if(n > -1 && n < size)
        break;
    
    if(n > -1)
        size= n+1;
    else
        size*= 2;
    
    buf= xresize(buf, size);
    
  }
  
  return buf;

}


/**
 * Redirect the standard file descriptors to /dev/null and route any
 * error messages to the log file.
 */
void redirect_stdfd() {

  int i;
  
  Run.have_tty= FALSE;
  for(i= 0; i < 3; i++)
      if(close(i) == -1 || open("/dev/null", O_RDWR) != i)
      error("Cannot reopen standard file descriptor (%d) -- %s\n", i, STRERROR);
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Returns the value of the variable if defined or the String
 * "(not defined)"
 */
static char *is_str_defined(char *var) {
  
  return (is_strdefined(var)?var:"(not defined)");
  
}


/**
 * Returns the value of the variable if defined or the String
 * "(default)"
 */
static char *is_str_defined_default(char *var) {

  return (is_strdefined(var)?var:"(default)");

}


/**
 * Returns TRUE if the given char is url unsafe
 * @param c A unsigned char
 * @return TRUE if the char is in the set of unsafe URL Characters
 */
static int is_unsafe(unsigned char *c) {

  int i;
  static unsigned char unsafe[]= "<>\"#{}|\\^~[]`";

  ASSERT(c);
  
  if(33>*c || *c>176)
      return TRUE;
  
  if(*c=='%') {
    if( isxdigit(*(c + 1)) && isxdigit(*(c + 2)) ) return FALSE;
    return TRUE;
  }

  for(i=0; unsafe[i]; i++)
    if(*c==unsafe[i]) return TRUE;

  return FALSE;

}

