/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <setjmp.h>
#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "monitor.h"
#include "alert.h"
#include "net.h"
#include "monit_process.h"

/* Private variables */
static sigjmp_buf timeout;

/* Private Prototypes */
static void do_validate(Process_T);
static void do_restart(Process_T, char *, ...) ;
static void do_start(Process_T, char *, ...) ;
static void do_stop(Process_T, char *, ...);
static int  do_not_validate(Process_T);
static int  check_connection(Process_T, Port_T, char *);
static int  check_resources(Process_T, Resource_T, char *);
static int  check_process_state(Process_T, char *);
static int  check_skip(Process_T);
static int  check_timeout(Process_T);
static int  check_checksum(Process_T);
static int  checksum_helper(Process_T, char *, char *);
static int  check_timestamp(Process_T, Timestamp_T, char *);
static void connection_timeout(int);
static void reset_resource_counter(Process_T);
static void vlog(char * report, int n, Process_T p, char *m,...);
static int  compare_value(int, int, int);

/**
 *  This function contains the main process and check machinery for
 *  monit. The validate function check processes in the process list
 *  to see if they are running and that they accepts connections if
 *  ports are defined. Not running processes are restarted. This
 *  function also handles the various tests per process that are
 *  defined in the control file.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Olivier Beyssac, <ob@r14.freenix.org> (check_skip)
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: validate.c,v 1.37 2003/02/18 13:55:32 hauk Exp $
 *
 *  @file
 */


/* ---------------------------------------------------------------- Public */


/**
 * Walk the process-list and validate each process
 */
void validate() {

  Process_T p;

  if(! update_loadavg()) {
    
    log("Update of loadavg has failed!\n");
    
  }
 
  for(p= processlist; p; p= p->next) {
    if(p->visited)
	continue;
    do_validate(p);
  }

  reset_depend();

}


/* --------------------------------------------------------------- Private */


/**
 * Validate the given process - If the process is not running, start
 * it. If the process listens on a port try to connect to the port and
 * test the stated protocol at the port. If any of the various tests
 * fails then do the specified action on the process.
 *
 * We block for the signals SIGTERM while in this function.
 */
static void do_validate(Process_T p) {
  
  Port_T pp;
  Resource_T pr;
  Timestamp_T tl;
  pid_t  pid= -1;
  sigset_t ns,os;
  char report[STRLEN];

  ASSERT(p);
  
  /* Obtain process action mutex */
  LOCK(p->mutex)

  /* First, check for pre-conditions */
  if(do_not_validate(p)) {
    /* Release process action mutex */
    pthread_mutex_unlock(&p->mutex);
    return;
  }
  
  /* Set a SIGTERM block */
  sigemptyset(&ns);
  sigaddset(&ns, SIGTERM);
  pthread_sigmask(SIG_BLOCK, &ns, &os);

  /* Test for running process */
  if(!(pid= is_process_running(p))) {

    /* Reset the proc info object to prevent false data in the first run */
    memset(p->procinfo, 0, sizeof *(p->procinfo));
    
    do_start(p, "Reason: Process is not running.");
    goto reinstall;
    
  } else {
    
    if(Run.debug) {
      
      log("'%s' is running with pid %d\n", p->name, (int)pid);

    }
    
  }

  if(Run.doprocess) {

    if(update_process_data(p, pid)) {
  
      if(! check_process_state(p, report)) {

	smtp_alert_resource(p, "Reason: %s\n", report);
    
      } else {
      
	if(Run.debug) {
	
	  log("'%s' check_process_state() passed.\n", p->name);
	
	}
      
      }
    
      for(pr= p->resourcelist; pr; pr= pr->next) {
      
	if(!check_resources(p, pr, report)) {
	
	  switch(pr->action) {
	  case ACTION_ALERT:
	      if(p->def_timeout) p->nstart++;
	      smtp_alert_resource(p, "Reason: %s\n", report);
	      /* We are also interested in other alerts/stops/restarts! */
	      pr->cycle=0;
	      break;

	  case ACTION_STOP:
	      do_stop(p, "Reason: %s\n", report);
	      reset_resource_counter(p);
	      goto reinstall;

	  case ACTION_RESTART:
	      do_restart(p, "Reason: %s\n", report);
	      reset_resource_counter(p);
	      goto reinstall;

	  default:
	      log("'%s' Unknow resource failure action.\n", p->name);
	      break;
	  }

	}
      
      }
	
    } else {

      log("'%s' failed to get process data\n", p->name);
      
    }

  }

  for(tl= p->timestamplist; tl; tl= tl->next) {
      
    if(!check_timestamp(p, tl, report)) {
	
      switch(tl->action) {
      case ACTION_ALERT:
          if(p->def_timeout) p->nstart++;
	  smtp_alert_timestamp(p, "Reason: %s\n", report);
	  break; /* continue */
	  
      case ACTION_STOP:
	  do_stop(p, "Reason: %s\n", report);
	  goto reinstall;

      case ACTION_RESTART:
	  do_restart(p, "Reason: %s\n", report);
	  goto reinstall;

      default:
	  log("'%s' Unknow timestamp failure action.\n", p->name);
	  break;

      }

    }

  } 	
      
  /* Test each host:port and protocol in the process's portlist */
  for(pp= p->portlist; pp; pp= pp->next) {
    
    if(!check_connection(p, pp, report)) {
      
      do_restart(p, "Reason: %s\n", report);
      goto reinstall;
      
    }
    
  }

  reinstall:
  
  /* Remove the SIGTERM block */
  pthread_sigmask(SIG_SETMASK, &os, NULL);

  /* Release process action mutex */
  END_LOCK;
 
}


/**
 * Restart the process if the start and stop
 * program is defined, and validate is TRUE.
 */
static void do_restart(Process_T p, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(p);
  
  if(!p->do_validate)
    return;

  if(p->def_timeout)
    p->nstart++;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  if(p->mode!=MODE_PASSIVE && p->start && p->stop) {

    log("Trying to restart '%s'\n", p->name);
    check_process(p->name, "restart");
    smtp_alert_restart(p, "%s", tmp);

  } else {

    smtp_alert_failed(p, "%s", tmp);

  }

  free(tmp);

}


/**
 * Start the process if the start program is defined,
 * and validate is TRUE.
 */
static void do_start(Process_T p, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(p);

  if(!p->do_validate)
    return;

  if(p->def_timeout)
    p->nstart++;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  if(p->mode!= MODE_PASSIVE && p->start) {

    check_process(p->name, "start");
    smtp_alert_restart(p, "%s", tmp);

  } else {

    smtp_alert_failed(p, "%s", tmp);

  }

  free(tmp);
  
}


/**
 * Stop the process if the stop program is defined,
 * and validate is TRUE.
 */
static void do_stop(Process_T p, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(p);

  if(!p->do_validate)
    return;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  LOCK(Run.mutex)
    p->do_validate= FALSE;
  END_LOCK;

  if(p->mode!= MODE_PASSIVE && p->stop) {

    smtp_alert_stop(p, "%s", tmp);
    check_process(p->name, "stop");

  } else {

    if(p->def_timeout)
      p->nstart++;

    smtp_alert_failed(p, "%s", tmp);

  }

  free(tmp);

}


/**
 * Returns the logical OR of pre-condition tests.
 * A variant of the template-method pattern
 */
static int do_not_validate(Process_T p) {

  ASSERT(p);
  
  return(!p->do_validate  ||
	 check_skip(p)    ||
	 check_timeout(p) ||
	 check_checksum(p));

}


/**
 * Returns TRUE if the connection and protocol test succeeded
 * otherwise FALSE.
 */
static int check_connection(Process_T p, Port_T pp, char *report) {

  volatile int rv= TRUE;

  ASSERT(p && pp);
  
  /* Control comes here if a timeout occures */
  if(sigsetjmp(timeout, TRUE)) {
    
    snprintf(report, STRLEN, 
	     "timed out when testing %s [%s]",
	     pp->address, pp->protocol->name);
    
    log("'%s' %s\n", p->name, report); 

    rv= FALSE;
    goto error;
    
  }

  /* Set a timeout handler and activate the timer */
  set_alarm_handler(connection_timeout);
  alarm(CHECK_TIMEOUT);

  /* Open a socket to the destination INET[hostname:port] or UNIX[pathname] */
  if((pp->socket= create_generic_socket(pp)) < 0) {
      
    snprintf(report, STRLEN, 
	     "does not accept connection at %s.",
	     pp->address);

    log("'%s' %s\n", p->name, report); 

    rv= FALSE;
    goto error;
    
  } else {
    
    if(Run.debug) {
      
      log("'%s' succeeded connecting to %s\n",
          p->name, pp->address);
      
    }
    
  }

  /* Verify that the socket is ready for i|o */
  if(!check_connection_io(pp)) {
    
    snprintf(report, STRLEN, 
	     "socket at %s is not ready for i|o -- %s",
	     pp->address, STRERROR);
    
    log("'%s' %s\n", p->name, report); 

    rv= FALSE;
    goto error;

  }

  if(pp->ssl != NULL) {

    if(!embed_ssl_socket(pp->ssl, pp->socket)) {

      snprintf(report, STRLEN, 
	       "failed establish SSL communication on socket at %s",
	       pp->address);
      
      log("'%s' %s\n", p->name, report); 
      rv= FALSE;
      goto error;
      
    }

    if(pp->certmd5 != NULL) {

      if(! check_ssl_md5sum(pp->ssl, pp->certmd5)) {

	snprintf(report, STRLEN, 
		 "md5sums of SSL certificates do not match at %s",
		 pp->address);
	
	log("'%s' %s\n", p->name, report); 
	rv= FALSE;
	goto error;

      }

    }

  }

  /* Run the protocol verification routine through the socket */
  if(! pp->protocol->check(pp)) {
    
    snprintf(report, STRLEN, 
	     "test with protocol [%s] failed at %s.",
	     pp->protocol->name, pp->address);

    log("'%s' %s\n", p->name, report); 
    
    rv= FALSE;
    goto error;
    
  } else {
    
    if(Run.debug) {
      
      log("'%s' succeeded testing protocol [%s] at %s\n",
          p->name, pp->protocol->name, pp->address);

    }
  }

  error:

  alarm(0);

  if(pp->ssl != NULL) {

    cleanup_ssl_socket(pp->ssl);
    
  }

  close_socket(pp->socket);
  
  return rv;
      
}


/**
 * Returns TRUE if the process state (e.g. Zombie) test succeeded
 * otherwise FALSE.
 */
static int check_process_state(Process_T p, char *report) {

  ProcInfo_T pi;

  ASSERT(p);
  
  pi= p->procinfo;
  
  if(pi->status_flag & PROCESS_ZOMBIE) {

    snprintf(report, STRLEN, 
	     "process with pid %d is a zombie\n", pi->pid);

    log("'%s' %s\n", p->name, report); 

    /* We do not check the process anymore if it's a zombie
     since such a process is (usually) unmanageable */
    LOCK(Run.mutex)
	p->do_validate= FALSE;
    END_LOCK;
    
    return FALSE;

  } else {

    if(Run.debug) {
      
      log("'%s' zombie check passed [status_flag=%04x]\n",
	  p->name,  pi->status_flag);
      
    }

  }

  return TRUE;

}


/**
 * Resets the resource counter
 */
static void reset_resource_counter(Process_T p) {
  
  Resource_T pr;

  ASSERT(p);
  
  for(pr= p->resourcelist; pr; pr= pr->next) {
    
      pr->cycle=0;

  }
  
}


/**
 * Returns TRUE if the proc test succeeded
 * otherwise FALSE.
 */
static int check_resources(Process_T p, Resource_T pr, char *report) {

  ProcInfo_T pi;
  int okay= TRUE;

  ASSERT(p);
  ASSERT(pr);

  pi= p->procinfo;
  
  switch(pr->resource_id) {
  case RESOURCE_ID_CPU_PERCENT: 
    if(compare_value(pr->operator, pi->cpu_percent, pr->limit)) {

      vlog(report, STRLEN, p, 
	   "cpu usage of %.1f%% matches resource limit [cpu usage%s%.1f%%]",
	   pi->cpu_percent/10.0, operatorshortnames[pr->operator],
	   pr->limit/10.0);

      okay= FALSE;

    } else {

      if(Run.debug) {
      
	log("'%s' cpu usage check passed [current cpu usage=%.1f%%]\n", 
	    p->name, pi->cpu_percent/10.0);

      }
    }
    break;

  case RESOURCE_ID_MEM_PERCENT: 
    if(compare_value(pr->operator, pi->mem_percent, pr->limit)) {

      vlog(report, STRLEN, p, 
	   "mem usage of %.1f%% matches resource limit [mem usage%s%.1f%%]",
	   pi->mem_percent/10.0, operatorshortnames[pr->operator],
	   pr->limit/10.0);      

      okay= FALSE;

    } else {
      
      if(Run.debug) {
	
	log("'%s' mem usage check passed [current mem usage=%.1f%%]\n", 
	    p->name, pi->mem_percent/10.0);
	
      }
    }
    break;

  case RESOURCE_ID_MEM_KBYTE: 

    if(compare_value(pr->operator, pi->mem_kbyte, pr->limit)) {

      vlog(report, STRLEN, p, 
	   "mem amount of %ldkB matches resource limit [mem amount%s%ldkB]", 
	   pi->mem_kbyte, operatorshortnames[pr->operator],
	   pr->limit);      
      
      okay= FALSE;
    
    } else {

      if(Run.debug) {
	
	log("'%s' mem amount check passed [current mem amount=%ldkB]\n", 
	    p->name, pi->mem_kbyte);
      
      }
    }
    break;

  case RESOURCE_ID_LOAD1: 

    if(compare_value(pr->operator, (int)(Run.loadavg[0]*10.0), pr->limit)) {

      vlog(report, STRLEN, p, 
	   "loadavg(1min) of %.1f matches resource limit "
	   "[loadavg(1min)%s%.1f]", 
	   Run.loadavg[0], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

      if(Run.debug) {
	
	log("'%s' loadavg(1min) check passed [current loadavg(1min)=%.1f]\n", 
	    p->name, Run.loadavg[0]);
      
      }
    }
    break;

  case RESOURCE_ID_LOAD5: 

    if(compare_value(pr->operator, (int)(Run.loadavg[1]*10.0), pr->limit)) {

      vlog(report, STRLEN, p, 
	   "loadavg(5min) of %.1f matches resource limit "
	   "[loadavg(5min)%s%.1f]", 
	   Run.loadavg[1], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

      if(Run.debug) {
	
	log("'%s' loadavg(5min) check passed [current loadavg(5min)=%.1f]\n", 
	    p->name, Run.loadavg[1]);
      
      }
    }
    break;

  case RESOURCE_ID_LOAD15: 

    if(compare_value(pr->operator, (int)(Run.loadavg[2]*10.0), pr->limit)) {

      vlog(report, STRLEN, p, 
	   "loadavg(15min) of %.1f matches resource limit "
	   "[loadavg(15min)%s%.1f]", 
	   Run.loadavg[2], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

      if(Run.debug) {
	
	log("'%s' loadavg(15min) check passed "
	    "[current loadavg(15min)=%.1f]\n", 
	    p->name, Run.loadavg[2]);
      
      }
    }
    break;

  default:

      log("'%s' error: unknow resource ID: [%d]\n", p->name, pr->resource_id);

  }

  if(okay && pr->cycle > 0) {

    pr->cycle--;

  } else if(! okay) {

    pr->cycle++;

  }

  if(pr->cycle >= pr->max_cycle) {

    return FALSE;

  }
    
  return TRUE;
  
}


/**
 * Returns TRUE if the process timed out, otherwise FALSE. 
 */
static int check_timeout(Process_T p) {

  ASSERT(p);
  
  if(!p->def_timeout) {

    return FALSE;

  }
  
  /*
   * Start counting cycles
   */
  if(p->nstart > 0) {
    
    p->ncycle++;
    
  }
  
  /*
   * Check timeout
   */
  if(p->nstart >= p->to_start && p->ncycle <= p->to_cycle) {
    
    /*
     * Got timeout - Set do not validate flag. This statement is
     * synchronized since the same data can be set from the http
     * thread
     */
    LOCK(Run.mutex)
	p->do_validate= FALSE;
    END_LOCK;
    
    /*
     * Log and notify the user that a timeout occured
     */
    log("**Alert** process '%s' timed out and will not be checked anymore.\n",
	p->name);
      
    smtp_alert_timeout(p, NULL);
    
    return TRUE;
    
  }
  
  /*
   * Stop counting and reset if the
   * cycle interval is passed
   */
  if(p->ncycle > p->to_cycle) {
    
    p->ncycle= 0;
    p->nstart= 0;
    
  }

  return FALSE;
  
}


/**
 * Returns TRUE if validation should be skiped for
 * this process in this cycle, otherwise FALSE
 */
static int check_skip(Process_T p) {

  ASSERT(p);
  
  if(!p->def_every) {

    return FALSE;

  }
  
  if(++p->nevery < p->every) {
    
      return TRUE;

  }
  
  p->nevery= 0;

  return FALSE;

}


/**
 * Returns TRUE if the checksum was changed for any associated
 * programs and send an alert warning. Also make sure that the program
 * will not be validated, started or restarted again in this
 * session. Returns FALSE if the checksum is the same or not defined
 * for this process.
 */
static int check_checksum(Process_T p) {

  Checksum_T c;

  ASSERT(p);

  if(!p->def_checksum) {

    return FALSE;

  }

  for(c= p->checksumlist; c; c= c->next) {
    
    if(! checksum_helper(p, c->file, c->md5)) {
	
      return TRUE;

    }
    
  }
  
  if(Run.debug) {
	
    log("'%s' have valid checksums\n", p->name);

  }
  
  return FALSE;

}


/**
 * Helper function that does all the checksum work; Returns TRUE if
 * the checksum is okay or not defined, otherwise issue a warning, do
 * alert and turn of any start/stop/restart action for the process and
 * return FALSE
 */
static int checksum_helper(Process_T p, char *program, char *sum) {

  int rv= TRUE;

  ASSERT(p);

  if(program && sum) {
    
    if(!check_md5(program, sum)) {
      
      log("'%s' **Alert** checksum error for %s\n", p->name, program);
      
      smtp_alert_checksum(p, NULL);
      
      LOCK(Run.mutex)
        p->do_validate= FALSE;
        p->has_checksum_error= TRUE;
      END_LOCK;
      
      rv= FALSE;
      
    }
    
  }
    
  return rv;

}


/**
 * Returns TRUE if the timestamp test succeded, otherwise FALSE
 */
static int check_timestamp(Process_T p, Timestamp_T t, char *report) {

  time_t now;
  time_t timestamp;

  ASSERT(p);
  ASSERT(t);

  if((int)time(&now) == -1) {
    vlog(report, STRLEN, p, "can't get actual time");      
    return FALSE;
  }

  if(!(timestamp= get_timestamp(t->pathname, S_IFDIR|S_IFREG))) {
    vlog(report, STRLEN, p, "can't get timestamp for %s", t->pathname);      
    return FALSE;
  }

  if(compare_value(t->operator, (int)(now - timestamp), t->time)) {
    vlog(report, STRLEN, p, "timestamp test failed for %s", t->pathname);      
    return FALSE;
  }

  if(Run.debug)
    log("'%s' timestamp test passed for %s\n", p->name, t->pathname); 
      
  return TRUE;

}


/**
 * Signal handler for connection timeout
 */
static void connection_timeout(int sig) {
  
  siglongjmp(timeout, TRUE);
  
}


/**
 * Logging with report generation
 */
static void vlog(char * report, int n, Process_T p, char *m,...) {

  va_list ap;
  char *tmp = NULL;

  va_start(ap, m);

  if(m) { 

    tmp=format(m,ap);

  }

  va_end(ap);

  strncpy(report, tmp, n);
  log("'%s' %s\n", p->name, report); 

  free(tmp);
}


/**
 * Comparison of values. Returns TRUE if comparison matches, otherwise
 * FALSE.
 */
static int compare_value(int operator, int left, int right) {

    switch(operator) {

    case OPERATOR_GREATER:

      if(left > right)
	return TRUE;
      break;

    case OPERATOR_LESS:

      if(left < right)
	return TRUE;
      break;

    case OPERATOR_EQUAL:

      if(left == right)
	return TRUE;
      break;

    case OPERATOR_NOTEQUAL:

      if(left != right)
	return TRUE;
      break;

    default:
      error("Unknow comparison operator\n");
      return FALSE;

    }

    return FALSE;

}

